//-------------------------------------------------------------------
//
//                         AGD.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------


#ifndef _AGD_AGD_H
#define _AGD_AGD_H


// AGD version macro

#define __AGD__ 083



#include <AGD/Keymapper.h>
#include <LEDA/graph_alg.h>


class AGDModule;
class KeySet;

leda_list<leda_edge> onehalf (leda_graph &GO, bool STRENG_COND = true);


/*{\Manpage {AGD} {} {AGD Base} {agd} }*/

class AGD {

	friend AGDModule;
	friend KeySet;

/*{\Mdefinition
	
The class |\Mname| provides a global key mapper, that maps strings
to key items of type |agd_key|, and a number of predefined key constants.
Furthermore, it defines static methods for testing properties
of graphs, and for augmentation (partially wrappers for LEDA functions).

We use key items to represent properties of instances of a particular type,
for example a specific instance |G| of type |graph| could have the
property |planar| meaning that |G| is a planar graph. We represent such
properties by arbitrary strings, which are stored in a global hash table by
the default keymapper of |\Mname|. Strings in this hash table are accessed via key items
of type |agd_key|. We use $\<s\>$ to denote a key item for string $s$.

The subclass |key| of |\Mname| defines a number of key constants
(e.g., |\Mname|::|key|::|directed|):\nopagebreak

\begin{tabular}{|p{2.75cm}|p{3cm}|p{9cm}|} \hline
\multicolumn{3}{|c|}{\rule[-2mm]{0mm}{6mm}\bf General properties}\\\hline
|animation|  & ``animation'' & A layout algorithm supports step by step animation.\\
|grid_animation|  & ``grid animation'' & A grid layout algorithm supports step by step
animation on the grid.\\
|planar_layout|  & ``planar layout'' & A layout algorithm works on a plane graph.\\\hline
\end{tabular}

\begin{tabular}{|p{2.75cm}|p{3cm}|p{9cm}|} \hline
\multicolumn{3}{|c|}{\rule[-2mm]{0mm}{6mm}\bf Properties of layouts}\\\hline
constant                & string value & meaning\\\hline\hline
|straight_line| & ``straight-line'' & All edge polylines have no bend points.\\
|orthogonal|    & ``orthogonal'' & All edge segments are either horizontal
or vertical.\\
|no_crossings|  & ``no crossings'' & The edge polylines do not intersect.\\
|upward_planar| & ``upward planar'' & All edges are directed upwards (downwards).\\
|convex| & ``convex'' & All faces are convex.\\\hline
\end{tabular}

\begin{tabular}{|p{2.75cm}|p{3cm}|p{9cm}|} \hline
\multicolumn{3}{|c|}{\rule[-2mm]{0mm}{6mm}\bf Properties of graphs}\\\hline
constant                & string value & meaning\\\hline\hline
|directed|      & ``directed'' & |G| is directed.\\
|undirected|    & ``undirected'' & |G| is undirected.\\
|planar|        & ``planar'' & |G| is planar.\\
|simple|        & ``simple'' & |G| has no parallel edges.\\
|no_self_loops| & ``no self loops'' & |G| has no self loops.\\
|no_reversals|  & ``no reversals'' & $(v,w)\in E\Rightarrow (w,v)\notin E$
	for all $v,w\in V$\\
|connected|     & ``connected''     & |G| is connected.\\
|biconnected|   & ``biconnected''   & |G| is biconnected.\\
|triconnected|  & ``triconnected''  & |G| is triconnected.\\
|bipartite|     & ``bipartite''     & |G| is bipartite.\\
|four_graph|    & ``four-graph''    & $|degree(v)|
	\leq 4$ for all $v \in V$\\
|acyclic|       & ``acyclic''       & $v_1,\ldots,v_k \in V, k\geq 3,
	(v_i,v_{i+1})\in E \Rightarrow v_1 \neq v_k$\\
|forest|        & ``forest''        & |G| represents a forest.\\
|tree|          & ``tree''          & |G| represents a tree.\\
|single_source| & ``single source'' & $\exists!\: v\in G : |indeg|(v)=0$\\
|single_sink|   & ``single sink''   & $\exists!\: v\in G : |outdeg|(v)=0$\\\hline
\end{tabular}\nopagebreak

where $G = (V,E)$.
}*/

public:
	enum Status { yes, no, unknown };

public:

	AGD ();

	virtual ~AGD () { }


	/*{\Moperations 2 3.2}*/


	// constants for predefined keys
	struct key {
		static const agd_key planar, simple, connected, biconnected,
			triconnected, four_graph, no_self_loops, directed, undirected,
			no_reversals, acyclic, tree, forest, single_source,
			single_sink, bipartite, straight_line,
			orthogonal, no_crossings,
			upward_planar, convex, animation, grid_animation, planar_layout;
	};

	friend AGD::key;


protected:


public:

	/*{\Mtext       
	\headerline{a) Testing Properties of Graphs} }*/

	static inline bool is_simple (const leda_graph& G) {
		return (G.number_of_edges() == 0) ? true : Is_Simple (G);
	}
	/*{\Mstatic returns |true| if $G$ is simple, e.g. has no parallel
	edges, |false| otherwise.}*/

	static inline bool is_connected (const leda_graph& G) {
		return (G.number_of_nodes() == 0) ? true : Is_Connected (G);
	}
	/*{\Mstatic returns |true| if $G$ is connected, |false| otherwise.}*/

	static inline bool is_biconnected  (const leda_graph& G) {
		return Is_Biconnected (G);
	}
	/*{\Mstatic returns |true| if $G$ is biconnected, |false| otherwise.}*/

	static bool is_triconnected (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is triconnected, |false| otherwise.}*/

	static bool is_planar       (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is planar, |false| otherwise.}*/

	static bool is_bipartite    (const leda_graph& G) {
		return Is_Bipartite (G);
	}
	/*{\Mstatic returns |true| if $G$ is bipartite, |false| otherwise.}*/

	static bool is_four_graph   (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is a four-graph, e.g. |degree|$(v)
	\leq 4$ holds for every node $v$ of $G$, |false| otherwise.}*/

	static bool has_no_self_loops (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ contains no self loops, |false|
	otherwise.}*/

	static bool has_no_reversals (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is undirected, or if $G$ contains
	no reversal edges, |false| otherwise.}*/

	static bool is_acyclic (const leda_graph& G) {
		return Is_Acyclic (G);
	}
	/*{\Mstatic returns |true| if $G$ is acyclic, |false| otherwise.}*/

	static bool is_forest (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ represents a forest, |false|
	otherwise.}*/

	static bool is_tree (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ represents a tree, |false|
	otherwise.}*/

	static bool has_single_source (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one node with no incoming edges, |false| otherwise.}*/

	static bool has_single_sink (const leda_graph& G);
	/*{\Mstatic returns |true| if $G$ is directed and contains exactly
	one node with no outgoing edges, |false| otherwise.}*/


	/*{\Mtext
	\bigskip
	\headerline{b) Establishing Properties of Graphs} }*/

	static inline leda_list<leda_edge> triangulate (leda_graph& G) {
		return TRIANGULATE_PLANAR_MAP (G);
	}
	/*{\Mstatic triangulates a directed graph $G$ representing a planar map.
	Returns the list of inserted edges.}*/

	static inline leda_list<leda_edge> make_bidirected (leda_graph& G) {
		return Make_Bidirected (G);
	}
	/*{\Mstatic makes $G$ bidirected by inserting missing reversal edges.
	Returns the list of inserted edges.}*/

	static inline leda_list<leda_edge> make_biconnected (leda_graph& G) {
		return Make_Biconnected (G);
	}
	/*{\Mstatic makes $G$ biconnected.
	Returns the list of inserted edges.}*/

	static inline leda_list<leda_edge> augment_planar_biconnected (leda_graph& G) {
		return onehalf (G, true);
	}
	/*{\Mstatic augments a planar graph $G$ to a planar and biconnected graph.
	Returns the list of inserted edges.}*/

	static inline leda_list<leda_edge> make_map (leda_graph& G) {
		leda_list<leda_edge> R;
		G.make_map (R);
		return R;
	}
	/*{\Mstatic makes $G$ bidirected and turns it into a map. Returns the
	list of inserted edges.}*/

	static inline bool embed_planar_graph (leda_graph& G) {
		return PLANAR (G, true);
	}
	/*{\Mstatic takes as input a bidirected graph $G$. Returns |true| and
	turns $G$ into a planar map if $G$ is planar, |false| otherwise.}*/


	/*{\Mtext
	\bigskip
	\headerline{c) Access to the Default Keymapper} }*/

	static int number_of_keys() {
		return keymapper.number_of_keys();
	}
	/*{\Mstatic returns the number of defined keys.}*/

	static agd_key ask_key (const leda_string& s) {
		return keymapper.ask_key (s);
	}
	/*{\Mstatic returns the key \<|s|\> if \<|s|\> is already
	defined, |nil| otherwise.}*/

	// Iteration

	static agd_key first_key() {
		return keymapper.first_key();
	}

	static agd_key next_key (agd_key k) {
		return keymapper.next_key (k);
	}

protected:

	static agd_key get_key (const leda_string& s) {
		return keymapper.get_key (s);
	}
	/*{\Mstatic returns the key \<|s|\>.}*/

public:

	static KeySet get_properties (const leda_graph& G);
	/*{\Mstatic returns the set of all properties that |G| satisfies.}*/

private:
	// the keymapper maps strings to keys (items)
	// there is exactly one keymapper for all AGD modules
	static Keymapper keymapper;

	static void extract_list (const leda_string& str, char delim,
		leda_list<leda_string>& L);

	static bool dfs_iua (leda_node v, leda_node father,
		leda_node_array<bool>& visited);

};

/*{\Mtext
\bigskip
\headerline{Iteration}
{\bf forall\_AGD\_keys}($k$) \\
\hspace*{2em}$\{$  ``the keys of |\Mname|'s default keymapper are successively assigned to $k$''  $\}$
}*/

#define forall_AGD_keys(k) \
	for((k)=AGD::first_key(); (k) != nil; (k) = AGD::next_key(k))



#endif
