//-------------------------------------------------------------------
//
//                        EdgeInsertModule.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_EDGE_INSERT_MODULE_H
#define _AGD_EDGE_INSERT_MODULE_H


#include <AGD/AGDModule.h>
#include <AGD/PlaneGraphCopy.h>


/*{\Manpage {EdgeInsertModule} {} {Edge Insert Modules} {I} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{0.5}{0}{6.5}{EdgeInsertModule}
\end{hierarchy}
}*/

class EdgeInsertModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for edge insert modules. An
edge insert module gets as input a planarized graph |PG|, which is
a copy of a graph $G$, and a list of edges
$e_1,\ldots,e_k\in G$. It inserts $e_1,\ldots,e_k$
into |PG| by replacing edge crossings with dummy nodes.

Edge insert modules are used for computing a planarization with
a planar subgraph. The edges that are not in the planar subgraph
are reinserted by means of an edge insert module
(see~\ref{Planarization with Planar Subgraph}).

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & $|in_graph|=1$ \\
 & edge list & $|in_list|=2$ \\
output parameter: & graph & $|out_graph|=1$
\end{tabular}

The integer constants |in_graph|, |in_list| and |out_graph| can be used to denote
the positions of the input and output parameters.

\headerline{Implementing Edge Insert Modules}

An implementation of an edge insert module must override the protected
method call\_insert(|PG|, |list<edge>|\& |e_insert|),
which gets as input a planarized graph $PG$, and a
list of edges $|e_insert|=e_1,\ldots,e_k$, which must be inserted.

The planarized graph $PG$ is a copy of $G$, and
$e_1,\ldots,e_k\in G$. For each edge $e_i$, the implementation
inserts a chain of edges starting at $|copy|(|source|(e_i))$ and ending at
$|copy|(|target|(e_i))$ that leads over dummy nodes. The dummy nodes are
new nodes resulting from splitting an edge, which will represent edge
crossings in a drawing of the graph.

The class |PlaneGraphCopy| stores these chains and provides methods
(split, new\_edge\_chain) for reinserting edges.
}*/

public:

	// the input and output parameters of a planarization module
	static const int in_graph, in_list;
	static const int out_graph;

	/*{\Minitialization I 4.5}*/

	EdgeInsertModule () : AGDModule(2,1) { }

	/*{\Mcreate initializes an edge insert module.}*/

	virtual ~EdgeInsertModule () { }


	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const PlaneGraphCopy& G, leda_list<leda_edge>& e_insert, leda_string& message);
	/*{\Mop returns |true| if $G$ satisfies the precondition of
	|\Mvar|. Otherwise, |false| is returned, and |message| contains the
	error message.}*/

	void call (PlaneGraphCopy& PG, leda_list<leda_edge>& e_insert);
	/*{\Mop calls the edge insert module for graph $PG$.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void _call (PlaneGraphCopy& G, leda_list<leda_edge>& e_insert)
	/*{\Moppure implements the edge insert module for the planarized graph $PG$,
	and the list of edges |e_insert|.}*/
	= 0;

};


inline void EdgeInsertModule::call (PlaneGraphCopy& G, leda_list<leda_edge>& e_insert)
{
	_call (G, e_insert);
}

inline bool EdgeInsertModule::check (const PlaneGraphCopy& G,
	leda_list<leda_edge>& /*e_insert*/, leda_string& message)

{
	return check_graph (in_graph, G, message);
}


#endif
