//-------------------------------------------------------------------
//
//                         FPPLayout.h
//
//    author: Carsten Gutwenger
//    date:   octobre 1997
//-------------------------------------------------------------------


#ifndef _AGD_FPP_LAYOUT_H
#define _AGD_FPP_LAYOUT_H


#include <AGD/GridLayoutModule.h>


/*{\Manpage {FPPLayout} {} {The de Fraysseix, Pach, Pollack Algorithm}
{L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{4.5}{0}{5}{FPPLayout}
\end{hierarchy}
\hfill\agdeps{fpp.epsi}{3.8cm}
}*/

class FPPLayout : public GridLayoutModule {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm by de Fraysseix, Pach,
Pollack~\cite{dpp-hdpgg-90}.

This algorithm draws a planar graph |G| straight-line
without crossings. |G| may not contain self-loops or mulitple edges.
The grid layout size is at most $(2n-4) \times (n-2)$ for a graph with
$n$ nodes ($n \geq 3$).

\horizline
\commoninfo{de Fraysseix et al.}{de Fraysseix, Pach, Pollack}
{de Fraysseix, Pach, Pollack}{Carsten Gutwenger}{february 1997}{1.0}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\{\, |planar|,|simple|,|no self loops| \,\}$ \\
postcondition(|PRE|) & $=$ & $\{\, \mbox{|straight|-|line|},|no crossings| \,\}$
\end{tabular}

\horizline
}*/

public:

	/*{\Mcreation}*/

	FPPLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~FPPLayout() { }


	/*{\Moperations 1.5 2.5}*/

	leda_string name () const {
		return "de Fraysseix et al.";
	}

	leda_string long_name () const {
		return "de Fraysseix, Pach, Pollack";
	}

	leda_string author () const {
		return "de Fraysseix, Pach, Pollack";
	}

	leda_string impl_author () const {
		return "Carsten Gutwenger";
	}

	leda_string impl_version () const {
		return "1.0";
	}

	leda_string impl_date () const {
		return "february 1997";
	}

	AGDModule *clone () const;


	/*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (PlaneGraphCopy& PG, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (PlaneGraphCopy& PG, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect get_grid_area ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_coordinate_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/

protected:
	void _call_grid (PlaneGraphCopy& PG, leda_face F, GridLayout &gl);

private:
	void compute_order (const leda_graph &G,
		leda_node_array<int>&  num,
		leda_node_array<leda_edge>& e_wp,
		leda_node_array<leda_edge>& e_wq,
		leda_edge e_12, leda_edge e_2n, leda_edge e_n1);

	void compute_coordinates (const leda_graph &G,
		leda_node_array<int>&  num,
		leda_node_array<leda_edge>& e_wp,
		leda_node_array<leda_edge>& e_wq,
		leda_node_map<int>&  xpos,
		leda_node_map<int>&  ypos);

};

/*{\Mimplementation
The computation of
the layout takes time $O(n)$ where $n$ is the number of nodes
of the drawn graph.}*/


#endif
