//-------------------------------------------------------------------
//
//                        FastHierarchyLayout.cpp
//
//
//    author: Christoph Buchheim
//    date:   march 1998
//-------------------------------------------------------------------


#ifndef _FastHierarchyLayout
#define _FastHierarchyLayout

#include <AGD/HierarchyLayoutModule.h>
#include <LEDA/list.h>

class LayoutInterface;
class GraphCopy;


/*** class withKey ***/

class withKey {

public:

  int element;
  double key;

  withKey operator=(const withKey& wk) {
    element=wk.element;
    key=wk.key;
    return *this;
  }
  friend ostream& operator<<(ostream& out,const withKey& wk) {
    out<<wk.element<<"("<<wk.key<<")";
    return out;
  }
  friend istream& operator>>(istream& in,withKey& wk) {
    in>>wk.element>>wk.key;
    return in;
  }
  friend int cmpWithKey(const withKey& wk1,const withKey& wk2) {
    if(wk1.key<wk2.key) return -1;
    if(wk1.key>wk2.key) return 1;
    return 0;
  }

};

/*** class kList ***/

class kList : public leda_list<withKey> {

  list_item actual;

public:

  bool pop(int&,double&);
  double peek();
  bool add(int,double);
  double median() const;
  void reduce(kList&);

};

/*** class FastHierarchyLayout ***/

/*{\Manpage {FastHierarchyLayout} {} {Fast Layout of Hierarchies} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{0}{1.5}{7.5}{HierarchyLayoutModule}
\arhook{7.5}{2}{8.5}
\hdclass{1.5}{0}{7.5}{FastHierarchyLayout}
\end{hierarchy}
}*/

class FastHierarchyLayout : public HierarchyLayoutModule {

/*{\Mdefinition
The class |\Mname| represents a simple layout algorithm for
hierarchies with given order of nodes on each layer. 
Thus it can be used as a third phase for the Sugiyama
algorithm.

In the layout, all edges will have at most two bends. Additionally,
for each edge having exactly two bends, the section between them is 
drawn vertically. This applies in particular to the long edges
arising in the first phase of the Sugiyama algorithm.

\horizline
\commoninfo{Phase 3}{Sugiyama Algorithm, Phase 3}{}{Christoph Buchheim}
{march 1998}{}

\headerline{Pre- and Postcondition}
\begin{tabular}{l@{ = }l}
precondition & $\emptyset$ \\
postcondition(|PRE|) & $\emptyset$
\end{tabular}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf node\_distance}\\
determines the minimal horizontal distance between two nodes on the
same layer.
\item |double| {\bf layer\_distance}\\
determines the minimal vertical distance between two nodes on
neighbored layers.
\item |bool| {\bf fixed\_layer\_distance}\\
if set to |true|, the distance between neighbored layers is fixed.
\end{options}\horizline
}*/

public:

	/*{\Mcreation L 4.5}*/

  FastHierarchyLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

  ~FastHierarchyLayout();


	/*{\Moperations 1.2 3}*/

  leda_string name() const;
  leda_string long_name() const;
  leda_string author() const;
  leda_string impl_author() const;
  leda_string impl_date() const;

  double node_distance() const;
  void node_distance(double);
  double layer_distance() const;
  void layer_distance(double);
  bool fixed_layer_distance() const;
  void fixed_layer_distance(bool);

  void call(const GraphCopy&,const Hierarchy&,LayoutInterface&);
	/*{\Mop calls the |\Mname| module for graph $GC$ and hierarchy
	$H$. Assigns the layout to $A$.}*/

  AGDModule *clone() const;

private:

  int n,m,k;
  int *layer,*first;
  leda_list<int> *adj[2],**longEdge;
  double minNodeDist,minLayerDist;
  bool fixedLayerDist;
  double *breadth,*height,*y,*x,*totalB,*mDist;
  bool *virt;

  void decrTo(double&,double);
  void incrTo(double&,double);
  bool sameLayer(int,int) const;
  bool isFirst(int) const;
  bool isLast(int) const;
  void sortLongEdges(int,int,double*,bool&,double&,int*,bool*);
  bool placeSingleNode(int,int,int,double&,int);
  void placeNodes(int,int,int,int,int);
  void moveLongEdge(int,int,bool*);
  void straightenEdge(int,bool*);
  void findPlacement();

};

#endif
