//-------------------------------------------------------------------
//
//                       GraphCopyLI.h
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_GRAPH_COPY_LI_H
#define _AGD_GRAPH_COPY_LI_H


#include <AGD/LayoutBuffer.h>
#include <AGD/GraphCopy.h>


/*{\Manpage {GraphCopyLI} {} {Layout Interface for Graph Copies} {A} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{LayoutInterface}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutBuffer}
\arhook{6.5}{2}{7.5}
\hdclass{3}{0}{5}{GraphCopyLI}
\end{hierarchy}
}*/

class GraphCopyLI : public LayoutBuffer {

/*{\Mdefinition
An instance |\Mvar| of type |\Mname| is a temporary buffer for storing
graphical attributes of a planarized graph |PG| that is a copy of a
graph $G$. The width and height of each node are accessed through a
layout interface $A$ for $G$. The update method assigns the computed layout
to $A$ by transforming dummy nodes in |PG| into bend points.
}*/

public:

	/*{\Mcreation}*/

	GraphCopyLI (GraphCopy& GC, LayoutInterface& A) :
		LayoutBuffer(GC)
	{
		gc_p = &GC;
		a_p  = &A;
	}
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a layout buffer for planarized graph |PG| with layout interface |A|.}*/

	~GraphCopyLI() { }


	/*{\Moperations}*/

	double get_width  (leda_node v) const {
		leda_node v_orig = gc_p->original(v);
		return (v_orig != nil) ? a_p->get_width (v_orig) : 0;
	}
	/*{\Mop returns the width of the graphics of node |original(v)|
	if $|original(v)|\neq |nil|$, $0$ otherwise.}*/

	void set_width  (leda_node v, double new_w) {
		leda_node v_orig = gc_p->original(v);
		if (v_orig != nil) a_p->set_width (v_orig, new_w);
	}
	/*{\Mop sets the width of the graphics of node |original(v)| to
	|new_w| if $|original(v)|\neq |nil|$.}*/

	double get_height (leda_node v) const {
		leda_node v_orig = gc_p->original(v);
		return (v_orig != nil) ? a_p->get_height (v_orig) : 0;
	}
	/*{\Mop returns the height of the graphics of node |original(v)|
	if $|original(v)|\neq |nil|$, $0$ otherwise.}*/

	void set_height (leda_node v, double new_h) {
		leda_node v_orig = gc_p->original(v);
		if (v_orig != nil) a_p->set_height (v_orig, new_h);
	}
	/*{\Mop sets the height of the graphics of node |original(v)| to
	|new_h| |new_w| if $|original(v)|\neq |nil|$..}*/


	void update();
	/*{\Mop assigns the stored layout to $A$ and calls the update method
	of $A$, where $A$ is the layout interface for the original graph $G$.}*/


protected:
	GraphCopy *gc_p;
	LayoutInterface *a_p;
};


#endif
