//-------------------------------------------------------------------
//
//                   GridCoordinateMapper.h
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#ifndef _AGD_GRID_COORDINATE_MAPPER_H
#define _AGD_GRID_COORDINATE_MAPPER_H


#include <AGD/IPolyline.h>
#include <AGD/IRect.h>
#include <AGD/LayoutInterface.h>
#include <AGD/GridLayout.h>
#include <LEDA/node_map.h>


class LayoutInterface;
class GraphCopy;
class GraphCopyLI;


/*{\Manpage {GridCoordinateMapper} {} {Grid Coordinate Mapper} {M} }*/

class GridCoordinateMapper {
/*{\Mdefinition
	|\Mname| is the base class for all grid coordinate mappers. Grid coordinate
	mappers are used to transform a grid layout of a graph into a
	real layout.
	
	We denote with $G_A$ the graph for which the grid coordinate mapper
	computes the real layout, and with $G_B$ the graph for which the
	grid layout is computed. $G_B$ may be equal to $G_A$, or a copy
	(|GraphCopy| or |PlanarizedGraph|) of $G_A$.

	\headerline{Implementation of Grid Coordinate Mappers}

	We refer to the input grid layout with $|column|(v)$, $|row|(v)$ for the position of
	node $v$, $|ip|_1^e,\ldots,|ip|_{k(e)}^e$ for the polyline of edge $e$,
	where $|ip|_i^e = (|col|_i^e,|row|_i^e)$, with $|width|(v)$, $|height|(v)$
	for the size of the graphics of node $v$, and with |xmin|, |xmax|, |ymin|,
	|ymax| for the bounding box of the grid layout.

	The output is given by $x(v)$, $y(v)$ for the position of node $v$,
	$|line|(e)$ for the polyline of edge $e$, and $|width|'(v)$,
	$|height|'(v)$ for the new width and height of node $v$.

	Basically, a grid coordinate mapper computes a mapping
	$\varphi : \nat^2 \rightarrow \real^2$ from the coordinates of
	a grid coordinate system into the coordinates of the standard coordinate
	system (see \ref{Layout Interfaces}, The coordinate system):
	\[ \varphi(|column|,|row|) = \left( \varphi_x(|column|),\varphi_y(|row|)
	\right) \]
	That means, the mapping of the x-coordinate is independent of the
	y-coordinate and vice versa.

	Usually, it is sufficient to define $\varphi_x$ and $\varphi_y$ by implementing
	map\_x and map\_y. The default
	implementations for the output will then return

	\centerline{\begin{minipage}{0.37\textwidth}
	\begin{eqnarray*}
	x(v) & = & \varphi_x\left(|column|(v)\right) \\
	y(v) & = & \varphi_y\left(|row|(v)\right) \\
	|line|(e) & = & \varphi\left(|ip|_1^e\right),\ldots,\varphi
	\left(|ip|_{k(e)}^e\right)
	\end{eqnarray*}
	\end{minipage}\hspace{1cm}\begin{minipage}{0.37\textwidth}
	\begin{eqnarray*}
	|width|'(v) & = & |width|(v) \\
	|height|'(v) & = & |height|(v)
	\end{eqnarray*}
	\end{minipage}}

	The mapping of the coordinates works as shown below. You must
	override the methods that have no default implementation, and
	you can change the default implementations if necessary, e.g.,
	if you want to change the size of nodes or to assign special
	edge anchors.

	\begin{enumerate}
	\item {\bfseries Initialization}

	The |init| method is called. The default implementation does nothing.

	\item {\bfseries Mapping of node coordinates}

	The |map_nodes| method is called. The default implementation computes
	the position of the nodes by calling |map_node| for each node in $G_A$.

	The default implementation of |map_node| calls |map_x| and |map_y|
	to transform grid coordinates in real coordinates.

	\item {\bfseries Mapping of edge coordinates}

	The |map_edges| method is called. The default implementation computes
	the position of the edges by calling |map_edge| for each edge in $G_A$.

	The default implementation of |map_edge| maps each bend point by
	calling |map_x| and |map_y|, and sets the source and target
	anchors to $(0,0)$.

	\item {\bfseries Computing the bounding box of the layout}

	The method |get_layout_size| is called.

	\item {\bfseries Cleaning up}

	Finally, the |cleanup| method is called. The default implementation
	does nothing.
	\end{enumerate}
}*/

public:

	/*{\Minitialization}*/

	GridCoordinateMapper() { }
	/*{\Mcreate initializes a grid coordinate mapper.}*/

	virtual ~GridCoordinateMapper() { }


	/*{\Moperations}*/

	/*{\Mtext       
	\headerline{a) Standard Interface}}*/

	DRect call (const leda_graph& G, const GridLayout &gl, const IRect& gbbox,
		LayoutInterface& A);
	/*{\Mop calls |\Mvar| for graph $G$ and grid layout |gl|. Assigns the
	transformed layout to |A| and returns its bounding box. \precond |gl|
	is a grid layout for $G$, |gbbox| is the bounding box of |gl|, and |A|
	is a layout interface for |G|.}*/

	DRect call (const GraphCopy& GC, const GridLayout &gl,
		const IRect& gbbox, LayoutInterface& A);
	/*{\Mop calls |\Mvar| for graph copy $GC$ and grid layout |gl|. Assigns the
	transformed layout to |A| and returns its bounding box. \precond |gl|
	is a grid layout for $GC$, |gbbox| is the bounding box of |gl|, and |A|
	is a layout interface for |original(GC)|.}*/


	/*{\Mtext       
	\bigskip
	\headerline{b) Access Methods} }*/

	const leda_graph& graph_A() const {
		return *_graph_A;
	}
	/*{\Mop returns a reference to $G_A$.}*/

	const leda_graph& graph_B() const {
		return *_graph_B;
	}
	/*{\Mop returns a reference to $G_B$.}*/

	leda_node in_B (leda_node v) const;
	/*{\Mop returns the node in $G_B$ that corresponds to $V$.
	\precond $v$ is a node in $G_A$.}*/

	int x (leda_node v) const {
		return _gl->x(in_B(v));
	}
	/*{\Mop returns the x-coordinate of $v$ in the grid layout.
	\precond $v\in G_A$.}*/

	int y (leda_node v) const {
		return _gl->y(in_B(v));
	}
	/*{\Mop returns the y-coordinate of $v$ in the grid layout.
	\precond $v\in G_A$.}*/

	IPolyline bends (leda_edge e);
	/*{\Mop returns the list of bend points of $e$ in the grid
	layout. \precond $e$ is an edge in $G_A$.}*/

	const IRect& gbbox() {
		return _gbbox;
	}
	/*{\Mop returns the bounding box of the grid layout.}*/

	double width (leda_node v) {
		return _A->get_width(v);
	}
	/*{\Mop returns the width of $v$. \precond $v$ is a node in $G_A$.}*/

	void width (leda_node v, double w) {
		_A->set_width(v,w);
	}
	/*{\Mop sets the width of $v$ to $w$. \precond $v$ is a
	node in $G_A$.}*/

	double height (leda_node v) {
		return _A->get_height(v);
	}
	/*{\Mop returns the height of $v$. \precond $v$ is a node in $G_A$.}*/

	void height (leda_node v, double h) {
		_A->set_height(v,h);
	}
	/*{\Mop sets the height of $v$ to $h$. \precond $v$ is a
	node in $G_A$.}*/


	/*{\Mtext       
	\bigskip
	\headerline{c) Overridables} }*/

	virtual void init() { }
	/*\Mop called for initialization.}*/

	virtual void map_nodes();
	/*{\Mop called for assigning node coordinates.}*/

	virtual void map_edges();
	/*{\Mop called for assigning edge polylines.}*/

	virtual void get_layout_size (DRect& bbox)
	/*{\Moppure called to get the bounding box of the layout.}*/
	= 0;

	virtual void cleanup() { }
	/*{\Mop called for cleaning up.}*/

	virtual DPoint map_node (leda_node v);
	/*{\Mop returns the position of $v$. \precond $v$ is a node in $G_A$.}*/

	virtual void map_edge (leda_edge e, DPolyline& dpl,
		DPoint& s_a, DPoint& t_a);
	/*{\Mop returns the polyline for edge $e$ in |dpl| (bend points),
	|s_a| (source anchor point) and |t_a| (target anchor point).
	\precond $e$ is an edge in $G_A$.}*/

	virtual double map_x (int column) const
	/*{\Moppure returns $\varphi_x(|column|)$.}*/
	= 0;

	virtual double map_y (int row) const
	/*{\Moppure returns $\varphi_y(|row|)$.}*/
	= 0;


	virtual GridCoordinateMapper *clone() const
	/*{\Moppure returns a pointer to a new |\Mname| instance with
	the same option settings as |\Mvar|.}*/
	= 0;

private:
	int _graph_type; // 0 = graph, 2 = GraphCopy

	const leda_graph *_graph_A, *_graph_B;
	const GraphCopy *_PG;
	LayoutInterface *_A;

	const GridLayout *_gl;
	IRect _gbbox;

	DRect _do_call (const GridLayout &gl, const IRect& gbbox);
};



#endif
