//-------------------------------------------------------------------
//
//                         GridLayout.h
//
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#ifndef _GRID_LAYOUT_H
#define _GRID_LAYOUT_H

#include <LEDA/graph.h>
#include <AGD/IPolyline.h>
#include <AGD/IRect.h>

class GraphCopy;


/*{\Manpage {GridLayout} {} {Grid Layouts} {gl} }*/

class GridLayout {

/*{\Mdefinition
A grid layout |\Mvar| of a graph $G$ stores for each node of $G$ an
integer x- and y-coordinate, and for each edge of $G$ a list of
bend points of type |IPolyline|.
}*/

	leda_node_map<int> _x, _y;
	leda_edge_map<IPolyline> _bends;
	const leda_graph *gp;

public:

	/*{\Mcreation}*/

	GridLayout() { gp = nil; }
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	the empty layout with no associated graph.}*/

	GridLayout(const leda_graph &G) : _x(G,0), _y(G,0), _bends(G) {
		gp = &G;
	}
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname| initialized to
	a grid layout for graph $G$. All x- and y-coordinates are initially $0$,
	and all bend point lists are empty.}*/


	/*{\Moperations}*/

	void init();
	/*{\Mop reinitializes |\Mvar| to the empty layout with no
	associated graph.}*/

	void init(const leda_graph &G);
	/*{\Mop reinitializes |\Mvar| to a grid layout for graph $G$. All x-
	and y-coordinates are set to $0$, and all bend point lists are empty.}*/

	const int& x(leda_node v) const {
		return _x[v];
	}
	int& x(leda_node v) {
		return _x[v];
	}
	/*{\Mop returns a reference to the x-coordinate of node $v$.}*/

	const int& y(leda_node v) const {
		return _y[v];
	}
	int& y(leda_node v) {
		return _y[v];
	}
	/*{\Mop returns a reference to the y-coordinate of node $v$.}*/

	const IPolyline &bends(leda_edge e) const {
		return _bends[e];
	}
	IPolyline &bends(leda_edge e) {
		return _bends[e];
	}
	/*{\Mop returns a reference to the bend point list of edge $e$.}*/

	leda_node_map<int> &x() {
		return _x;
	}
	/*{\Mop returns a reference to the node map storing the
	x-coordinates of the nodes.}*/

	leda_node_map<int> &y() {
		return _y;
	}
	/*{\Mop returns a reference to the node map storing the
	y-coordinates of the nodes.}*/

	leda_edge_map<IPolyline> &bends() {
		return _bends;
	}
	/*{\Mop returns a reference to the edge map storing the
	bend point lists of the edges.}*/

	IRect area() const;
	/*{\Mop returns a tight bounding box of the grid layout.}*/

	int number_of_bends() const;
	/*{\Mop returns the total number of bend points.}*/

	IPolyline get_compact_bends(leda_edge e) const;
	/*{\Mop returns the bend point list for edge $e$ without
	unnecessary bends.}*/

	IPolyline get_compact_bends(const GraphCopy& GC, leda_edge e) const;
	/*{\Mop returns the bend point list for edge $e$ of graph copy
	|GC| without unnecessary bends. \precond |\Mvar| is a grid layout
	for |GC| and $e$ is an edge in $|original|(e)$.}*/

	void compact_all_bends();
	/*{\Mop removes all unnecessary bends.}*/

	void supplement_map_layout();
	
private:
	static bool _is_redundant(IPoint &p1, IPoint &p2, IPoint &p3);
	static void _compact(IPolyline& ip);

};


#endif
