
//-------------------------------------------------------------------
//
//                       GridLayoutModule.h
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#ifndef _AGD_GRID_LAYOUT_H
#define _AGD_GRID_LAYOUT_H


#include <AGD/LayoutModule.h>
#include <AGD/MaxAllMapper.h>
#include <AGD/IRect.h>

class PlaneGraphCopy;


/*{\Manpage {GridLayoutModule} {} {Grid Layout Algorithms} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{2}{0}{6}{GridLayoutModule}
\end{hierarchy}
}*/

class GridLayoutModule : public LayoutModule {

/*{\Mdefinition
The class |\Mname| is the base class for grid layout algorithms. It
defines an extended interface for computing a grid layout.

\headerline{Implementation of Grid Layout Algorithms}
The AGD provides mechanisms to work on a copy of the input graph,
rather than the input graph itself. An implementation can
manipulate the copy in certain ways and compute a grid layout for
the resulting graph. Afterwards, this grid layout is transformed
automatically into a respective grid layout for the input graph.
The type of the graph copy depends on whether the grid layout
algorithm is a planar grid layout algorithms (general postcondition
|planar_layout| is set, or not. The type is |PlaneGraphCopy| for
planar grid layout algorithms, and |GraphCopy| otherwise.
The following graph manipulations are possible if |GraphCopy| is
used:
\begin{itemize}
\item Adding new nodes or new edges. This nodes or edges will not
appear in the final grid layout.
\item Reversing edges. The respective polylines will be reversed
automatically in the final grid layout.
\item Splitting edges. The respective polylines will be concatenated
in the final grid layout.
\end{itemize}
If you want to reverse and split an edge, you must first reverse it.
Edges resulting from splitting may not be reversed. 

An implementation must override the one of the {\sl\_call\_grid} methods
depending on the parameter |use_copy| passed to the constructor of
|\Mname|, and whether the general postcondition |planar_layout| is set
or not. The following table shows which {\sl\_call\_grid} method has
to be overridden:

\begin{tabular}{cc|l}
|use_copy|  & |planar_layout| & overridable \\\hline
|false| & no  & {\sl \_call\_grid(const graph \&G, GridLayout \&gl)} \\
|true|  & no  & {\sl \_call\_grid(GraphCopy \&GC, GridLayout \&gl)} \\
|false| & yes & {\sl \_call\_grid(const graph \&G, face f, GridLayout \&gl)} \\
|true|  & yes & {\sl \_call\_grid(PlaneGraphCopy \&GC, face f, GridLayout \&gl)}
\end{tabular}

Initially, all x- and y-coordinates for nodes are set to $0$, and all
bend point lists for edges are empty. Finally, all unnecessary bend
points are removed automatically. |\Mname| also maintains the bounding
box of the least computed grid layout. If an algorithm computes this
bounding box itself, it can set it with {\sl set\_grid\_bbox}. Otherwise,
the bounding box is computed by |\Mname|. 
}*/

public:

	/*{\Minitialization}*/

	GridLayoutModule (bool use_copy = false) : LayoutModule(use_copy)
	{
			gcm = new MaxAllMapper (16, 8, 8);
	}
	/*{\Mcreate initializes a grid layout algorithm.}*/

	~GridLayoutModule() {
			delete gcm;
	}

	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext       
	\headerline{a) Extended Standard Interface} }*/

	virtual void call_grid (const leda_graph& G, GridLayout &gl);
	/*{\Mop calls |\Mvar| for graph $G$. Returns the grid layout in |gl|.}*/

	IRect get_grid_area () const {
		return _grid_bbox;
	}
	/*{\Mop returns the bounding box of the computed grid layout.
	\precond {\sl call\_grid} has been called before.}*/

	void set_coordinate_mapper (const GridCoordinateMapper& gcm);
	/*{\Mop sets the grid coordinate mapper for |\Mvar| to a copy of |gcm|.}*/

	const GridCoordinateMapper& get_coordinate_mapper() const {
		return *gcm;
	}
	/*{\Mop returns the currently used grid coordinate mapper.}*/

	void call (const leda_graph& G, LayoutInterface& A);
	void call (const leda_graph& G, leda_face F, LayoutInterface& A);
	void call (PlaneGraphCopy& PG, leda_face F, LayoutInterface& A);

	/*{\Mtext       
	The following methods are only supported by grid layout modules
	with general postcondition |planar_layout|,
	i.e., $|planar_layout|\in|post|(0)(\emptyset)$. }*/

	void call_grid (const leda_graph& G, leda_face f, GridLayout &gl);
	/*{\Mop calls |\Mvar| for graph $G$ and external face $f$. Returns
	the grid layout in |gl|. \precond $G$ represents a planar map, and
	$f$ is |nil| or a face in $G$.}*/

	void call_grid (PlaneGraphCopy& PG, leda_face f, GridLayout &gl);
	/*{\Mop calls |\Mvar| for planarized graph $PG$ and external face $f$.
	Returns the grid layout |gl|. \precond $f$ is |nil| or a face in $PG$.}*/


	/*{\Mtext       
	\bigskip
	The following methods are only supported by grid layout modules
	with general postcondition |animation|,
	i.e., $|animation|\in|post|(0)(\emptyset)$. }*/

	virtual bool start_grid_animation (const leda_graph& G, GridLayout &gl);
	/*{\Mop initializes step by step animation on the grid for graph |G|;
	the grid layout is stored in |gl|. Must be called before {\sl next\_grid\_step}
	is used. Returns |false|, if animation is not possible.}*/

	virtual bool next_grid_step ();
	/*{\Mop executes the next step of the grid animation. Returns |false|,
	if no more steps are possible, i.e., animation is finished.}*/

	virtual void stop_grid_animation ();
	/*{\Mop finish explicitly grid animation.}*/


protected:

	/*{\Mtext       
	\bigskip
	\headerline{c) Implementation of the Module Call}
	}*/

	bool _start_animation (const leda_graph& G, LayoutInterface& A);
	bool _start_animation (GraphCopy& G, LayoutInterface& A);

	bool _next_step (const leda_graph& G, LayoutInterface& A);
	bool _next_step (GraphCopy& G, LayoutInterface& A);

	void _stop_animation (const leda_graph& G, LayoutInterface& A);
	void _stop_animation (GraphCopy& G, LayoutInterface& A);

	virtual void _call_grid (const leda_graph& G, GridLayout &gl);
	/*{\Mop implements the computation of the grid layout for graph $G$.
	Returns the grid layout in |gl|.}*/

	virtual void _call_grid (GraphCopy& GC, GridLayout &gl);
	/*{\Mop implements the computation of the grid layout for graph
	copy |GC|. Returns the grid layout in |gl|.}*/

	virtual void _call_grid (const leda_graph& G, leda_face f, GridLayout &gl);
	/*{\Mop implements the computation of the grid layout for graph $G$ and
	external face $f$. $G$ represents a planar map, and $f$ is either a
	face in $G$ or |nil| indicating that the algorithm selects the external
	face. Returns the grid layout in |gl|.
	}*/

	virtual void _call_grid (PlaneGraphCopy& PG, leda_face f, GridLayout &gl);
	/*{\Mop implements the computation of the grid layout for planarized
	graph |PG| and external face $f$. $f$  is either a face in |PG| or |nil|
	indicating that the algorithm selects the external face. Returns the grid
	layout in |gl|.}*/

	
	/*{\Mtext       
	\bigskip
	\headerline{d) Implementation of Animation}
	The following methods are called if const graph references are used
	(|use_copy| is |false|) and |\Mvar| is not a planar layout module:
	}*/

	virtual bool _start_grid_animation (const leda_graph& G, GridLayout &gl);
	/*{\Mop called to initialize step by step grid animation for graph $G$
	and grid layout |gl|. Returns |true| if animation is possible,
	|false| otherwise.}*/

	virtual bool _next_grid_step (const leda_graph& G, GridLayout &gl);
	/*{\Mop called to perform the next step of the grid animation for graph $G$
	and grid layout |gl|. Returns |false| if this was the last possible step,
	|true| otherwise.}*/

	virtual void _stop_grid_animation (const leda_graph& G, GridLayout &gl);
	/*{\Mop called after the grid animation has been finished (by request of
	the client or the algorithm).}*/

	
	/*{\Mtext\medskip
	The following methods are called if graph copies are used
	(|use_copy| is |true|).
	}*/

	virtual bool _start_grid_animation (GraphCopy& GC, GridLayout &gl);
	/*{\Mop called to initialize step by step grid animation for graph copy |GC|
	and grid layout |gl|. Returns |true| if animation is possible,
	|false| otherwise.}*/

	virtual bool _next_grid_step (GraphCopy& GC, GridLayout &gl);
	/*{\Mop called to perform the next step of the grid animation for graph
	copy |GC| and grid layout |gl|. Returns |false| if this was the last
	possible step, |true| otherwise.}*/

	virtual void _stop_grid_animation (GraphCopy& G, GridLayout &gl);
	/*{\Mop called after the grid animation has been finished (by request of
	the client or the algorithm).}*/


protected:

	/*{\Mtext
	\medskip
  	\headerline{Protected Functions}
	}*/

	void set_grid_bbox(const IRect &bbox);
	/*{\Mop sets the computed grid size.}*/

	void transform_grid_layout (const GraphCopy& GC,
		GridLayout &gl, GridLayout &_gl);

private:
	void _cg_helper (const leda_graph& G, GridLayout *gl,
		LayoutInterface *ap);
	void _cg_helper (const leda_graph& G, leda_face F, GridLayout *gl,
		LayoutInterface *ap);
	void _cg_helper (PlaneGraphCopy& PG, leda_face F,
		GridLayout &gl, LayoutInterface *ap);

	void _init_grid(const leda_graph &G, GridLayout &gl);
	void _cleanup_grid(const GridLayout &gl);

	GridCoordinateMapper *gcm;
	IRect _grid_bbox;

	GridLayout the_gl;
	bool _bbox_set;

	const leda_graph *grid_step_graph;
	GraphCopy *grid_step_copy_graph;
	GridLayout *step_gl;
	GridLayout step_gl_2;
};



#endif
