//-------------------------------------------------------------------
//
//                       HierarchicalRanking.h
//
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#ifndef _AGD_HIERARCHICAL_RANKING_H
#define _AGD_HIERARCHICAL_RANKING_H


#include <AGD/RankAssignment.h>


/*{\Manpage {HierarchicalRanking} {} {Hierarchical Ranking} {R} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5.5}{AGDModule}
\arhook{5.5}{3.5}{6.5}
\hdclass{1.5}{1.5}{5.5}{RankAssignment}
\arhook{7}{2}{8}
\hdclass{2.5}{0}{6}{HierarchicalRanking}
\end{hierarchy}
}*/

class HierarchicalRanking : public RankAssignment {

/*{\Mdefinition
The |\Mname| module computes the following ranking of a graph $G=(V,E)$:
First, an acyclic graph $G=(V,E')$ is computed such that $G$ and
$G'$ have the same underlying undirected graph. If $G$ is acyclic,
then $G'$ equals $G$. Then, we compute a ranking satisfying
\begin{itemize}
\item If $v,w$ are two nodes in $G'$ with $v$ is a source, $w$ is a sink,
and there is a path from $v$ to $w$ in $G'$, then exists also a
path $v=u_1,\ldots,u_p=w$ with $rank(u_{i+1})=1+rank(u_i)$ for
$1\leq i<p$.
\item $\min\{rank(v)\mid v\in V\} = 0$
\end{itemize}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf separate\_deg0\_layer} = |true|\\
a boolean value determining if the nodes with degree $0$ are placed
on a separate layer, i.e., they get the highest rank.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

	HierarchicalRanking(bool sdl = true);
	/*{\Mcreate  creates an instance |\Mvar| of type |\Mname|. Sets
	the option |separate_deg0_layer| to |sdl|.}*/

	~HierarchicalRanking() { }


	/*{\Moperations}*/

	void call (const leda_graph& G, leda_node_array<int>& rank);
	/*{\Mop calls |\Mvar| for graph $G$. Returns the rank for each
	node $v\in G$ in |rank[v]|.}*/

	bool separate_deg0_layer() {
		return _sep_deg0;
	}

	void separate_deg0_layer (bool sdl) {
		_sep_deg0 = sdl;
	}

	AGDModule *clone() const;

private:
	bool _sep_deg0;

	void dfs(leda_node v);
	void get_tmp_rank(leda_node v, leda_node_array<int>& rank);
	void dfs_add(leda_node v, leda_node_array<int>& rank);

	int offset, MAX_N;

	leda_node_array<bool> is_source, finished;
	leda_node_array<leda_list<leda_node> > adjacent;
	leda_node_array<int> ingoing;
};

/*{\Mimplementation
The computation of the ranking takes time $O(\Labs{V}+\Labs{E})$.
}*/


#endif
