//-------------------------------------------------------------------
//
//                         KeySet.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------



#ifndef _AGD_KEY_SET_H
#define _AGD_KEY_SET_H


#include <AGD/AGD.h>
#include <LEDA/int_set.h>


/*{\Manpage {KeySet} {} {Key Sets} {S} }*/

const int _KeySet_min_table_size = 8*sizeof(unsigned long);

class KeySet {

/*{\Mdefinition
An instance |\Mvar| of the data type |\Mname| is a collection
of elements of the data type |agd_key| with an associated key mapper.
}*/

public:

	typedef agd_key item;

	/*{\Mcreation}*/

	KeySet();
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to the empty set. Sets the associated key mapper
	to |AGD|'s default key mapper. }*/

	KeySet (Keymapper& mapper);
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to the empty set. Sets the associated key mapper
	to |mapper|. }*/

	KeySet (const KeySet& x);
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to a copy of |x|. }*/

	~KeySet();

	/*{\Moperations}*/

	/*{\Mtext
	\headerline{a) Common Set Operations}
	}*/

	void insert (agd_key k);
	/*{\Mop adds |k| to |\Mvar|.}*/

	void insert (const leda_string& s) {
		insert(_mapper->get_key(s));
	}
	/*{\Mop adds $\<s\>$ to |\Mvar|.}*/

	void del (agd_key k);
	/*{\Mop deletes |k| from |\Mvar|.}*/

	void del (const leda_string& s) {
		del(_mapper->get_key(s));
	}
	/*{\Mop deletes $\<s\>$ from |\Mvar|.}*/

	bool member (agd_key k) const;
	/*{\Mop returns |true| iff $k\in \Mvar$.}*/

	KeySet join (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp += T;
	}
	/*{\Mop returns $\Mvar\cup T$.}*/

	KeySet diff (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp -= T;
	}
	/*{\Mop returns $\Mvar\setminus T$.}*/

	KeySet intersect (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp &= T;
	}
	/*{\Mop returns $\Mvar\cap T$.}*/

	KeySet symdiff (const KeySet& T) const {
		KeySet tmp (*this);
		return tmp %= T;
	}
	/*{\Mop returns $(\Mvar\cup T) \setminus (\Mvar\cap T)$.}*/

	bool empty() const;
	/*{\Mop returns |true| iff $\Mvar = \emptyset$.}*/

	void clear() {
		_table->clear();
	}
	/*{\Mop makes |\Mvar| the empty set.}*/


	/*{\Mtext
	\bigskip
	\headerline{b) Operators}
	The following operators return a new key set instead of changing
	|\Mvar|.\nopagebreak
	}*/

	KeySet operator+ (const KeySet& T) const {
		return join(T);
	}
	/*{\Mbinop returns |\Mvar|.join(|T|).}*/

	KeySet operator- (const KeySet& T) const {
		return diff(T);
	}
	/*{\Mbinop returns |\Mvar|.diff(|T|).}*/

	KeySet operator& (const KeySet& T) const {
		return intersect(T);
	}
	/*{\Mbinop returns |\Mvar|.intersect(|T|).}*/

	KeySet operator% (const KeySet& T) const {
		return symdiff(T);
	}
	/*{\Mbinop returns |\Mvar|.symdiff(|T|).}*/

	/*{\Mtext
	\medskip
	The following operators manipulate |\Mvar| directly.\nopagebreak
	}*/

	KeySet& operator+= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.join(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator-= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.diff(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator&= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.intersect(|T|) to |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator%= (const KeySet& T);
	/*{\Mbinop assigns |\Mvar|.symdiff(|T|) to |\Mvar| and returns |\Mvar|.}*/


	KeySet& operator<< (agd_key k) {
		insert(k);
		return *this;
	}
	/*{\Mbinop inserts |k| into |\Mvar| and returns |\Mvar|.}*/

	KeySet& operator<< (const leda_string& str) {
		insert(str);
		return *this;
	}

	/*{\Mtext
	\medskip
	Relational and assignment operators:\nopagebreak
	}*/

	bool operator<= (const KeySet& T) const;
	/*{\Mbinop returns |true| iff $\Mvar\subseteq T$.}*/

	bool operator>= (const KeySet& T) const {
		return T <= (*this);
	}
	/*{\Mbinop returns |true| iff $\Mvar\supseteq T$.}*/

	KeySet& operator= (const KeySet& T);
	/*{\Mbinop assigns |T| to |\Mvar|. Returns |\Mvar|.}*/


	/*{\Mtext
	\bigskip
	\headerline{c) Input and Output}
	}*/

	void read (istream& is, int delim = (char)EOF);
	/*{\Mop reads a sequence of strings $s_1,\ldots,s_k$ terminated by
	|delim| from the input stream |is|, and assigns the set $\left\{
	\<s_1\>,\ldots,\<s_k\>\right\}$ to |\Mvar|.}*/

	void read (char delim = '\n') {
		read (cin, delim);
	}
	/*{\Mop calls |\Mvar|.read(|cin,delim|) to read |\Mvar| from the
	standard input stream |cin|.}*/

	void print (ostream& os, char space = ' ') const;
	/*{\Mop prints the string representations of the elements of |\Mvar|
	to the output stream |os|. The elements are separated by |space|.}*/

	void print (char space = ' ') const {
		print (cout, space);
	}
	/*{\Mop calls |\Mvar|.print(|cout,space|) to print |\Mvar| to the
	standard output stream |cout|.}*/


	// Iteration

	item first_item() const {
		return find_next_item(_mapper->first_key());
	}

	item next_item (item it) const {
		return find_next_item(_mapper->next_key(it)); 
	}

	const agd_key inf (item it) const {
		return it;
	}


private:
	void _alloc_table (int sz);
	void _expand_table (int sz);

	item find_next_item (item it) const;

	leda_int_set *_table;
	int _table_size;
	Keymapper *_mapper;
};


inline istream& operator>> (istream& is, KeySet& S)
{
	S.read (is);
	return is;
}
/*{\Mbinopfunc reads the key set |S| from the input stream |is|.}*/

inline ostream& operator<< (ostream& os, const KeySet& S)
{
	S.print (os);
	return os;
}
/*{\Mbinopfunc writes the key set |S| to the output stream |os|.}*/


/*{\Mtext
\bigskip
\headerline{Iteration}
{\bf forall}($k,S$) 
$\{$  ``the elements of $S$ are successively assigned to $k$''  $\}$
}*/


/*{\Mimplementation
Key sets are implemented by bit vectors.
}*/


inline int compare (const KeySet&, const KeySet&)
{ return 0; }


#endif
