//-------------------------------------------------------------------
//
//                      LayoutInterface.h
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#ifndef _AGD_LAYOUT_INTERFACE_H
#define _AGD_LAYOUT_INTERFACE_H


#include <AGD/DPolyline.h>
#include <AGD/DRect.h>
#include <LEDA/graph.h>


/*{\Manpage {LayoutInterface} {} {Layout Interfaces} {A} }*/

class LayoutInterface {

/*{\Mdefinition
	The abstract base class |\Mname| defines the generic interface to
	access the graphical
	attributes of a graph. The LEDA |graph| class contains only the topology
	of a graph, e.g., the nodes and edges. If you want to produce a graphical
	representation of a graph, you need additional attributes such as the graphics
	and the position of each node, the line represantation of each edge,
	and some more.

	Our graphical model for graphs contains the following attributes:
	\begin{itemize}
	\item a rectangular bounding box for the graphics of each node.
	\item a position of type |DPoint| for each node |v| which is the center
	of the graphics of |v|.
	\item a source and a target anchor point of type |DPoint| for each edge |e|
	with coordinates in the range of $[-1\ldots1]\times[-1\ldots1]$. The
	anchor points define the start and endpoint of the line representation
	of |e|.
	\item a list of bend points of type |DPolyline| for each edge.
	\end{itemize}

	\headerline{The Coordinate System}
	We assume a rectilinear coordinate system, where positive
	$x$ is to the right, positive $y$ is down, and the origin is in the upper left
	corner. One unit in $x$- and $y$-direction should be 1/72 inch. We call this
	coordinate system the {\em standard coordinate system}.

}*/

public:

	/*{\Minitialization}*/

	/*{\Mtext       
	|\Mname| is an abstract base class, so you cannot create instances of it.
	}*/

	LayoutInterface(const leda_graph &G) {
		gp = &G;
	}
	/*{\Mcreate initializes a layout interface for graph $G$.}*/

	virtual ~LayoutInterface() { }
	
	/*{\Moperations}*/

	/*{\Mtext
	\headerline{a) Access to Graphical Attributes} 
	The following methods define the generic interface to the graphical
	attributes of a graph. All of them are pure virtual functions and must be
	overridden by a specif implementation of |\Mname|, except for
	|get_position| and |set_position|, which have a default implementation.
	}*/

	virtual double get_x (leda_node v) const
	/*{\Moppure returns the x-coordinate of the position of node |v|.}*/
	= 0;

	virtual void   set_x (leda_node v, double new_x)
	/*{\Moppure sets the x-coordinate of the position of node |v| to |new_x|.}*/
	= 0;

	virtual double get_y (leda_node v) const
	/*{\Moppure returns the y-coordinate of the position of node |v|.}*/
	= 0;

	virtual void   set_y (leda_node v, double new_y)
	/*{\Moppure sets the y-coordinate of the position of node |v| to |new_y|.}*/
	= 0;


	virtual DPoint get_position (leda_node v) const {
		return DPoint (get_x(v), get_y(v));
	}
	/*{\Mop returns the position of node |v|.}*/

	virtual void set_position (leda_node v, const DPoint& pos) {
		set_x (v, pos.xcoord());
		set_y (v, pos.ycoord());
	}
	/*{\Mop sets the position of node |v| to |pos|.}*/
	
	virtual DPolyline get_bend_points (leda_edge e) const
	/*{\Moppure returns the bend points of edge |e|.}*/
	= 0;

	virtual void set_bend_points (leda_edge e, const DPolyline& l)
	/*{\Moppure sets the bend points of edge |e| to |l|.}*/
	= 0;

	virtual double get_width  (leda_node v) const
	/*{\Moppure returns the width of the graphics of node |v|.}*/
	= 0;

	virtual void set_width  (leda_node v, double new_w)
	/*{\Moppure sets the width of the graphics of node |v| to |new_w|.}*/
	= 0;

	virtual double get_height (leda_node v) const
	/*{\Moppure returns the height of the graphics of node |v|.}*/
	= 0;

	virtual void set_height (leda_node v, double new_h)
	/*{\Moppure sets the height of the graphics of node |v| to |new_h|.}*/
	= 0;

	virtual DPoint get_source_anchor (leda_edge e) const
	/*{\Moppure returns the source anchor position of edge |e|.}*/
	= 0;

	virtual void set_source_anchor (leda_edge e, const DPoint& pos)
	/*{\Moppure sets the source anchor position of edge |e| to |pos|.}*/
	= 0;

	virtual DPoint get_target_anchor (leda_edge e) const
	/*{\Moppure returns the target anchor position of edge |e|.}*/
	= 0;

	virtual void set_target_anchor (leda_edge e, const DPoint& pos)
	/*{\Moppure sets the target anchor position of edge |e| to |pos|.}*/
	= 0;

	/*{\Mtext
	\bigskip
	\headerline{b) Notification Methods}
	The following methods can be overridden by derived classes if
	necessary. The default implementations do nothing.
	}*/

	virtual void update() { }
	/*{\Mop called to update the layout, e.g., update the window displaying
	the graph, or set temporarily buffered informations.}*/

	virtual void refresh() { }
	/*{\Mop called to indicate that the associated graph has been
	changed, e.g., by adding or deleting nodes or edges.}*/


	/*{\Mtext
	\bigskip
	\headerline{c) Further Operations}
	}*/

	const leda_graph &get_graph() const {
		return *gp;
	}
	/*{\Mop returns the associated graph.}*/

	virtual LayoutInterface& original() {
		return *this;
	}

	DRect area() const;
	/*{\Mop computes a tight bounding box.}*/

	virtual void make_straight_line ();
	/*{\Mop deletes all edge bends and sets all edge anchors to
	$(0,0)$.}*/

	virtual void set_all_anchors (DPoint p = DPoint(0,0));
	/*{\Mop sets all source and target anchors to |p|.}*/


protected:

	void set_graph(const leda_graph& G) {
		gp = &G;
	}
	/*{\Mop protected method for setting the associated graph.}*/

private:
	const leda_graph *gp;
};


/*{\Mexample
The following function translates all nodes displayed in
a GraphWin $gw$ by a vector $p$, removes all bends points,
and returns the bounding box of the layout.

\small\tt
\begin{verbatim}
DRect translate(GraphWin &gw, DPoint p)
{
  GraphWinInterface gwi(gw);        // create layout interface
  const graph &g = gwi.get_graph();

  gwi.make_straight_line();  // remove alle bends

  // shift all nodes by p
  node v;
  forall_nodes(v,g)
    gwi.set_position(v, p + gwi.get_position());

  gwi.update();  // update the display

  return gwi.area();  // compute bounding box
}
\end{verbatim}
}*/


#endif
