//-------------------------------------------------------------------
//
//                        MaxAllMapper.h
//
//    author: Carsten Gutwenger
//    date:   octobre 1997
//-------------------------------------------------------------------


#ifndef _AGD_MAX_ALL_MAPPER_H
#define _AGD_MAX_ALL_MAPPER_H


#include <AGD/GridCoordinateMapper.h>


/*{\Manpage {MaxAllMapper} {} {The Grid Coordinate Mapper MaxAll} {M} }*/

/*{\Mbaseclasses

\begin{hierarchy}{7.5}{2.5}
\hdclass{0}{1.5}{7.5}{GridCoordinateMapper}
\arhook{7.5}{2}{8.5}
\hdclass{3}{0}{6}{MaxAllMapper}
\end{hierarchy}
}*/

class MaxAllMapper : public GridCoordinateMapper {

/*{\Mdefinition
	The MaxAll mapper maps the grid coordinates such that each row
	and each column has the same size |size|. This is given by
	\[ |size| = \max \{ |width|(v),|height|(v) \mid v\in V\} \]
	where $G=(V,E)$ is the given graph, and |width(v)|, |height(v)| are the
	width and the height of the node graphics, respectively.

	Rows and columns are separated by an additional distance |separation|.
	Furthermore, a left margin |left_margin| and a right margin
	|right_margin| can be set. The mapping is then defined as
	\[ \varphi_x(|column|) = (|size| + |separation|) * (|column|-|xmin|) +
	|left_margin| + \frac{|size|}{2} \]
	\[ \varphi_y(|row|) = (|size| + |separation|) * (|ymax|-|row|) +
	|upper_margin| + \frac{|size|}{2} \]

	The size of the node graphics is not changed, that means
	\[  \left. \begin{array}{c@{\;=\;}c}
		|width|'(v) & |width|(v) \\
		|height|'(v) & |height|(v)
		\end{array} \right\} \mbox{for all $v\in V$} \]
}*/

public:

	/*{\Mcreation}*/

	MaxAllMapper (double sep, double lm, double um) {
		separation (sep);
		left_margin (lm);
		upper_margin (um);
	}
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|, and sets
	|separation| to |sep|, |left_margin| to |lm|, and |upper_margin| to
	|um|.}*/

	virtual ~MaxAllMapper() { }


	/*{\Moperations}*/

	/*{\Mtext       
	\headerline{a) Initialization} }*/

	void init ();
	/*{\Mop computes the mapping for the grid layout given by the graph |G|, 
	the node coordinates |xpos|, |ypos|, the bendpoints |bends| of the edge
	polylines, the layout interface |a|, and the bounding box of the grid
	layout |grid_bbox|.}*/


	/*{\Mtext       
	\bigskip
	\headerline{b) Mapping of coordinates} }*/

	double map_x (int column) const {
		return ax * column + bx;
	}
	/*{\Mop returns $\varphi_x(|column|)$.}*/

	double map_y (int row) const {
		return ay * row + by;
	}
	/*{\Mop returns $\varphi_y(|row|)$.}*/

	void get_layout_size (DRect& bbox) {
		bbox = the_bbox;
	}
	/*{\Mop returns the bounding box of the mapped layout.}*/

	/*{\Mtext       
	\bigskip
	\headerline{c) Further operations} }*/

	GridCoordinateMapper *clone() const {
		return new MaxAllMapper (separation(), left_margin(), upper_margin());
	}
	/*{\Mop returns a pointer to a new |\Mname| object that behaves like
	|\Mvar|, i.e. |clone| only copies options, not computed informations.}*/

	double separation() const {
		return the_sep;
	}
	/*{\Mop returns the current setting of |separation|.}*/

	void separation (double sep) {
		the_sep = sep;
	}
	/*{\Mop sets |separation| to |sep|.}*/

	double left_margin() const {
		return the_lm;
	}
	/*{\Mop returns the current setting of |left_margin|.}*/

	void left_margin (double lm) {
		the_lm = lm;
	}
	/*{\Mop sets |left_margin| to |lm|.}*/

	double upper_margin() const {
		return the_um;
	}
	/*{\Mop returns the current setting of |upper_margin|.}*/

	void upper_margin (double um) {
		the_um = um;
	}
	/*{\Mop sets |upper_margin| to |um|.}*/

protected:
	void get_max_width_height (double& max_w, double& max_h);
	void set_coeff (double size);

	double the_sep, the_lm, the_um;

	double ax, bx, ay, by;
	DRect the_bbox;
};

/*{\Mimplementation
The initialization of the
mapping (|init|) takes time $O(n)$ where $n$ is the number of nodes of the
given graph. All other operations take constant time.
}*/


#endif
