
#ifndef _AGD_ORTHO_LAYOUT_H
#define _AGD_ORTHO_LAYOUT_H

#include <AGD/GridLayoutModule.h>
#include <AGD/Ovid.h>

void random_4_planar_graph(leda_graph& G, int n, int m, bool biconnected);

/*{\Manpage {Ortho} {} {The Orthogonal Drawing Algorithm} {L} }*/

/*{\Mbaseclasses
\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3.5}{0}{6}{Ortho}
\end{hierarchy}
\hfill\agdeps{Ortho.eps}{5.5cm}
}*/

class Ortho : public GridLayoutModule, public Ovid {

/*{\Mdefinition
The class |\Mname| represents Tamassia's algorithm~\cite{t-eggmn-87} for
constructing orthogonal grid embeddings of 4--planar graphs. This class of
graphs contains planar graphs $G$ whose maximal vertex degrees $\Delta(G)$ do
not exceed 4.  An input graph |G| will be drawn on a grid such that the edges
run along grid lines, vertices are placed on different grid points, and no
crossings occur. The number of bends will be minimal with respect to the given
embedding. $G$ must be a planar graph without multiple edges and self loops. 

\horizline

\commoninfo{Ortho}{Ortho (Tamassia)}{Roberto Tamassia}{Gunnar W.
Klau}{June 1998}{1.4}

\headerline{Pre- and Postcondition}

\begin{tabular}{lcl}
precondition & $=$ & $\{\,$ {\em planar, simple, no self loops, 4--graph, connected} $\,\}$ \\
postcondition(|PRE|) & $=$ & $\{\,$ {\em no crossings, orthogonal} $\,\}$
\end{tabular}\\

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf dissect} $= true$\\ 
Truth value indicating if the faces should be dissected in rectangles during the compaction phase. This is the original method proposed in \cite{t-eggmn-87}. Switching this off may lead to violated embeddings in the current version.
\item |bool| {\bf check} $= true$\\
If set, the checking routines are activated. Especially the 
orthogonal representation is verified, which is helpful for debugging and
certificates the correctness of the computation.  
\end{options} \horizline }*/

public:
  /*{\Mcreation}*/
    
  Ortho();
  /*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/
  
  virtual ~Ortho() {}
  
  /*{\Moperations 1.5 2.5}*/
  
  leda_string name() const { return leda_string("Orthogonal (Tamassia)"); }
  
  leda_string long_name() const { 
    return leda_string("Bend minimal orthogonal grid drawings (fixed topology)"); 
  }
  
  leda_string author() const { return leda_string("Roberto Tamassia"); }
  
  virtual leda_string impl_author() const { return leda_string("Gunnar W. Klau"); }
  
  virtual leda_string impl_version() const { return leda_string("1.4"); }
  
  virtual leda_string impl_date() const { return leda_string("June 1998"); }

  AGDModule* clone() const;

  void output_level(int o) { _output_level = o; }

  int output_level() const { return _output_level; }

  bool dissect() const { return _dissect; }
  /*{\Mop returns the optional parameter |dissect| }*/ 

  void dissect(bool d) { _dissect = d; }
  /*{\Mop sets the parameter |dissect| to |d| }*/ 
  
  bool optimally() const { return _optimally; }
  /*{\Mop returns the optional parameter |optimally| }*/ 

  void optimally(bool o) { _optimally = o; }
  /*{\Mop sets the parameter |optimally| to |d| }*/ 

  bool verify() const { return _check; }
  /*{\Mop returns the verify status }*/ 

  void verify(bool c) { _check = c; }
  /*{\Mop sets the verify status to |c| }*/ 

  /*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (PlaneGraphCopy& PG, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (PlaneGraphCopy& PG, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect get_grid_area ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_coordinate_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/


protected:

  bool _dissect, _optimally;
  int _output_level;
  
  void _call_grid(PlaneGraphCopy& G, leda_face f_0, GridLayout& I);

};

/*{\Mimplementation
The computation of the layout takes time $O(n^2 \log n \log n)$ where $n$ is the number of nodes of the graph to be drawn. This is due to the flow algorithm, we chose {\tt LEDA}'s [[MIN_COST_FLOW]] because of its simplicity and stability. A different choice would influence the running time.
}*/

/*{\Mexample\label{ortho_example}
The following program shows how to use the orthogonal layout algorithm in
combination with a [[TopologyModule]]. The user can draw or load a graph in a
[[GraphWin]], pressing the ``done'' button initiates the layout process. If
the input drawing divides the plane into faces, the same faces will emerge in
the output drawing (i.e. the topological representation and the external face
will be preserved during the computations).

{\small \tt
\begin{verbatim}
#include <AGD/TopologyModule.h>
#include <AGD/Ortho.h>
#include <AGD/GraphWinInterface.h>
#include <LEDA/graphwin.h>

main() {
  
  GraphWin W; W.display();

  while (W.edit()) {
  
    GraphWinInterface I(W);
    graph& G = W.get_graph(); face f_0; string msg;
    TopologyModule T; Ortho O;
    T.call(G, f_0, I);
    if (O.check(G, msg)) O.call(G, f_0, I);
    else W.acknowledge(msg);
    W.fill_window();

  }

}
\end{verbatim}
}
}*/

#endif

