//-------------------------------------------------------------------
//
//                    PlanarStraightLayout.h
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_BIC_STRAIGHT_LAYOUT_H
#define _AGD_BIC_STRAIGHT_LAYOUT_H


#include <AGD/GridLayoutModule.h>
#include <AGD/AugmentModule.h>
#include <AGD/AGDModuleOption.h>


class LmcOrder;


/*{\Manpage {PlanarStraightLayout} {} {The PlanarStraight Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{9.5}{5.5}
\hdclass{0}{4.5}{5}{AGDModule}
\arhook{5}{5}{6}
\hdclass{1.5}{3}{5}{LayoutModule}
\arhook{6.5}{3.5}{7.5}
\hdclass{2}{1.5}{6}{GridLayoutModule}
\arhook{8}{2}{9}
\hdclass{3}{0}{6.5}{PlanarStraightLayout}
\end{hierarchy}
\hfill\agdeps{planar_straight.epsi}{3.2cm}
}*/

class PlanarStraightLayout : public GridLayoutModule {

/*{\Mdefinition
The class |\Mname| represents the layout algorithm
PlanarStraight (Gutwenger and Mutzel~\cite{GM97}), which
is a generalization of the algorithm Convex (Kant~\cite{Kan96}).

This algorithm draws a planar graph |G| straight-line
without crossings. |G| may not contain self-loops or mulitple edges.
The grid layout
size is at most $(2n-4) \times (n-2)$ for a graph with $n$ nodes ($n \geq 3$).

\horizline
\commoninfosz{PlanarStraight}{PlanarStraight (Gutwenger/Mutzel)}{Carsten Gutwenger, Petra Mutzel}
{Carsten Gutwenger}{march 1997}{2.0}{6cm}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\{\, |planar|,|simple|,|no self loops| \,\}$ \\
postcondition(|PRE|) & $=$ & $\{\, \mbox{|straight|-|line|},|no crossings|,|convex| \,\}$
\end{tabular}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf augmenter}\\
an |AugmentModule| option used for planar biconnected augmentation.

\begin{tabular}{ll}
guaranteed precondition: & |planar|, |simple|, |no self loops|, |directed|\\
required postcondition: & |planar|, |biconnected|, |simple|, |no self loops|\\
initial module: & |PlanAug|
\end{tabular}
\end{options}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |bool| {\bf size\_opt} = |true|\\
a boolean value determining if the grid size of the
layout is optimized compared to the original Convex algorithm. This
option reduces the grid size especially for graphs with few edges.
If the graph is (almost) triangulated, the grid size will be (almost)
the same.
\item |bool| {\bf max\_outer} = |true|\\
a boolean value determining if the algorithm chooses
the face of |G| with maximal degree as outerface (|true|), or the
first face in the list of faces.
\item |double| {\bf base\_ratio} $= 0.33$\\
a double value determining, how many nodes are drawn on the base line.
If |base_ratio| is $r$ and the outerface is $F$, the number of nodes
on the base line is $\min\left(2,\max\left(\ell, r\Labs{F}\right)\right)$,
where $\ell = \max\{k \mid v_1,\ldots,v_k$ is a path on $F_{out}$ with
$(v_i,v_j)\notin E$ for all $j\neq i+1 \}$.
\end{options}
\horizline
}*/

public:

	/*{\Mcreation L 5}*/

	PlanarStraightLayout ();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

	virtual ~PlanarStraightLayout() { }


	/*{\Moperations 1.5 2.5}*/

	leda_string name () const {
		return "PlanarStraight";
	}

	leda_string long_name () const {
		return "PlanarStraight (Gutwenger/Mutzel)";
	}

	leda_string author () const {
		return "Carsten Gutwenger, Petra Mutzel";
	}

	leda_string impl_author () const {
		return "Carsten Gutwenger";
	}

	leda_string impl_version () const {
		return "2.0";
	}

	leda_string impl_date () const {
		return "march 1997";
	}


	bool size_opt() const {
		return the_size_opt;
	}

	void size_opt (bool so) {
		the_size_opt = so;
	}

	bool max_outer() const {
		return the_max_outer;
	}

	void max_outer (bool mo) {
		the_max_outer = mo;
	}

	double base_ratio() const {
		return the_base_ratio;
	}
	
	void base_ratio (double br) {
		the_base_ratio = br;
	}

	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base classes (|LayoutModule|, |GridLayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (const graph& G, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	virtual void call (PlaneGraphCopy& PG, face f, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (const graph& G, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	void call_grid (PlaneGraphCopy& PG, face f, GridLayout &gl);
	*/
	/*{\Mop}*/

	/* inherited
	IRect get_grid_area ();
	*/
	/*{\Mop}*/

	/* inherited
	void set_coordinate_mapper (const GridCoordinateMapper& gcm);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/

	AGD_DECLARE_MODULE (AugmentModule, augmenter)

protected:
	 void _call_grid (PlaneGraphCopy& PG, leda_face F, GridLayout &gl);

private:
	bool the_size_opt;
	bool the_max_outer;
	double the_base_ratio;

	void compute_coordinates (leda_graph &G, LmcOrder &lmc,
		leda_node_map<int>& xpos, leda_node_map<int>& ypos);
};


#endif
