//-------------------------------------------------------------------
//
//                      PlanarizationLayout.h
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_PLANARIZATION_LAYOUT_H
#define _AGD_PLANARIZATION_LAYOUT_H


#include <AGD/LayoutModule.h>
#include <AGD/PlanarizerModule.h>
#include <AGD/AGDModuleOption.h>


/*{\Manpage {PlanarizationLayout} {} {The Planarization Method} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1}{1.5}{5.5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{1.5}{0}{6.5}{PlanarizationLayout}
\end{hierarchy}
\hfill
\agdeps{planarization.epsi}{5.5cm}
}*/

class PlanarizationLayout : public LayoutModule {

/*{\Mdefinition
The class |\Mname| represents a general framework for drawing with planarization.
It transforms a general graph into a planar graph using a planarization module, and
draws the resulting planar graph using a layout algorithm for planar graphs.
The planarization and planar layout algorithm are exchangeable modules allowing
to use |\Mname| with different methods for planarization and planar drawing.

\horizline
\commoninfo{Planarization}{Planarization}{}
{Carsten Gutwenger}{decembre 1997}{1.0}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\{\, |simple|,|no self loops| \,\}$ \\
postcondition(|PRE|) & $=$ & $\emptyset$
\end{tabular}

\headerline{Exchangeable Modules}
Instances of type |\Mname| provide the following module options:
\begin{options}
\item {\bf planarizer}\\
a |PlanarizerModule| option used to compute the planarization of the input graph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\, |simple|, |no self loops|, |directed| \,\}$\\
required postcondition: & $\{\, |planar|, |simple|, |no self loops| \,\}$\\
initial module: & |PlanarSubgraph|
\end{tabular}

\item {\bf planar\_layouter}\\
a |LayoutModule| option used to draw the planarized graph.

\begin{tabular}{ll}
guaranteed precondition: & $\{\, |simple|, |no self loops|, |planar| \,\}$\\
required postcondition: & $\{ |planar layout| \} , \{\, |no crossings| \,\}$\\
initial module: & |Quod|
\end{tabular}
\end{options}
\horizline
}*/

AGD_DECLARE_MODULE(PlanarizerModule,planarizer)

protected:
	AGDModuleOption<LayoutModule> planar_layouter;
public:
	void set_planar_layouter (const LayoutModule& M);

public:

	PlanarizationLayout();
	~PlanarizationLayout() { }

	/*{\Moperations}*/

	leda_string name () const {
		return "Planarization";
	}

	leda_string impl_author () const {
		return "Carsten Gutwenger";
	}

	leda_string impl_version () const {
		return "1.0";
	}

	leda_string impl_date () const {
		return "decembre 1997";
	}

	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	virtual bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/

protected:

	void _call (const leda_graph& G, LayoutInterface& A);
};


#endif
