//-------------------------------------------------------------------
//
//                         PlaneGraphCopy.h
//
//    author: Carsten Gutwenger
//    date:   june 1998
//-------------------------------------------------------------------


#ifndef _AGD_PLANARIZED_GRAPH_H
#define _AGD_PLANARIZED_GRAPH_H


#include <AGD/GraphCopy.h>

template<class E> class stack;
struct MMDelNode;


/*{\Manpage {PlaneGraphCopy} {} {Plane Graph Copies} {PG} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{graph}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{GraphCopy}
\arhook{6.5}{2}{7.5}
\hdclass{2.5}{0}{5.5}{PlaneGraphCopy}
\end{hierarchy}
}*/

class PlaneGraphCopy : public GraphCopy {

/*{\Mdefinition
Plane graph copies represent copies of planar graphs or
planarized graphs as planar maps.
}*/

public:

	/*{\Mcreation PG 6}*/

	PlaneGraphCopy () : GraphCopy(), S(0) { }
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to an empty graph wothout original graph.}*/
	
	PlaneGraphCopy(const PlaneGraphCopy& PG);

	PlaneGraphCopy(const leda_graph& G);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of |G|.}*/

	PlaneGraphCopy(const leda_graph& G, leda_list<leda_edge>& E);
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname| initialized
	to a copy of $G$ without the edges in $E$. \precond Each $e\in E$
	is an edge in $G$.}*/

	~PlaneGraphCopy();


	/*{\Moperations 2 3.5}*/

	PlaneGraphCopy& operator= (const PlaneGraphCopy& PG);

	leda_face copy_of_face (leda_face f) const {
		return face_of(_chain[original().first_face_edge(f)].head());
	}
	/*{\Mop returns a face of |\Mvar| that corresponds to $F$.
	\precond $F$ is a face in $G$.}*/

	virtual leda_edge split (leda_edge e);
	/*{\Mop split edge $e=(v,w)$ and its reversal $r=(w,v)$ into
	edges $(v,u)$, $(u,w)$ and their reversals by introducing a new
	node $u$. Returns the edge $(u,w)$.}*/

	void new_edge_chain (leda_edge e, leda_edge from, leda_edge to,
		const leda_list<leda_edge>& crossings);
	/*{\Mop Inserts a chain of edges for edge e into |PG|. Let
	$|crossings|=e_1,\ldots,e_k$ for a $k\geq0$,
	$e_{k+1}=|to|$, and $e=(v,w)$.
	Then, we split each $e_i$ $(1\leq i\leq k)$ with a node $u_i$,
	and insert the
	edges $(|source|(|from|),u_1)$,$(u_i,u_{i+1})$,$(u_k,|source|(|to|))$ into
	|\Mvar|. The inserted edges are appended to |chain(e)|.
	\precond $e$ is an edge in $G$, all other edges are edges in |\Mvar|,
	$|source|(|from|)=|copy|(v)$,
	$|source|(|to|)=|copy|(w)$,
	$|face|(|from|)=|face|(e_1)$, and $|face|(|reversal|(e_i))=|face|(e_{i+1})$.}*/

	leda_face expand (leda_node v);
	/*{\Mop replaces node $v$ by a face $f$. Let $v_1,\ldots,v_k$ be the
	adjacent nodes of $v$. Then $f = w_1,\ldots,w_k$, where deg$(w_i)=3$
	and $w_i$ is connected to $v_i$. Returns $f$. \precond deg$(v)\geq 3$.}*/

	leda_node collapse (leda_face f, leda_node v_orig);
	/*{\Mop collapses face $f$ to a single node $v$. Sets the original
	node of $v$ to |v_orig|. Serves as undo for expand$(v)$. Returns $v$.
	\precond Each node on $f$ has degree $3$.}*/

	leda_face max_face() const;
	/*{\Mop returns a face of |\Mvar| with maximal degree.}*/

	virtual bool is_map() const {
		return true;
	}
	/*{\Mop returns |true|.}*/

	void remove_deg1_nodes(const leda_node_array<bool> &mark);
	void restore_deg1_nodes(leda_list<leda_node> &deg1s);

private:

	stack<MMDelNode> *S;

	void extend_chain(leda_edge e1, leda_edge e2);
	leda_edge add_edge(leda_edge orig, leda_edge rev_orig, leda_edge e, leda_node w,
		int dir = LEDA::leda_after);
	leda_edge add_edge(leda_edge orig, leda_edge rev_orig, leda_node v, leda_node w);


	void _init (const leda_graph& G, leda_list<leda_edge>& E);
};



#endif
