//-------------------------------------------------------------------
//
//                       SpringLayout.h
//
//    author: Stefan Naeher, David Alberts
//    date:   April 1997
//-------------------------------------------------------------------


#ifndef _AGD_SPRING_LAYOUT_H
#define _AGD_SPRING_LAYOUT_H


#include <AGD/LayoutModule.h>
#include <LEDA/array2.h>


/*{\Manpage {SpringLayout} {} {Spring Embedder} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{3}{0}{5}{SpringLayout}
\end{hierarchy}
\hfill\agdeps{spring.epsi}{3cm}
}*/

class SpringLayout : public LayoutModule
{

/*{\Mdefinition
The class |\Mname| represents the spring embedder layout algorithm
by Fruchtermann and Rheingold~\cite{fr-gdfdp-91}.

This algorithm draws a general graph |G| straight-line. The drawing
of a planar graph may contain crossings.

\horizline
\commoninfo{Spring Embedder}{Spring Embedder (Fruchtermann/Rheingold)}{Fruchtermann, Rheingold}
{Stefan Naeher, David Alberts}{April 1997}{}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\{\,  \,\}$ \\
postcondition(|PRE|) & $=$ & $\{\, \mbox{|straight|-|line|} \,\}$
\end{tabular}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |int| {\bf iterations} = 400\\
the number of main steps that the algorithm performs.
\item |bool| {\bf noise} = |true|\\
determines if the algorithm performs random perturbations.
\end{options}\horizline
}*/

  public:

	/*{\Mcreation}*/

    SpringLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

    ~SpringLayout() { }

  /*{\Moperations}*/

    leda_string name () const { return "Spring Embedder"; }

    leda_string long_name () const
    { return "Spring Embedder (Fruchtermann/Rheingold)"; }

    leda_string author () const { return "Fruchtermann, Rheingold"; }

    leda_string impl_author () const
    { return "Stefan Naeher, David Alberts"; }

    leda_string impl_date () const
    { return "April 1997"; }

    void do_spring(const leda_graph& G, LayoutInterface& a)
    { leda_list<leda_node> L; do_spring(G,a,L); }

    void do_spring(const leda_graph& G, LayoutInterface& a,
                   const leda_list<leda_node>& fixed_nodes);


    int iterations() const { return the_iterations; }

    void iterations(int i)
    { if(i>0) the_iterations = i; }

    bool noise() const { return the_noise; }

    void noise(bool on)
    { the_noise = on; }
      
	AGDModule *clone () const;


	/*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	bool start_animation (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/

	/* inherited
	bool next_step ();
	*/
	/*{\Mop}*/

	/* inherited
	void stop_animation ();
	*/
	/*{\Mop}*/

	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/


  private:
    
    bool initialize(const leda_graph& G, LayoutInterface& a,
                    const leda_list<leda_node>& fixed_nodes);
    void main_step(const leda_graph& G, LayoutInterface& a);
    void cleanup() { if(A) delete A; }

    void _call(const leda_graph& G, LayoutInterface& a);
    void _call(GraphCopy& GC, LayoutInterface& a)
	{ LayoutModule::_call (GC,a); }
    
    bool _start_animation(const leda_graph& G, LayoutInterface& a)
    { leda_list<leda_node> L; initialize(G,a,L); return true; }
    
	bool _start_animation(GraphCopy& GC, LayoutInterface& A)
	{ return LayoutModule::_start_animation (GC,A); }

    bool _next_step(const leda_graph& G, LayoutInterface& a)
    { main_step(G,a); return true; }
    
	bool _next_step(GraphCopy& GC, LayoutInterface& A)
	{ return LayoutModule::_next_step (GC,A); }

    void _stop_animation(const leda_graph& /*G*/, LayoutInterface& /*a*/)
    { cleanup(); }

	void _stop_animation(GraphCopy& GC, LayoutInterface& A)
	{ LayoutModule::_stop_animation (GC,A); }

    
    leda_node_array<list_item>  lit; 
    leda_node_array<bool>       fixed;

    int c_f;
    
    double width;
    double height; 

    double tx_null;
    double ty_null;
    double tx;
    double ty;

    double k;
    double k2;
    double kk;
    int ki;
    
    int xA;
    int yA;

    leda_array2<leda_list<leda_node> > *A;


    float log2(int x)
    { float l = 0.0f; while(x) { l++; x >>= 1; } return l/2; }

    int    the_iterations;     // # of iterations

    double xleft;              // coordinates of bounding box
    double xright;
    double ysmall;
    double ybig;

    bool the_noise;            // perform random perturbations

};


#endif
