//-------------------------------------------------------------------
//
//                        SubgraphModule.h
//
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#ifndef _AGD_SUBGRAPH_MODULE_H
#define _AGD_SUBGRAPH_MODULE_H


#include <AGD/AGDModule.h>


class GraphCopy;


/*{\Manpage {SubgraphModule} {} {Subgraph Modules} {S} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1}{0}{6}{SubgraphModule}
\end{hierarchy}
}*/

class SubgraphModule : public AGDModule {

/*{\Mdefinition
The class |\Mname| is the base class for subgraph modules. A
subgraph module computes a subgraph $G'=(V,E')$ of an input
graph $G=(V,E)$, where $E'\subseteq E$. Usually, $G'$ satisfies
a certain condition, e.g, planar or acyclic. This condition is
expressed by the postcondition of the module.

\headerline{Input and Output Parameters}

\begin{tabular}{lcc}
input parameter: & graph & $|in_graph|=1$ \\
output parameter: & graph & $|out_graph|=1$
\end{tabular}

The integer constants |in_graph| and |out_graph| can be used to denote
the positions of the input and output parameters.

\headerline{Implementing Subgraph Modules}

An implementation of a subgraph module must override the protected
method call\_subgraph($G$,$L$), which gets as input a reference to a
const graph $G$. It determines the set of edges $\{e_1,\ldots,e_k\}=E\cap E'$,
and returns it as a list in $L$.
}*/

public:

	// the input and output parameters of a subgraph module
	static const int in_graph;
	static const int out_graph;

	/*{\Minitialization S 4.5}*/

	SubgraphModule () : AGDModule(1,1) { }

	/*{\Mcreate initializes a subgraph module.}*/

	virtual ~SubgraphModule () { }


	/*{\Moperations 1.5 3.5}*/

	/*{\Mtext
	\headerline{Standard Interface} }*/

	bool check (const leda_graph& G, leda_string& message);
	/*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
	Otherwise, |false| is returned, and |message| contains the
	error message.}*/

	void call (const leda_graph& G, leda_list<leda_edge>& L);
	/*{\Mop calls the subgraph module for graph $G$. Returns in $L$ the
	list of edges $e_1,\ldots,e_k$.}*/

	void call (leda_graph& G);
	/*{\Mop calls the subgraph module for graph $G$, and deletes all
	edges $e_1,\ldots,e_k$ in $G$.}*/

	void call (GraphCopy& G, leda_list<leda_edge>& L);
	/*{\Mop calls the subgraph module for graph copy $G$. Deletes all
	edges $e_1,\ldots,e_k$ in $G$, and returns in $L$ the list of
	corresponding edges $|original|(e_1),\ldots,|original|(e_k)$ in
	|original(G)|.}*/

	int num_del_edges() const {
		return _num_del_edges;
	}
	/*{\Mop returns the number $k$ of edges in $E\cap E'$.}*/

protected:

	/*{\Mtext
	\bigskip
	\headerline{Protected Overridables to Implement Functionality} }*/

	virtual void _call (const leda_graph& G, leda_list<leda_edge>& L)
	/*{\Moppure implements the computation of the subgraph for graph copy $G$.
	Returns in $L$ the list of edges in |original(G)| that correspond to
	deleted edges in $G$.}*/	
	= 0;

private:
	int _num_del_edges;
};



#endif
