//-------------------------------------------------------------------
//
//                         TaggedRule.h
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------


#ifndef _AGD_TAGGED_RULE_H
#define _AGD_TAGGED_RULE_H

#include <AGD/tr_literal.h>
#include <LEDA/list.h>


/*{\Manpage {TaggedRule} {} {Tagged Rules} {r} }*/

class TaggedRule {

/*{\Mdefinition
An instance |\Mvar| of the data type |\Mname| is a rule
$(p_1,i_1) \wedge \ldots \wedge (p_k,i_k) \rightarrow p$,
where $(p_j,i_j)$ $(1\leq j\leq k)$ are tagged rule
literals, and $p$ is a key. Tagged rules are used by modules to
define postcondition rules.
}*/

private:
	leda_list<tr_literal> _left;
	agd_key _right;

public:

	/*{\Mcreation}*/

	TaggedRule () {
		_right = nil;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow |nil|$.}*/

	TaggedRule (agd_key p) {
		_right = p;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $\rightarrow p$.}*/

	TaggedRule (const leda_list<tr_literal>& L, agd_key p) {
		_left = L;
		_right = p;
	}
	/*{\Mcreate creates an instances |\Mvar| of type |\Mtype|
	initialized to $l_1 \wedge\ldots\wedge l_k\rightarrow p$ if
	$L = l_1,\ldots,l_k$.}*/


	/*{\Moperations}*/

	agd_key right () const {
		return _right;
	}
	/*{\Mop returns the key p.}*/

	const leda_list<tr_literal>& left () const {
		return _left;
	}
	/*{\Mop returns the list of tagged literals
	$(p_1,i_1),\ldots,(p_k,i_k)$.}*/

	int length () const {
		return _left.length();
	}
	/*{\Mop returns the number $k$ of literals on the left side of |\Mvar|.}*/

	TaggedRule& and (agd_key q, int j = 1) {
		_left.append (tr_literal(q, j));
		return *this;
	}
	/*{\Mop extends |\Mvar| to $(p_1,i_1) \wedge \ldots \wedge
	(p_k,i_k) \wedge (q,j) \rightarrow p$. Returns |\Mvar|.}*/

	// Input & Output (currently dummy only)

	friend istream& operator>> (istream& is, TaggedRule& ) {
		return is;
	}

	friend ostream& operator<< (ostream& os, const TaggedRule& r);
	/*{\Mbinopfunc writes the tagged rule |r| to the output stream |os|.}*/

};

/*{\Mexample
We create a tagged rule $(|planar|,1) \wedge (|triconnected|,1)
\wedge (|simple|,1) \wedge (|no_self_loops|,1) \to |convex|$ and set it
as postcondition rule of a layout algorithm.

\small\tt
\begin{verbatim}
SampleLayout::SampleLayout() : LayoutModule (true)
{
  TaggedRule r(convex);

  r.and(planar).and(triconnected).and(simple).and(no_self_loops);

  add_post_rule(r);
  
    ....
}
\end{verbatim}
}*/


#endif
