
#ifndef _AGD_TOPOLOGY_MODULE_H
#define _AGD_TOPOLOGY_MODULE_H

#include <AGD/AGDModule.h>
#include <AGD/LayoutInterface.h>

/*{\Manpage {TopologyModule} {} {Topology Modules} {T} }*/

/*{\Mbaseclasses

\begin{hierarchy}{6.5}{2.5}
\hdclass{0}{1.5}{5.5}{AGDModule}
\arhook{5.5}{2}{6.5}
\hdclass{1.5}{0}{5.5}{TopologyModule}
\end{hierarchy}
}*/

class TopologyModule : public AGDModule {

/*{\Mdefinition
The purpose of class |\Mname| is to relate the geometrical and topological
structure of a graph.  An instance of type |\Mname| transports the information
from the actual drawing in the combinatorial structure of the underlying graph
$G$.  The output is a {\em planar representation} for $G$, i.e., a
combinatorial embedding defining the faces and an external face $f_0$. 

\headerline{Input and Output Parameters}

\begin{tabular}{lrl}
input parameter: & graph                  & $|in_graph|=1$\\
                 & layout                 & $|in_layout|=2$ \\
output parameter: & planar representation & $|out_plan_rep|=1$
\end{tabular}

The integer constants |in_graph|, |in_layout|, and |out_plan_rep| denote the
positions in the input and output parameter lists. The values can be used to
reference the desired parameter when setting the pre-- and postcondition.  
}*/

public:

  static const int in_graph, in_layout;
  static const int out_plan_rep;

  /*{\Minitialization T 5}*/
  
  TopologyModule();
  
  /*{\Mcreate initializes a topology module.}*/
  
  virtual ~TopologyModule() { }
  
  /*{\Moperations 1.5 2.5}*/

  leda_string name() const { return leda_string("Topology Module"); }
  
  leda_string long_name() const { 
    return leda_string("Topology Module"); 
  }
  
  leda_string author() const { return leda_string("Gunnar W. Klau"); }
  
  virtual leda_string impl_author() const { 
    return leda_string("Gunnar W. Klau"); 
  }
  
  virtual leda_string impl_version() const { return leda_string("1.0"); }
  
  virtual leda_string impl_date() const { return leda_string("June 1998"); }
  
  /*{\Mtext
  \headerline{Standard Interface} }*/

  bool check(const leda_graph& G, leda_string& message);
  /*{\Mop returns |true| if $G$ satisfies the precondition of |\Mvar|.
  Otherwise, |false| is returned, and |message| contains the
  error message.}*/

  leda_list<leda_edge> call(leda_graph& G, leda_face& f_0, 
    LayoutInterface& L);
  /*{\Mop calls the topology module for graph $G$ and layout interface
     $L$. The method changes the order of edges in $G$ such that the topological structure of the drawing is reflected. If the imposed combinatorial embedding is proper, the parameter $f_0$ contains the external face, otherwise |NULL|.}*/

  leda_list<leda_edge> call(leda_graph& G, LayoutInterface& L);
  /*{\Mop calls the topology module but only changes the order of edges
in $G$ according to the layout given by $L$. The external face is not determined.}*/    

  AGDModule *clone() const;

protected:

  bool get_planar_representation(leda_graph& G, LayoutInterface& L,
    leda_list<leda_edge>& M);

  leda_face get_outer_face(leda_graph& G, const LayoutInterface& L);

  double angle(DPoint p, DPoint q, DPoint r);

};

/*{\Mtext \headerline{Example}
An example program showing how to use a [[TopologyModule]] in combination with
a planar layout algorithm can be found in Section~\ref{The Orthogonal Drawing Algorithm} (p.~\pageref{ortho_example}).
}*/

#endif

