/************************************************************************

            Filename    :  TreeLayout.h

            Version     :  01.1997

            Author      :  Sebastian Leipert

            Language    :  C++

************************************************************************/


#ifndef _AGD_TREE_LAYOUT_H
#define _AGD_TREE_LAYOUT_H

#include <AGD/impl/TreeLayout/tree.h>
#include <AGD/LayoutModule.h>


/*{\Manpage {TreeLayout} {} {Walker's Algorithm} {L} }*/

/*{\Mbaseclasses

\begin{hierarchy}{8}{4}
\hdclass{0}{3}{5}{AGDModule}
\arhook{5}{3.5}{6}
\hdclass{1.5}{1.5}{5}{LayoutModule}
\arhook{6.5}{2}{7.5}
\hdclass{3}{0}{5}{TreeLayout}
\end{hierarchy}
\hfill\agdeps{tree.epsi}{5.5cm}
}*/

class TreeLayout : public LayoutModule
{
/*{\Mdefinition
The class |\Mname| represents the layout algorithm by Walker II
(Reingold and Tilford, Walker~\cite{rt-tdt-81,w-npagt-90,Lei96}).

This algorithm draws binary and general rooted trees. The graph has 
to be directed, connected, and has to have exactly one source. 
Furthermore the equality $\Labs{V}-1 = \Labs{E}$ must hold. 

The drawing is straight-line.
Nodes of the same level are placed on a horizontal line.   

\horizline
\commoninfo{TreeLayout}{Layout algorithm for general trees}{John Q. Walker II}
{Sebastian Leipert}{may 1997}{0.1}

\headerline{Pre- and Postcondition}
\begin{tabular}{lcl}
precondition & $=$ & $\{\, |tree|,|single source| \,\}$ \\
postcondition(|PRE|) & $=$ & $\{\, \mbox{|straight|-|line|},|no crossings|,|upward planar| \,\}$
\end{tabular}

\headerline{Optional Parameters}
Instances of |\Mname| provide the following optional parameters:
\begin{options}
\item |double| {\bf sib\_separation}\\
The spacing between two sibling nodes.
\item |double| {\bf subtree\_separation}\\
The spacing between two adjacent subtrees.
\item |double| {\bf level\_separation}\\
The spacing between the nodes of two adjacent levels.
\item |double| {\bf left\_margin}\\
Additional spacing at the left side of the drawing.
\item |double| {\bf upper\_margin}\\
Additional spacing at the upper side of the drawing.
\end{options}\horizline
}*/

public:

	/*{\Mcreation}*/

   TreeLayout();
	/*{\Mcreate creates an instance |\Mvar| of type |\Mname|.}*/

   virtual ~TreeLayout();


  /*{\Moperations}*/

   leda_string name() const;

   leda_string long_name() const;

   leda_string author() const;

   leda_string impl_author() const;

   leda_string impl_version() const;

   leda_string impl_date() const;


   double sib_separation() const;

   void sib_separation(double s);

   double level_separation() const;

   void level_separation(double s);

   double subtree_separation() const;

   void subtree_separation(double s);

   double left_margin() const;

   void left_margin(double s);

   double upper_margin() const;

   void upper_margin(double s);

	AGDModule *clone () const;

	/*{\Moptions warnings=no}*/

	/*{\Mtext       
	\headerline{Generic Interface (Inherited Methods)}
	The detailed description of these methods can be found in the
	manual entries of the base class (|LayoutModule|).
	\setlength{\entrysep}{-1ex}}*/

	/* inherited
	virtual bool check (const graph& G, string& message);
	*/
	/*{\Mop}*/

	/* inherited
	void call (const graph& G, LayoutInterface& A);
	*/
	/*{\Mop}*/
	
	/*{\Mtext \setlength{\entrysep}{0.5ex}}*/

protected:

   void _call(const leda_graph& G, LayoutInterface& a);
   void _call (GraphCopy& GC, LayoutInterface& A)
   { LayoutModule::_call (GC,A); }

private:

   bool Initialize(const leda_graph& G,LayoutInterface &a);
   
   Tree    _T;
   leda_node    _root;
   leda_node  *_arrayOfNodes;
};

/*{\Mimplementation
The running time of the algorithm is $O(\Lvert V\Lvert)$.}*/

#endif


