//-------------------------------------------------------------------
//
//                           bctree.h
//
//    author: Sergey Fialko
//    date:   february 1997
//-------------------------------------------------------------------


#ifndef BCTREE
#define BCTREE

#include <time.h>
#include <LEDA/stack.h>
#include <LEDA/set.h>
#include <LEDA/p_queue.h>
#include <LEDA/map.h>
#include <LEDA/d_array.h>
#include <LEDA/array2.h>
#include <LEDA/node_set.h>
#include <LEDA/graph.h>
#include <LEDA/graph_alg.h>

class block_struct;
class cutpoint_struct;
class bc_node_struct;
class bctree;

typedef bc_node_struct* bc_node;
typedef cutpoint_struct* cutpoint;
typedef block_struct* block;

// inserted forward declarations for compare functions
// in order to compile with MS Visual C++
// (Carsten Gutwenger, feb. 1997) changed july 97
// Compare Funktionen
//***************************************************************
inline int compare (cutpoint const &cp1, cutpoint const &cp2)
              { return compare ((long)cp1, long(cp2)); }

//***************************************************************
inline int compare (block const &bl1, block const &bl2)
              { return compare ((long)bl1, (long)bl2); }
//***************************************************************
inline int compare (bc_node const &bcn1, bc_node const &bcn2)
              { return compare ((long)bcn1, (long)bcn2);}


typedef list<node> nodelist;
typedef list<edge> edgelist;
typedef list<block> blocklist;
typedef list<cutpoint> cutpointlist;
typedef list<bc_node> bcnodelist;
typedef set<cutpoint> cutpointset;
typedef set<face> facesset;
typedef set<node> nodeset;
typedef set<block> blockset;

list<edge> onehalf (graph &);

class bc_node_struct { //Common basic class for Block and Cutpoint (CP)
   friend class bctree;

   list<bc_node> adj_nodes; //adj bc_nodes
   int dfs_number;

   int continue_DFS (int, bc_node); // For DFS
   bc_node father;  //according to DFS-Numbering
   bc_node sohn (); //according to DFS-Numbering

   inline bc_node first_adj ()
     { return (adj_nodes.empty ()) ? nil : adj_nodes.head ();}

   inline bc_node last_adj () 
     { return (adj_nodes.empty ()) ? nil : adj_nodes.tail (); }

   inline void add_bc_node ( bc_node bcn)
     { adj_nodes.push (bcn); }

   inline void del_bc_node ( bc_node bcn); //Removes bcn from adj_nodes

 protected:
// These Functions will be overwritten in Block and Cutpoint
   virtual node first_node () { return nil; }
   virtual bool is_cutpoint () { return true; }
   virtual list<node> all_nodes () { list<node> result; return result; }
   virtual int number_of_nodes () { return 0; }

 public: 
   virtual ~bc_node_struct () {}
   inline friend int degree ( bc_node const& bcn )
     { return (bcn->adj_nodes).size(); }

   inline friend int dfs_number ( bc_node const& bcn )
     { return bcn->dfs_number; }

   inline friend bc_node father ( bc_node bcn)
     { return bcn->father; }
   inline friend bc_node sohn (bc_node bcn)
     { return bcn->sohn (); }
   inline friend bool is_cutpoint ( bc_node bcn )
     { return bcn->is_cutpoint (); }

// Returns the first node of bc_node: 
// if bc_node is cutpoint then it is its corresponding node from graph.
// bc_node is block: first non-cutpoint node belonging to block. 
// (Note: a block can consist of only cutpoints. In this case the function
// will return nil.

   inline friend node first_node ( bc_node const& bcn )
     { return bcn->first_node (); }

//=1 for cutpoint, # of non-cutpoint nodes for blocks
   inline friend int number_of_nodes (bc_node const& bcn)
     { return bcn->number_of_nodes (); }

   inline friend bc_node first_adj ( bc_node bcn )
     { return bcn->first_adj (); } 

   inline friend bc_node last_adj ( bc_node bcn )
     { return bcn->last_adj (); }
};

//************************************************************************
class cutpoint_struct : public bc_node_struct {
  friend bctree;
  node knoten;

  cutpoint_struct (node v) { knoten = v; }
  inline bool is_cutpoint () { return true; }
  inline node first_node () { return knoten; }
  inline list<node> all_nodes ()
    { list<node> result; result.push (knoten); return result; }
  inline int number_of_nodes () { return 1; }

public:
  friend int degree_ord ( cutpoint const& cp ) //a compare function
   { return degree (cp); }
  friend node vertex ( cutpoint const& cp)
   { return cp->knoten; } 
};

//**********************************************************************
class block_struct : public bc_node_struct {
   friend class bctree;
   list<node> nodes;

   inline bool is_cutpoint () { return false; }
   inline int number_of_nodes () { return nodes.size (); }
   inline node first_node () { return (nodes.empty ()) ? nil : nodes.head (); }
   inline list<node> all_nodes () { return nodes; }

  public:
   inline friend list<node>& all_block_nodes (block bl)
     { return bl->nodes; }
};

//***************************************************************
class bctree {
   list<cutpoint> cutpoints;
   list<block> blocks;
   bc_node root_; //the Root of tree
   const graph* G;

// Here will stored to which bc_nodec belong the nodes of graph
   node_map<bc_node> node_owner;

   int num_connected_comps; //the number of connected components

   list<bc_node> roots_; //Roots of connected components (CC)

   int compute_SubBCTree (node start = nil); //computes bctree for a CC

   void compute_bcTree (); //computes bctree for all CC's of graph

// Auxiliary for compute_bcTree
   node find_unchecked (node v,edge_array<bool> &, bool only_check = false );

   cutpoint new_cutpoint (node v); //Adds new cutpoint with vertex v to list

// Adds new  cutpoint and makes it adjacent to bl.
   cutpoint new_cutpoint (node v, block bl);

   block new_block (stack<node>&, node);
   inline block new_block (node n) 
     { block bl = new_block (); add_node (n, bl);  return bl; }

   inline block new_block ()
     { block bl = new block_struct; blocks.push (bl); return bl; }

   inline void make_nodes_adj (bc_node bcn1, bc_node bcn2 );

   inline void add_node ( node v, block bl )
     { (bl->nodes).push (v); node_owner[v] = bl; }

   void del_node (node v, block bl);//Removes v from list of nodes of bl

   void del_bc_node (bc_node);

   void coalesce_path (bc_node start, bc_node end, list<block>&);

 public:
   bctree () {}

//Constructor: construct = true -> initialize the bctree.
// DFS_num -> DFS-Numbering
   bctree (const graph* G, bool construct = false, bool DFS_num = false);

   ~bctree ();
// bctree will be computed automatically
   void init (graph* G, bool DFS = false);

//reinit
   void update (bool DFS = false);

//Creates a graph in LEDA-Format. First parameter is the Graph that will be 
//created, every node in this Graph has a pointer to original bc_node from bctree.
//The type of edges is no matter. Second Parameter stores the back correspondence
//from bc_nodes to nodes in LEDA-Graph
   void LEDA_Graph ( GRAPH <bc_node, int>& LEDA_Gr,
                     d_array<bc_node, node>& to_copy );

   bool biconnected () { return !number_of_cutpoints (); }

   void DFS_num (bc_node = nil); //Parameter - the start node 

   inline int dfs_number (bc_node bcn) { return bcn->dfs_number; }

   inline bc_node root () { return root_;}

   inline list<bc_node>& roots () { return roots_; }

   inline bc_node father (bc_node bcn ) { return bcn->father; }

   inline bc_node sohn (bc_node bcn) { return bcn->sohn (); }

//Finds the lowest  common ancestor for two nodes
   bc_node common_ancestor (bc_node, bc_node);

//Depth of the bc_node in graph
   int depth (bc_node);

//Computes path between two bc_nodes
   list <bc_node> path (bc_node, bc_node);

   bc_node granddad ( bc_node );

   inline int degree (bc_node bcn )
     { return (bcn->adj_nodes).size (); }

   inline bc_node first_adj_bcnode (bc_node bcn)
     { return bcn->first_adj (); }

   inline list<bc_node>& adj_nodes (bc_node bcn)
     { return bcn->adj_nodes; }

//TRUE if v is a CP and this CP is adjacent to bl.
   bool is_adj_cutpoint (node v, block bl);

   inline bool are_adj (bc_node, bc_node);

// Teue, iff c1 and c2 belong to the same block
   block same_block (cutpoint c1, cutpoint c2);
   
//Returns the vertex of first cutpoint adjacent to block
   node head_node ( block bl); 

   inline bool is_cutpoint (bc_node bcn)
     { return bcn->is_cutpoint(); }

   inline cutpoint is_cutpoint (node v);

   inline node vertex (cutpoint cp) { return cp->knoten; }

   inline node first_node ( bc_node bcn )
     { return bcn->first_node (); }

   inline list<node> all_nodes ( bc_node bcn )
     { return bcn->all_nodes (); }

   inline int number_of_nodes ( bc_node bcn )
     { return bcn->number_of_nodes (); }

//Checks for bc_node in cutpoints or blocks
   bool is_valid (bc_node );

   inline bc_node owner (node v)
     { return node_owner[v]; }

   inline int number_of_blocks ()
     { return blocks.length(); }

   inline int number_of_cutpoints ()
     { return cutpoints.length(); }

   inline int number_of_bcnodes () 
     { return blocks.length()+ cutpoints.length(); }

   inline int number_of_components ()
     { return num_connected_comps; }

   int number_blocks_with_degree (int deg );

   inline block first_block ()
     { return ( blocks.empty ()) ? nil : blocks.head (); }

   inline list<node>& all_nodes (block bl)
     { return bl->nodes; }

   inline list<block>& all_blocks() 
     { return blocks; }

   inline list<cutpoint>& all_cutpoints() 
     { return cutpoints; }
   list<bc_node> all_bcnodes ();

   list<block> all_blocks_with_degree (int deg );

   void print_node (bc_node);

//Two following functions update *this after adding of new edge between
//two different blocks. The second function works only if connected blocks
//are pendants
   block union_nodes (bc_node, bc_node);

   block union_pendants (cutpoint, list<block>&);

   void connect_components (edge e); //is not tested

//Returns the first node which:
// a) is adjacent to cp
// b) belongs to bl
// c) is not cutpoint
// If such node no exist - returns nil.
   node adj_to_cutpoint_node ( block bl, cutpoint cp = nil );
};

//***************************************************************************
class non_planar_heaps  {
  d_array <cutpoint, cutpointset> non_planars;

public:
  non_planar_heaps() { }
  virtual ~non_planar_heaps() { }

  inline cutpointset* heap_of (cutpoint cp) {
     if (!non_planars.defined (cp))  return nil;
     else return &(non_planars [cp]);
  }
  bool are_non_planar (cutpoint, cutpoint);
  void add_info (cutpoint, set<cutpoint>);
  void add_np_pair (cutpoint, cutpoint);
};
#endif
