#ifndef PL_AUGMENT
#define PL_AUGMENT

#include <LEDA/matrix.h>
#include <LEDA/d_array.h>
#include <LEDA/slist.h>
#include <LEDA/stream.h>

#if __LEDA__ >= 360
#undef set
#undef string
#endif

#include <abacus/master.h>
#include <abacus/variable.h>
#include <abacus/constraint.h>
#include <abacus/sub.h>
#include <abacus/numvar.h>
#include <abacus/srowcon.h>
#include <abacus/pool.h>
#include <abacus/standardpool.h>
#include <abacus/nonduplpool.h>
#include <abacus/fsvarstat.h>
#include <abacus/string.h>
#include <abacus/cutbuffer.h>
#include <abacus/setbranchrule.h>
#include <abacus/poolslot.h>
#include <abacus/opensub.h>

#if __LEDA__ >= 360
#define set leda_set
#define string leda_string
#endif

#include <AGD/impl/bctree.h>
#include <AGD/impl/cons.h>
#include <AGD/impl/edgevar.h>

class PLAU_SUB;
class PLANAR_CONSTRAINT;

typedef list<PLAU_EDGEVAR*> edgevarlist;
typedef set<PLAU_EDGEVAR*> edgevarset;
typedef PLAU_EDGEVAR* evar;

//***************************************************************************
// The derived from ABA_MASTER class.

class PLAU_MASTER : public ABA_MASTER
{
    friend class PLAU_SUB;
    friend class PLAU_EDGEVAR;
    friend class PLANAR_CONSTRAINT;

 private:
    inline ABA_SUB* firstSub ();
    graph* originalG;

// G is the scratch Graph. Every node in G has a pointer to its corresponding
// node in originalG. We store also for every edge in G pointer to its
// corresponding PLAU_EDGEVAR
    GRAPH <node, PLAU_EDGEVAR*> G;

    bctree BC_;

// the Pool for dynamically generated constraints (Kuratowski subdivisions)
    ABA_NONDUPLPOOL <ABA_CONSTRAINT, ABA_VARIABLE> *Cuts;

// returns a PLAU_EDGEVAR* with number i
    inline PLAU_EDGEVAR* evar (int);
   
    int counter;
    
    int nPlanar;
    
    int nNormalSeparate;

    ABA_CPUTIMER *separTime;

    int MaxSolutionLevel;
 
    int OptSolutionLevel;

    int AveLPLevel;

    bool balanced_;

    int LowerB; //result of Approximation

// We will add every new *PLAU_EDGEVAR to this sets in this dictionary: the set of
// source-Block and one of target-Block. All Pendants and "BadBlocks" have
// its own set of incident edges.
    d_array <bc_node, edgevarset*> block_edges;
    
    d_array <bc_node, int> good_edges;
    
    inline int ngood_edges (bc_node b) { return good_edges[b]; }

//Returns the set of edges incident to bl
    inline set<PLAU_EDGEVAR*>* edges_for_block (bc_node bl);

    list<cutpoint> adj_to_BadBlocks;

    slist <ABA_CONSTRAINT*> bicons;

    list<int> ToFixVars;

//    ABA_CONSTRAINT* not_violated;

    void compute_all_partitions (list<int>, set<PLAU_EDGEVAR*>, d_array <int, edgevarset*>& sets, int );

    void initializeOptimization ();

    void terminateOptimization ();

    void computeSparseGraph (ABA_BUFFER<ABA_VARIABLE*> &);

    void make_copy ();

    set<PLAU_EDGEVAR*>* compute_bicons (bc_node bcn);
    
    ABA_CONSTRAINT* createCon (set<PLAU_EDGEVAR*>*); //BiConnectivity constraint
    
    PLANAR_CONSTRAINT* createCon (list<int>& evars);

    edgelist best_found_solution;//The best solution will be stored here.
             //If we have not found a solution which is better than
	     //our approxim. solution (LowerB) then the list remains empty.

    void output ();

    inline PLAU_EDGEVAR* evar_for ( edge e ) { return G.inf(e); }

    bool update_solution ( list<int>& );

    inline graph* Graph ()  { return &G; }

    inline bctree* BCtree () { return &BC_; }

    inline void removeVar ( PLAU_EDGEVAR* evar) { remove_edge (evar); }

    edge restore_edge ( PLAU_EDGEVAR* evar );

    void remove_edge (PLAU_EDGEVAR *evar);
    
    bool planar_edge (int i);

    void print_evar (PLAU_EDGEVAR* evar, double val = 0 );

    inline int nVars () 
    { return varPool()->number (); }

//    inline ABA_STANDARDPOOL<ABA_VARIABLE, ABA_CONSTRAINT>* vPool ()
//    { return (ABA_STANDARDPOOL<ABA_VARIABLE, ABA_CONSTRAINT>*)varPool (); }

//    inline ABA_POOLSLOT<ABA_VARIABLE,ABA_CONSTRAINT>* slot (int i)
//    { return varPool()->slot (i); }

    inline ABA_NUMVAR* variable (int i)
    { return (ABA_NUMVAR*)(varPool()->slot(i)->conVar ()); }

    int UpperBound;

    bool branch;

    bool wrongSet;
    
// Parameters
    bool LogStatisticsToFile;
    ABA_STRING StatisticsFileName;
    bool UseNewBranchingStrategy;
    int MaxOpenSubs;
    ENUMSTRAT origStrat;

  protected:
    int equalSubCompare(ABA_SUB*, ABA_SUB*);
  public:
    PLAU_MASTER (graph& Gr,
         int LowerBound = 0,
	 double eps = 1.0e-6,
	 double machineEps = 1.0e-10);
    ~PLAU_MASTER ();

    inline edgelist& solution () { return best_found_solution; }
    inline bool found ()
       { return !best_found_solution.empty () ? true : false; }
};

//****************************************************************************
class PLAU_SUB : public ABA_SUB //Derived from class ABA_SUB for Planar Augmentation
{
  public :
    PLAU_SUB (ABA_MASTER* );

    PLAU_SUB (ABA_MASTER* , ABA_SUB *father, ABA_BRANCHRULE *brule);

    bool feasible ();

    ABA_SUB* generateSon (ABA_BRANCHRULE *brule)
    {  return new PLAU_SUB (master_, this, brule); }
 
    inline int NumOfSet2One () { return NumOfSet2One_; }

  private:
    inline PLAU_MASTER* plau_master () { return (PLAU_MASTER*)master_; }

    inline graph* Graph () { return plau_master()->Graph (); }

    inline bctree* BCtree () { return plau_master()->BCtree (); }
    
    inline bc_node s_owner (int i) { return BCtree()->owner (edge_var (i)->source ());}
    
    inline bc_node t_owner (int i) { return BCtree()->owner (edge_var (i)->target ());}

    inline PLAU_EDGEVAR* edge_var ( ABA_VARIABLE* var ) { return (PLAU_EDGEVAR*)var; }

    inline PLAU_EDGEVAR* edge_var ( int i ) { return (PLAU_EDGEVAR*)variable (i);  }

    inline void remove_edge (int i) { plau_master()->remove_edge (edge_var(i)); }

    inline edge restore_edge (int i) { return plau_master()->restore_edge (edge_var(i)); }

    int separatePlanarCuts (list<int>&, list<int>& scnd_choice);
    
    int quickSeparate (list<int>& edges, ABA_BUFFER <ABA_CONSTRAINT*> &buff, bool& violated);

    int SeparationFromCutsPool (list<int>&, list<PLANAR_CONSTRAINT*>&);

    void filter_solution (list<int>&, list<int>&, list<int>&);

    inline edge Edge ( ABA_VARIABLE* var ) { return ((PLAU_EDGEVAR*)var)->Edge (); } 

    inline edge Edge ( int i ) { return Edge (variable(i)); }

    inline kind_of_edge kind (int i)  { return edge_var(variable (i))->kind (); }

    inline bool IsFixed (int i) { return variable(i)->fsVarStat()->fixed (); }

    inline bool IsFixed (PLAU_EDGEVAR* e) { return e->fsVarStat()->fixed (); }
#undef set    
    inline bool IsLocalSet (int i) { return fsVarStat (i)->set (); }
#define set leda_set
   
    int NumOfSet2One_;

    int BranchingVar;

    int  oldBranchingStrategy (list<int>&);

    int  newBranchingStrategy (list<int>&, list<int>&);
    
  protected:
    void activate ();
 
    int  selectBranchingVariable ( int& );

    void fixByLogImp(ABA_BUFFER<int>&, ABA_BUFFER<ABA_FSVARSTAT*>&);
    
    bool exceptionBranch();

    bool exceptionFathom ();

    void fathom (bool reoptimize);
};

/**********************************************************************/
//******************************************************************************
extern inline int number (PLAU_EDGEVAR* const &evar);

#endif
