//-------------------------------------------------------------------
//
//                        FPPLayout.cpp
//
//    author: Carsten Gutwenger
//    date:   octobre 1997
//-------------------------------------------------------------------


#include <AGD/FPPLayout.h>
#include <AGD/PlaneGraphCopy.h>


FPPLayout::FPPLayout () : GridLayoutModule(true)
{
	add_precondition (key::planar);
	add_precondition (key::simple);
	add_precondition (key::no_self_loops);

	add_post_rule (key::planar_layout, 0);
	add_post_rule (key::straight_line);
	add_post_rule (key::no_crossings);
}

AGDModule *FPPLayout::clone() const
{
	FPPLayout *M = new FPPLayout;

	M->set_coordinate_mapper (get_coordinate_mapper());

	return M;
}


void FPPLayout::_call_grid (PlaneGraphCopy& G, face F, GridLayout &gl)
{
	node v1, v2;

	// handle graphs with less than 3 nodes
	switch (G.number_of_nodes()) {
	case 0:
		return;
	case 1:
		v1 = G.first_node();
		gl.x(v1) = gl.y(v1) = 0;
		return;
	case 2:
		v1 = G.first_node();
		v2 = G.succ_node (v1);
		gl.x(v1) = gl.y(v1) = gl.y(v2) = 0;
		gl.x(v2) = 1;
		return;
	}
//----
	make_biconnected (G);
	make_bidirected (G);
	if (embed_planar_graph (G) == false)
		error_handler(0,"FPPLayout: cannot embed graph!");
	make_map (G);
//----
	G.triangulate_planar_map();

	edge e_12 = G.first_edge(),
		 e_2n = G.cyclic_adj_pred (G.reversal(e_12));

	node_array<int>  num (G);
	node_array<edge> e_wp (G), e_wq (G);

	compute_order (G, num , e_wp, e_wq, e_12, e_2n,
		G.cyclic_adj_pred(G.reversal(e_2n)));
	compute_coordinates (G, num, e_wp, e_wq, gl.x(), gl.y());
}

void FPPLayout::compute_order (const graph &G, node_array<int> &num, node_array<edge> &e_wp,
	node_array<edge> &e_wq, edge e_12, edge e_2n, edge e_n1)
{
	node_array<int>       num_diag (G, 0);	// Anzahl der Diagonalen
	node_array<list_item> link (G, nil);	// link[v] = Listitem in possible, das v enthlt, falls diag[v] = 0 und outer[v] = TRUE
	node_array<int>       outer (G, false);	// outer[v] = TRUE <=> v ist Knoten der aktuellen Auenregion

	list<node> possible;	// Liste aller Knoten v mit outer[v] = TRUE und diag[v] = 0
	// Knoten der Auenregion ( = Dreieck(v_1,v_2,v_n) )
	node v_1 = G.source (e_12), v_2 = G.target (e_12), v_n = G.target (e_2n);
	node v_k, wp, wq, u;
	edge e, e2;
	int k;

	 // Initialisierung (Beginne mit Auenflche (v_1,v_2,v_n) und v_n einziger mglicher nchster Knoten
	num [v_1] = 1;
	num [v_2] = 2;
	outer [v_1] = outer [v_2] = outer [v_n] = true;
	link [v_n] = possible.append (v_n);
	e_wq [v_1] = G.reversal(e_n1); e_wp [v_2] = e_2n;
	e_wp [v_n] = e_n1; e_wq [v_n] = G.reversal(e_2n);

	// Whle nacheinander v_k und entferne v_k
	for (k = G.number_of_nodes(); k >= 3; k--) {
		v_k = possible.pop();	// Whle beliebigen Knoten aus possible als v_k
		num [v_k] = k;

		// Bestimme Vorgnger wp und Nachfolger wq von vk in C_k (= aktuelle Auenregion)
		wp = G.target (e_wp [v_k]);
		wq = G.target (e_wq [v_k]);

		// v_k nicht mehr in C_k-1 (= neue Auenregion)
		outer [v_k] = false;

		// Wegfall einer Diagonalen ?
		if (G.target (G.cyclic_adj_pred (e_wp[v_k])) == wq) {	// wp, wq einzige Nachfolger von vk in G_k
			// dann: wp, wq verlieren eine Diagonale
			if (--num_diag[wp] == 0) link [wp] = possible.append (wp);
			if (--num_diag[wq] == 0) link [wq] = possible.append (wq);
		}

		// Aktualisiere bzw. initialisiere e_wq, e_wp
		e_wq [wp] = G.cyclic_adj_succ (e_wq[wp]);
		e_wp [wq] = G.cyclic_adj_pred (e_wp[wq]);
		e = e_wq[wp];
		for (u = G.target(e); u != wq; u = G.target(e)) {
			outer [u] = true;
			e_wp [u] = G.reversal(e);
			e = e_wq [u] = G.cyclic_adj_succ (G.cyclic_adj_succ(e_wp[u]));

			// Suche neue Diagonalen
			for (e2 = G.cyclic_adj_pred(e_wp[u]); e2 != e_wq[u]; e2 = G.cyclic_adj_pred(e2)) {
				node w = G.target(e2);
				if (outer [w] == true) {
					++num_diag [u];
					if (w != v_1 && w != v_2)
						if (++num_diag [w] == 1) possible.del_item (link[w]);
				}
			}
			if (num_diag [u] == 0) link [u] = possible.append (u);
		}

	}
}

void FPPLayout::compute_coordinates (const graph &G, node_array<int> &num,
	node_array<edge> &e_wp, node_array<edge> &e_wq,
	node_map<int>& x, node_map<int>& y)
{
	const int n = G.number_of_nodes();
	node_array<int>  x_rel (G);
	node_array<node> upper (G);
	node_array<node> next (G);
	array<node> v (1, n);
	node w, vk, wp, wq;
	int k, xq, dx;

	// Initialisierung
	forall_nodes (w, G)
		v [num [w]] = w;

	x_rel [v[1]] = x_rel [v[2]] = y [v[1]] = y [v[2]] = 0;
	next [v[1]] = v[2]; next [v[2]] = nil;

	// Hauptschleife
	for (k = 3; k <= n; k++) {
		vk = v[k];
		wp = G.target (e_wp [vk]);
		wq = G.target (e_wq [vk]);

		xq = 2;
		w = wp;
		do {
			w = next [w];
			xq += x_rel [w];
		} while (w != wq);

		x_rel [vk] = (xq + y[wq] - y[wp]) / 2;
		y  [vk] = (xq + y[wq] + y[wp]) / 2;
		x_rel [wq] = xq - x_rel[vk];

		dx = 1;
		for (w = next[wp]; w != wq; w = next[w]) {
			dx += x_rel[w];
			x [w] = dx-x_rel[vk]; upper [w] = vk;
		}

		next [wp] = vk; next [vk] = wq;
	}

	// Bestimme absolute x-Koordinate auf Auenregion
	x [v[n]] = x_rel[v[n]];
	x [v[2]] = x [v[n]] + x_rel[v[2]];
	x [v[1]] = 0;

	// Bestimme absolute x-Koordinate fr alle anderen (inneren) Knoten
	for (k = n-1; k >= 3; k--)
		x [v[k]] += x [upper[v[k]]];

	// compute grid bounding box

	int xl = 0, xr = 0, yb = 0, yt = 0;
	switch(n) {
	case 0:
	case 1:
		break;
	case 2:
		xr = 1;
		break;
	default:
		yt = n-2;
		xr = 2*(n-2);
	}
	set_grid_bbox(IRect(xl,xr,yb,yt));
}
