//-------------------------------------------------------------------
//
//                      GridLayoutModule.cpp
//
//    author: Carsten Gutwenger
//    date:   may 1998
//-------------------------------------------------------------------


#include <AGD/GridLayout.h>
#include <AGD/GridLayoutModule.h>
#include <AGD/PlaneGraphCopy.h>
#include <AGD/GraphCopyLI.h>

//-------------------------------------------------------------------
//                           GridLayout
//-------------------------------------------------------------------

void GridLayout::init (const graph &G)
{
	_x.init(G,0);
	_y.init(G,0);
	_bends.init(G);
	gp = &G;
}

void GridLayout::init ()
{
	_x.init();
	_y.init();
	_bends.init();
	gp = nil;
}

int GridLayout::number_of_bends() const
{
	int num = 0;

	edge e;
	forall_edges(e,*gp)
		num += _bends[e].length();

	return num;
}

IPolyline GridLayout::get_compact_bends(edge e) const
{
	IPolyline ip = _bends[e];

	if (ip.length() == 0) return ip;

	ip.push_front(IPoint(_x[source(e)],_y[source(e)]));
	ip.push_back (IPoint(_x[target(e)],_y[target(e)]));

	_compact(ip);
	ip.pop_front(); ip.pop_back();
	return ip;
}

IPolyline GridLayout::get_compact_bends(const GraphCopy& GC, edge e) const
{
	IPolyline ipl;
	IPoint ip;
	const list<edge>& ch = GC.chain(e);
	edge e1;

	forall (e1, ch) {
		ipl.append(IPoint(_x[source(e1)],_y[source(e1)]));

		if (!_bends[e1].empty()) {
			forall (ip, _bends[e1])
				ipl.append(ip);
		} else if (GC.is_map()) {
			forall_rev (ip, _bends[GC.reversal(e1)])
				ipl.append(ip);
		}
	}
	ipl.append(IPoint(_x[target(e1)],_y[target(e1)]));

	if (GC.is_reversed(e))
		ipl.reverse_items();

	_compact(ipl);
	ipl.pop_front(); ipl.pop_back();
	return ipl;
}

void GridLayout::compact_all_bends()
{
	edge e;
	forall_edges(e,*gp)
		_bends[e] = get_compact_bends(e);
}

bool GridLayout::_is_redundant(IPoint &p1, IPoint &p2, IPoint &p3)
{
	int dzy1 = p3.xcoord() - p2.xcoord();
	int dzy2 = p3.ycoord() - p2.ycoord();
	int dyx1 = p2.xcoord() - p1.xcoord();

	if (dzy1 == 0) return (dyx1 == 0 || dzy2 == 0);

	int f = dyx1 * dzy2;

	return (f % dzy1 == 0 && (p2.ycoord() - p1.ycoord()) == f / dzy1);
}

void GridLayout::_compact(IPolyline& ip)
{
	IPoint p;
	list_item it = ip.first();
	p = ip[it]; it = ip.succ(it);
	for (it = ip.succ(it); it != nil; it = ip.succ(it)) {
		list_item it_pred = ip.pred(it);
		if (_is_redundant(p,ip[it_pred],ip[it])) {
			ip.del_item(it_pred);
		} else {
			p = ip[it_pred];
		}
	}
}


IRect GridLayout::area() const
{
	if (!gp || gp->empty()) return IRect();

	node v = gp->first_node();
	int xmin = _x[v], xmax = _x[v],
		ymin = _y[v], ymax = _y[v];

	forall_nodes (v, *gp) {
		if (_x[v] > xmax) xmax = _x[v];
		if (_x[v] < xmin) xmin = _x[v];
		if (_y[v] > ymax) ymax = _y[v];
		if (_y[v] < ymin) ymin = _y[v];
	}

	edge e;
	forall_edges (e, *gp) {
		const IPolyline& points = _bends[e];

		IPoint ip;
		forall (ip, points) {
			if (ip.xcoord() > xmax) xmax = ip.xcoord();
			if (ip.xcoord() < xmin) xmin = ip.xcoord();
			if (ip.ycoord() > ymax) ymax = ip.ycoord();
			if (ip.ycoord() < ymin) ymin = ip.ycoord();
		}
	}

	return IRect(xmin,xmax,ymax,ymin);
}

void GridLayout::supplement_map_layout()
{
	edge e;
	forall_edges (e, *gp)
		if (_bends[e].empty()) {
			_bends[e] = _bends[gp->reversal(e)];
			_bends[e].reverse_items();
		}
}

//-------------------------------------------------------------------
//                         GridLayoutModule
//-------------------------------------------------------------------

// overridden call methods

void GridLayoutModule::call (const graph& G, LayoutInterface& A)
{
	_cg_helper(G,0,&A);
}

/*{
	GridLayout gl;

	if (postcondition(0).member(key::planar_layout)) {
		PlaneGraphCopy PG(G);

		_init_grid(PG,gl);

		if (use_copy())
			_call_grid (PG, nil, gl);
		else
			_call_grid ((const graph&)PG, nil, gl);

		_cleanup_grid(gl);
		gcm->call (PG, gl, _grid_bbox, A);

	} else {
		if (use_copy()) {
			GraphCopy GC(G);
			_init_grid(GC,gl);
			_call_grid(GC,gl);
			_cleanup_grid(gl);
			gcm->call (GC,gl,_grid_bbox,A);

		} else {
			_init_grid(G,gl);
			_call_grid(G,gl);
			_cleanup_grid(gl);
			gcm->call(G,gl,_grid_bbox,A);
		}
	}
	_finish_call(A,0);
}*/


void GridLayoutModule::call_grid (const graph& G, GridLayout &gl)
{
	_cg_helper(G,&gl,0);
}

void GridLayoutModule::_cg_helper (const graph& G, GridLayout *gl,
	LayoutInterface *ap)
{
	if (postcondition(0).member(key::planar_layout)) {
		PlaneGraphCopy PG(G);
		GridLayout _gl;

		_init_grid(PG,_gl);
		if (!PG.empty()) {
			if (use_copy())
				_call_grid (PG, nil, _gl);
			else
				_call_grid ((const graph&) PG, nil, _gl);
		}
		_cleanup_grid(_gl);

		if (ap)
			gcm->call (PG, _gl, _grid_bbox, *ap);
		else
			transform_grid_layout (PG, *gl, _gl);

	} else {
		if (use_copy()) {

			GraphCopy GC (G);
			GridLayout _gl;

			_init_grid(GC,_gl);
			if (!GC.empty())
				_call_grid (GC, _gl);
			_cleanup_grid(_gl);

			if (ap)
				gcm->call (GC, _gl, _grid_bbox, *ap);
			else
				transform_grid_layout (GC, *gl, _gl);

		} else {
			if (ap) gl = new GridLayout;
			_init_grid(G,*gl);
			if (!G.empty())
				_call_grid (G, *gl);
			_cleanup_grid(*gl);
			if (ap) {
				gcm->call (G, *gl, _grid_bbox, *ap);
				delete gl;
			}
		}
	}

	if (ap) 
		ap->update();
	else
		gl->compact_all_bends();
}

void GridLayoutModule::call (const graph& G, face F, LayoutInterface& A)
{
	_cg_helper(G, F, 0, &A);
}

void GridLayoutModule::call_grid (const graph& G, face F, GridLayout &gl)
{
	_cg_helper (G, F, &gl, 0);
}

void GridLayoutModule::_cg_helper (const graph& G, face F, GridLayout *gl,
	LayoutInterface *ap)
{
	if (!postcondition(0).member(key::planar_layout)) {
		error_handler(0,"GridLayoutModule::call_grid(G,F,gl): not a planar layout algorithm.");
		return;
	}

	if (use_copy()) {
		PlaneGraphCopy PG(G);
		GridLayout _gl;

		_init_grid(PG,_gl);
		if (!PG.empty()) {
			face F0 = (F != nil) ? PG.copy_of_face(F) : nil;
			_call_grid (PG, F0, _gl);
		}
		_cleanup_grid(_gl);
		
		if (ap)
			gcm->call (PG, _gl, _grid_bbox, *ap);
		else
			transform_grid_layout (PG, *gl, _gl);

	} else {
		if (ap) gl = new GridLayout;
		_init_grid(G,*gl);
		if (!G.empty())
			_call_grid (G, F, *gl);
		_cleanup_grid(*gl);
		if (ap) {
			gl->supplement_map_layout();
			gcm->call (G, *gl, _grid_bbox, *ap);
			delete gl;
		}
	}

	if (ap) 
		ap->update();
	else
		gl->compact_all_bends();
}

void GridLayoutModule::call (PlaneGraphCopy& PG, face F, LayoutInterface& A)
{
	GridLayout gl;
	_cg_helper (PG, F, gl, &A);
}

void GridLayoutModule::call_grid (PlaneGraphCopy& PG, face F,
	GridLayout &gl)
{
	_cg_helper (PG, F, gl, 0);
}

void GridLayoutModule::_cg_helper (PlaneGraphCopy& PG, face F,
	GridLayout &gl, LayoutInterface *ap)
{
	if (!postcondition(0).member(key::planar_layout)) {
		error_handler(0,"GridLayoutModule::call_grid(PG,F,gl): not a planar layout algorithm.");
		return;
	}

	_init_grid(PG,gl);
	if (!PG.empty()) {
		if (use_copy())
			_call_grid (PG, F, gl);
		else
			_call_grid ((const graph&)PG, F, gl);
	}
	_cleanup_grid(gl);

	if (ap) {
		gcm->call ((const graph&)PG, gl, _grid_bbox, *ap);
		ap->update();
	} else
		gl.compact_all_bends();
}


// further methods of standard interface

void GridLayoutModule::set_coordinate_mapper (const GridCoordinateMapper& new_gcm)
{
	delete gcm;
	gcm = new_gcm.clone();
}

void GridLayoutModule::_init_grid(const graph &G, GridLayout &gl)
{
	gl.init(G);
	_bbox_set = false;
}

void GridLayoutModule::set_grid_bbox(const IRect &bbox)
{
	_grid_bbox = bbox;
	_bbox_set = true;
}

void GridLayoutModule::_cleanup_grid(const GridLayout &gl)
{
	if (!_bbox_set)
		_grid_bbox = gl.area();
}

// standard interface for grid animation

bool GridLayoutModule::start_grid_animation (const graph& G, GridLayout &gl)
{
	gl.init(G);

	grid_step_graph = &G;
	step_gl = &gl;

	if (use_copy()) {

		grid_step_copy_graph = new GraphCopy (G);

		_init_grid(*grid_step_copy_graph, step_gl_2);
		if (_start_grid_animation (*grid_step_copy_graph,step_gl_2) == false)
			return false;
		_cleanup_grid(step_gl_2);
		transform_grid_layout (*grid_step_copy_graph, step_gl_2, gl);

	} else {
		_init_grid(G,gl);
		if (_start_grid_animation (G, gl) == false)
			return false;
		_cleanup_grid(gl);

	}

	return true;
}


void GridLayoutModule::stop_grid_animation ()
{
	if (use_copy()) {

		_stop_grid_animation (*grid_step_copy_graph, step_gl_2);

		delete grid_step_copy_graph;
		step_gl_2.init();

	} else {
		_stop_grid_animation (*grid_step_graph,*step_gl);
	}

}


bool GridLayoutModule::next_grid_step ()
{
	_bbox_set = false;
	if (use_copy()) {
		if (_next_grid_step (*grid_step_copy_graph,	step_gl_2) == false)
			return false;
		_cleanup_grid(step_gl_2);

		transform_grid_layout (*grid_step_copy_graph,
			step_gl_2, *step_gl);

	} else {
		if (_next_grid_step (*grid_step_graph, *step_gl) == false)
			return false;
		_cleanup_grid(*step_gl);
	}

	return true;
}


// implementation of animation using grid animation

bool GridLayoutModule::_start_animation (const graph& G, LayoutInterface& A)
{
	the_gl.init(G);

	if (_start_grid_animation (G, the_gl) == false)
		return false;

	//compute_grid_area (G, the_gl);

	gcm->call (G, the_gl, _grid_bbox, A);

	return true;
}

bool GridLayoutModule::_start_animation (GraphCopy& GC, LayoutInterface& A)
{
	the_gl.init(GC);

	if (_start_grid_animation (GC, the_gl) == false)
		return false;

	//compute_grid_area (GC, the_gl);

	gcm->call (GC, the_gl, _grid_bbox, A);

	return true;
}

bool GridLayoutModule::_next_step (const graph& G, LayoutInterface& A)
{
	bool result = _next_grid_step (G, the_gl);

	//compute_grid_area (G, the_gl);

	gcm->call (G, the_gl, _grid_bbox, A);

	return result;
}

bool GridLayoutModule::_next_step (GraphCopy& GC, LayoutInterface& A)
{
	bool result = _next_grid_step (GC, the_gl);
	
	//compute_grid_area (GC, the_gl);

	gcm->call (GC, the_gl, _grid_bbox, A);

	return result;
}

void GridLayoutModule::_stop_animation (const graph& G, LayoutInterface& A)
{
	_stop_grid_animation (G, the_gl);

	the_gl.init();
}

void GridLayoutModule::_stop_animation (GraphCopy& G, LayoutInterface& A)
{
	_stop_grid_animation (G, the_gl);

	the_gl.init();
}


// transforming grid layouts
/*
void GridLayoutModule::get_bends (const GraphCopy& GC,
	edge e_orig, GridLayout &gl, IPolyline& ip)
{
	edge e, e1 = GC.chain(e_orig).head();

	forall (e, GC.chain(e_orig)) {
		if (e != e1) ip.append(IPoint(gl.x(source(e)),gl.y(source(e))));

		if (!gl.bends(e).empty()) {
			ip.conc (gl.bends(e));
		} else if (GC.is_map()) {
			gl.bends(GC.reversal(e)).reverse_items();
			ip.conc (gl.bends(GC.reversal(e)));
		}
	}
	if (GC.is_reversed(e_orig))
		ip.reverse_items();
}
*/
void GridLayoutModule::transform_grid_layout (const GraphCopy& GC,
	GridLayout &gl, GridLayout &_gl)
{
	const graph& G = GC.original();
	gl.init(G);

	node v;
	forall_nodes (v, G) {
		gl.x(v) = _gl.x(GC.copy(v));
		gl.y(v) = _gl.y(GC.copy(v));
	}

	edge e;
	forall_edges (e, G) {
		gl.bends(e) = _gl.get_compact_bends (GC, e);
	}
}


// default implementations for overridable _call methods

void GridLayoutModule::_call_grid (const graph& /*G*/, GridLayout &)
{
	error_handler(0,"GridLayoutModule::_call_grid(const graph&,GridLayout&) not overridden");
}

void GridLayoutModule::_call_grid (GraphCopy& /*GC*/, GridLayout &)
{
	error_handler(0,"GridLayoutModule::_call_grid(GraphCopy&,GridLayout&) not overridden");
}

bool GridLayoutModule::_start_grid_animation (const graph& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_start_grid_animation(const graph&,GridLayout&) not overridden");
	return false;
}

bool GridLayoutModule::_next_grid_step (const graph& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_next_grid_step(const graph&,GridLayout&) not overridden");
	return false;
}

void GridLayoutModule::_stop_grid_animation (const graph& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_stop_grid_animation(const graph&,GridLayout&) not overridden");
}

bool GridLayoutModule::_start_grid_animation (GraphCopy& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_start_grid_animation(GraphCopy&,GridLayout&) not overridden");
	return false;
}

bool GridLayoutModule::_next_grid_step (GraphCopy& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_next_grid_step(GraphCopy&,GridLayout&) not overridden");
	return false;
}

void GridLayoutModule::_stop_grid_animation (GraphCopy& /*G*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_stop_grid_animation(GraphCopy&,GridLayout&) not overridden");
}


void GridLayoutModule::_call_grid (const graph& /*G*/, face /*F*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_call_grid(const graph&,face,GridLayout&) not overridden");
}

void GridLayoutModule::_call_grid (PlaneGraphCopy& /*PG*/, face /*F*/,
	GridLayout &)
{
	error_handler(0,"GridLayoutModule::_call_grid(PlaneGraphCopy&,face,GridLayout&) not overridden");
}
