//-------------------------------------------------------------------
//
//                         KeySet.cpp
//
//
//    author: Carsten Gutwenger
//    date:   july 1997
//-------------------------------------------------------------------


#include <AGD/KeySet.h>
#include <ctype.h>


KeySet::KeySet()
{
	_mapper = &AGD::keymapper;
	_table_size = _KeySet_min_table_size;
	_alloc_table(_mapper->number_of_keys());
}

KeySet::KeySet(Keymapper& mapper)
{
	_mapper = &mapper;
	_table_size = _KeySet_min_table_size;
	_alloc_table(_mapper->number_of_keys());
}

KeySet::KeySet (const KeySet& S)
{
	_mapper = S._mapper;
	_table_size = S._table_size;
	_table = new int_set(*S._table);
}

KeySet::~KeySet()
{
	delete _table;
}

void KeySet::_alloc_table (int sz)
{
	while (_table_size < sz)
		_table_size <<= 1;

	_table = new int_set(_table_size);
}

void KeySet::_expand_table (int sz)
{
	int_set *old_table = _table;
	int i, old_size = _table_size;

	_alloc_table(sz);
	for (i = 0; i < old_size; i++)
		if (old_table->member(i)) _table->insert(i);

	delete old_table;
}

void KeySet::insert (agd_key k)
{
	int id = k->id();

	if (id >= _table_size) _expand_table(id+1);
	_table->insert(id);
}

void KeySet::del (agd_key k)
{
	if (k->id() < _table_size)
		_table->del(k->id());
}

bool KeySet::member (agd_key k) const
{
	int id = k->id();

	return (id < _table_size) ? (_table->member(id) ? true : false) : false;
}

bool KeySet::empty() const
{
	agd_key k;

	forall (k, *this)
		return true;
	return false;
}


KeySet& KeySet::operator+= (const KeySet& T)
{
	if (_mapper != T._mapper)
		error_handler(1,"KeySet::join: associated keymapper differ.");
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	_table->join(*T._table);

	return *this;
}


KeySet& KeySet::operator-= (const KeySet& T)
{
	if (_mapper != T._mapper)
		error_handler(1,"KeySet::diff: associated keymapper differ.");

	agd_key k;
	forall (k, T)
		del (k);
	return *this;
}


KeySet& KeySet::operator&= (const KeySet& T)
{
	if (_mapper != T._mapper)
		error_handler(1,"KeySet::intersect: associated keymapper differ.");
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	_table->intersect(*T._table);

	return *this;
}


KeySet& KeySet::operator%= (const KeySet& T)
{
	if (_mapper != T._mapper)
		error_handler(1,"KeySet::symmdiff: associated keymapper differ.");
	if (T._table_size > _table_size)
		_expand_table(T._table_size);

	KeySet S1(*this);
	S1 &= T;
	*this += T;
	*this -= S1;

	return *this;
}


bool KeySet::operator<= (const KeySet& T) const
{
	agd_key k;

	forall (k, (*this))
		if (!T.member(k)) return false;
	return true;
}

KeySet& KeySet::operator= (const KeySet& T)
{
	delete _table;
	_table = new int_set(*T._table);
	_table_size = T._table_size;
	_mapper = T._mapper;

	return *this;
}

KeySet::item KeySet::find_next_item (item it) const
{
	for ( ; ; it = _mapper->next_key(it))
		if (it == nil || it->id() >= _table_size)
			break;
		else if (_table->member(it->id()))
			return it;
	return nil;

	//while (it != nil && S.inf(it) == false)
	//	it = S.next_item (it);
}


void KeySet::read (istream& is, int delim)
{
	string str;
	char c;

	clear();

	if (delim == EOF) {
		for (; ;) {
			while (is.get(c) && isspace(c)) ;
			if (!is) break;
			is.putback(c);
			is >> str;
			insert (str);
		}
	} else {
		for (; ;) {
			while (is.get(c) && isspace(c) && c != delim) ;
			if (!is || c == delim) break;
			is.putback (c);
			is >> str;
			insert (str);
		}
	}
}


void KeySet::print (ostream& os, char space) const
{
	agd_key k;
	bool first = true;

	forall (k, (*this)) {
		if (first) first = false; else os << space;
		os << (k->info());
	}
}

