//-------------------------------------------------------------------
//
//                 PlanarStraightDrawLayout.cpp
//
//    author: Carsten Gutwenger
//    date:   august 1997
//-------------------------------------------------------------------


#include <AGD/PlanarDrawLayout.h>
#include <AGD/PlaneGraphCopy.h>
#include <LEDA/b_stack.h>
#include <AGD/PlanAug.h>
#include <AGD/CanonicalOrder.h>


PlanarDrawLayout::PlanarDrawLayout () :
	GridLayoutModule(true)
{
	add_precondition (key::planar);
	add_precondition (key::simple);
	add_precondition (key::no_self_loops);

	add_post_rule (key::planar_layout, 0);
	add_post_rule (key::straight_line);
	add_post_rule (key::no_crossings);

	size_opt (true);
	side_opt (false);
	max_outer (true);
	base_ratio (0.33);

	array<KeySet> PRE(1,1), POST(0,1);
	PRE [1] << key::planar << key::simple << key::no_self_loops
		    << key::directed;
	POST[1] << key::planar << key::biconnected << key::simple
		    << key::no_self_loops;
	augmenter.init (PRE,POST,PlanAug());
}

AGDModule *PlanarDrawLayout::clone() const
{
	PlanarDrawLayout *M = new PlanarDrawLayout;

	M->size_opt (size_opt());
	M->side_opt (side_opt());
	M->max_outer (max_outer());
	M->base_ratio (base_ratio());

	M->set_augmenter (augmenter.get());
	M->set_coordinate_mapper (get_coordinate_mapper());

	return M;
}

void PlanarDrawLayout::_call_grid (PlaneGraphCopy& G, face F, GridLayout &gl)
{
	// handle graphs with less than 3 nodes
	node v1, v2;

	switch (G.number_of_nodes()) {
	case 0:
		return;
	case 1:
		v1 = G.first_node();
		gl.x(v1) = gl.y(v1) = 0;
		return;
	case 2:
		v1 = G.first_node();
		v2 = G.succ_node (v1);
		gl.x(v1) = gl.y(v1) = gl.y(v2) = 0;
		gl.x(v2) = 1;
		return;
	}

	augmenter.get().call (G);

	make_map (G);
	G.make_planar_map();

	CanonicalOrder c_o (G, co_biconnected, G.max_face(), false, base_ratio());
	compute_coordinates (G, c_o, gl.x(), gl.y());
}

void PlanarDrawLayout::compute_coordinates (graph &G, CanonicalOrder &c_o,
	node_map<int>& x, node_map<int>& y)
{
	// Fr Auenflche c_1,...,c_q ist next[c_i] = c_i+1, prev[c_i] = c_i-1
	node_array<node> next(G), prev(G),
	// upper[v] = w bedeutet x-Koord. von v ist relativ zu w (abs. x-Koord.
					 upper(G, nil); // von v = x[v] + abs. x-Koord von w)

	node_array<int>	max_neighbour(G, 0);
	b_stack<node> internals (G.number_of_nodes());

	const OrderSet& V1 = c_o[1];
	node v1 = V1 [1],	// v1 und v2 der Grundlinie
		 v2 = V1 [V1.len()],
		 right_side = v2;

	node v, z1, cl, cr, alpha, beta;
	edge e;
	int	 k, i, sum;
	bool is_outer;

	forall_nodes (v, G)
		forall_out_edges (e, v) {
			int r = c_o.rank (G.target(e));
			if (r > max_neighbour[v]) max_neighbour[v] = r;
		}

	for (i = 1; i <= V1.len(); i++) {
		y [V1[i]] = 0;
		x [V1[i]] = (i == 1) ? 0 : 1;
		if (i < V1.len())
			next[V1[i]] = V1[i+1];
		if (i > 1)
			prev[V1[i]] = V1[i-1];
	}
    prev[v1] = next[v2] = nil;

	for (k = 2; k <= c_o.length(); k++) {
		// Referenz auf aktuelle Menge Vk (als Abkrzung)
		const OrderSet &Vk = c_o[k];
		int l = Vk.len(), offset, eps, x_cr, y_z;    // Vk = { z_1,...,z_l }

		z1 = Vk[1];
		// left- und rightvertex in Menge gespeichert
		cl = Vk.left(); cr = Vk.right();
		if (side_opt() && cr == right_side && max_neighbour[cr] <= k) {
			is_outer = true; right_side = Vk[l];
		} else
			is_outer = false;

		// Berechne relativen x-Abstand con ci zu cl fr i = l+1, ..., r
		for (sum = 0, v = next[cl]; v != cr; v = next[v]) {
			sum += x[v]; x[v] = sum;
		}
		x[cr] += sum;

		eps = (max_neighbour [cl] <= k && k > 2) ? 0 : 1;

		if (size_opt()) {
			int y_max;
			if (is_outer) {
				y_max = Max (y[cl]+1-eps,
							 y[cr] + ((x[cr]==1 && eps==1) ? 1 : 0));
				for (v = next[cl]; v != cr; v = next[v])
					if (x[v] < x[cr]) {
						int y1 = (y[cr]-y[v])*(eps-x[cr])/(x[cr]-x[v])+y[cr];
						if (y1 >= y_max) y_max = 1+y1;
					}
				for (v = cr; v != cl; v = prev[v])
					if (y[prev[v]] > y[v] && max_neighbour[v] >= k) {
						if (y_max <= y[v] + x[v] - eps) {
							eps = 1; y_max = y[v] + x[v];
						}
						break;
					}
				x_cr = Max (x[cr]-eps-l+1, (y[cr]==y_max) ? 1 : 0);
				y_z  = y_max;
			} else {
				// y_max = max { y[c_i] | l <= i <= r }
				y_max = y[cl] - eps;
				for (v = cr; v != cl; v = prev[v])
					if (y[v] > y_max) y_max = y[v];
				offset = Max (y_max-x[cr]+l+eps-y[cr],
							  (y[prev[cr]] > y[cr]) ? 1 : 0);
				y_z  = y[cr] + x[cr] + offset - l + 1 - eps;
				x_cr = y_z - y[cr];
			}
		}
		else {
			y_z  = y[cr] + x[cr] + 1 - eps;
			x_cr = y_z - y[cr];
		}

		alpha = cl;
		for (v = next[cl]; max_neighbour[v] <= k-1 &&
			c_o.rank(v) <= c_o.rank(prev[v]); v = next[v])
		{
			if (c_o.rank (v) < c_o.rank (alpha)) alpha = v;
			if (v == cr) break;
		}
		beta = prev [cr];
		for (v = prev[cr]; max_neighbour[v] <= k-1 &&
			c_o.rank(v) <= c_o.rank(next[v]); v = prev[v])
		{
			if (c_o.rank (v) <= c_o.rank (beta))
				beta = v;
			if (v == cl) break;
		}

		for (i = 1; i <= l; i++) {
			x[Vk[i]] = 1;
			y[Vk[i]] = y_z;
		}
		x[z1] = eps;

		for (v = alpha; v != cl; v = prev [v]) {
			upper [v] = cl;
			internals.push (v);
		}
		for (v = next [beta]; v != cr; v = next [v]) {
			upper [v] = cr; x [v] -= x[cr];
			internals.push (v);
		}
		for (v = beta; v != alpha; v = prev[v]) {
			upper [v] = z1; x [v] -= x[z1];
			internals.push (v);
		}

		x[cr] = x_cr;

		// Aktualisiere Auenregion nach Einfgen von z1,...,zl
		for (i = 1; i <= l; i++) {
			if (i < l) next[Vk[i]] = Vk[i+1];
			if (i > 1) prev[Vk[i]] = Vk[i-1];
		}
		next [cl] = z1;    next [Vk[l]] = cr;
		prev [cr] = Vk[l]; prev [z1] = cl;
	}
	// Berechne die endgltigen x-Koordinaten fr die Knoten der (endgltigen)
	// Auenregion
	for (sum = 0, v = v1; v != nil; v = next[v])
		x [v] = (sum += x[v]);

	// Berechne endgultige x-Koordinaten fr die inneren Knoten
	while (!internals.empty()) {
		v = internals.pop();
		x[v] += x[upper[v]];
	}

}
