/************************************************************************

            Filename    :  TreeLayout.cc

            Version     :  01.1997

            Author      :  Sebastian Leipert

            Language    :  C++

************************************************************************/

#include <AGD/TreeLayout.h>
#include <AGD/LayoutInterface.h>
#include <AGD/impl/TreeLayout/tree.h>
#include <AGD/DPolyline.h>
#include <LEDA/queue.h>

/****************************************************************
                        Constructor
****************************************************************/

TreeLayout::TreeLayout()
    :  LayoutModule()
{
	add_precondition(key::tree);
	add_precondition(key::single_source);

    add_post_rule(key::straight_line);
    add_post_rule(key::no_crossings);
    add_post_rule(key::upward_planar);

    _arrayOfNodes = NULL;
}


AGDModule *TreeLayout::clone() const
{
	TreeLayout *M = new TreeLayout;

	M->sib_separation (sib_separation());
	M->subtree_separation (subtree_separation());
	M->level_separation (level_separation());
	M->left_margin (left_margin());
	M->upper_margin (upper_margin());

	return M;
}

/****************************************************************
                        Destructor
****************************************************************/

TreeLayout::~TreeLayout()
{
}


/****************************************************************
                         name
****************************************************************/

string TreeLayout::name() const
{
   return "TreeLayout";
}


/****************************************************************
                       long_name
****************************************************************/

string TreeLayout::long_name() const
{
   return "Layout algorithm for general trees";
}


/****************************************************************
                         author
****************************************************************/

string TreeLayout::author() const
{
   return "John Q. Walker II";
}


/****************************************************************
                        impl_author
****************************************************************/

string TreeLayout::impl_author() const
{
   return "Sebastian Leipert";
}



/****************************************************************
                        impl_version
****************************************************************/

string TreeLayout::impl_version() const
{
   return "0.1";
}



/****************************************************************
                         impl_date
****************************************************************/

string TreeLayout::impl_date() const
{
   return "may 1997";
}



/****************************************************************
                           _call
****************************************************************/

void TreeLayout::_call(const graph& G, LayoutInterface& a)

{

   int       i  = 0;
   DPolyline _l;   
   double    _x1   = 0.0;
   double    _x2   = 0.0;   
   double    _y1   = 0.0;
   double    _y2   = 0.0;
   double    x_max = 0.0;
   double    y_max = 0.0;

   if (Initialize(G,a))
   {
      _T.positiontree(&x_max,&y_max);

	  a.make_straight_line();
      for (i = 1; i <= G.number_of_nodes(); i++)
      {
	 a.set_x(_arrayOfNodes[i],_T.get_Xcoord(i));
	 a.set_y(_arrayOfNodes[i],_T.get_Ycoord(i));
      }

   }
}


/****************************************************************
                         Initialize
****************************************************************/

bool TreeLayout::Initialize(const graph& G,LayoutInterface &a)

{

   node          _v;
   node          _u;
   edge          _e;
   queue<node>  _Q;
   queue<int>    _QID;
   int           _tree           = 1;
   int           _identification = 1;
   int           _rootDetected   = 0;
   int           _edgeNumber     = 1;
   int           _nodeID         = 0;

   _arrayOfNodes   = new node[G.number_of_nodes()+1];

   if (    has_no_self_loops(G) 
       &&  is_simple(G)
       &&  is_connected(G)
       &&  G.is_directed()
       &&  (G.number_of_nodes() == (G.number_of_edges()+1)))
       
   {
      forall_nodes(_v,G)
      {
	 if (G.indeg(_v) == 0)
	 {
             if (!_rootDetected)
             {
		_root = _v;
	        _rootDetected = 1;
	     }
	     else
	        _tree = 0;
	 }
	 else if (G.indeg(_v) != 1)
	   _tree = 0;
      }	      
   }
   if (_tree)
   {
      _T.clean_tree();
      _T.allocateTree(G.number_of_nodes(),G.number_of_edges());
      _Q.append(_root);
      _arrayOfNodes[_identification] = _root;
      _QID.append(_identification++);
      while (!_Q.empty())
      {
	 _v = _Q.pop();
	 _nodeID = _QID.pop();

	 if (G.outdeg(_v) > 0)
	 {
	    _e = G.first_adj_edge(_v);
	    while (_e != NULL)
	    {
	       _u = G.target(_e);
	       if (!_T.enter_edge(_nodeID,_identification,_edgeNumber))
		  _tree = 0;
	       else
		  _edgeNumber++;
	       _Q.append(_u);
	       _QID.append(_identification);
	       _arrayOfNodes[_identification++] = _u;
	       _e = G.adj_succ(_e);
	    }
	 }
      }
      if (_tree)
      {
	 _T.allocate_node_sizes(_arrayOfNodes,a);
         return 1;
      }
      else
      {
         error_handler(1,"TreeLayout: sorry. No Tree.\n");
         return 0;
      }
      
   }
   else
   {
      error_handler(1,"TreeLayout: sorry. No Tree.\n");
      return 0;
   }
}



/****************************************************************
                        sib_separation
****************************************************************/

void TreeLayout::sib_separation(double value)
{
   _T.set_sib_separation(value);
}


/****************************************************************
                       level_separation
****************************************************************/

void TreeLayout::level_separation(double value)
{
    _T.set_level_separation(value);
}



/****************************************************************
                      subtree_separation
****************************************************************/

void TreeLayout::subtree_separation(double value)
{
   _T.set_subtree_separation(value);
}




/****************************************************************
                       sib_separation
****************************************************************/

double TreeLayout::sib_separation() const
{
   return _T.get_sib_separation();
}


/****************************************************************
                       level_separation
****************************************************************/

double TreeLayout::level_separation() const
{
    return _T.get_level_separation();
}



/****************************************************************
                      subtree_separation
****************************************************************/

double TreeLayout::subtree_separation() const
{
   return _T.get_subtree_separation();
}



/****************************************************************
                    left_margin
****************************************************************/

double TreeLayout::left_margin() const
{
   return _T.get_left_margin();
}


void TreeLayout::left_margin(double s)
{
   _T.set_left_margin(s);
}



/****************************************************************
                    upper_margin
****************************************************************/

double TreeLayout::upper_margin() const
{
   return _T.get_upper_margin();
}


void TreeLayout::upper_margin(double s)
{
   _T.set_upper_margin(s);
}



