//-------------------------------------------------------------------
//
//                         TutteLayout.c
//
//    author: David Alberts
//    date:   April 1997
//-------------------------------------------------------------------


#include <AGD/TutteLayout.h>
#include <AGD/GraphCopy.h>
#include <AGD/LayoutInterface.h>
#include <LEDA/vector.h>
#include <LEDA/matrix.h>
#include <math.h>

TutteLayout::TutteLayout() : LayoutModule()
{
  add_precondition(key::connected);
  add_precondition(key::no_self_loops);

  add_post_rule(key::straight_line);
  add_post_rule(key::planar, in_graph, key::no_crossings);

  // set options to defaults
  the_bbox = DRect (0.0, 250.0, 0.0, 250.0);
}


AGDModule *TutteLayout::clone() const
{
	TutteLayout *M = new TutteLayout;

	M->bbox (bbox());

	return M;
}


void TutteLayout::set_fixed_nodes(const graph& G, list<node>& nodes,
                                  list<DPoint>& pos, double radius)
// sets the positions of the nodes in a largest face of G in the form
// of a regular k-gon. The corresponding nodes and their positions are
// stored in nodes and pos, respectively.
{
  // compute faces of a copy of G
  GraphCopy GC(G);

  make_map(GC);
  PLANAR(GC,true);
  GC.compute_faces();

  // search for largest face
  face max_face = 0;
  int  max_size = -1;
  face f;
  forall_faces(f,GC)
  {
    int current_size = GC.size(f);
    if(current_size > max_size)
    {
      max_size = current_size;
      max_face = f;
    }
  }

  if(!max_face)
    error_handler(999,"TutteLayout::set_fixed_nodes: graph contains no face?!");

  // delete possible old entries in nodes and pos
  nodes.clear();
  pos.clear();

  // set nodes and pos
  node_array<bool> add_me(G,true);
  node v;

  list<node> max_nodes = GC.adj_nodes(max_face);

  forall(v,max_nodes)
  {
    node w = GC.original(v);
    if(add_me[w])
    { 
      nodes.append(w);
      add_me[w] = false;
    }
  }

  double step  = 2.0 * LEDA_PI / double(nodes.length());
  double alpha = 0.0;
  forall(v,nodes)
  {
    pos.append(DPoint(radius * sin(alpha), radius * cos(alpha)));
    alpha += step;
  }
}



void TutteLayout::_call(const graph& G, LayoutInterface& a)
{
  list<node> fixed_nodes;
  list<DPoint> positions;

  a.make_straight_line();

  double diam =
      sqrt((the_bbox.left() - the_bbox.right()) * (the_bbox.left() - the_bbox.right())
         + (the_bbox.bottom() - the_bbox.top()) * (the_bbox.bottom() - the_bbox.top()));

  // handle graphs with less than two nodes
  switch (G.number_of_nodes()) {
  case 0:
	  return;
  case 1:
	  node v = G.first_node();
	  a.set_position (v, the_bbox.center());
	  return;
  }
		
  set_fixed_nodes(G,fixed_nodes,positions,diam/4.2);

  do_tutte(G,a,fixed_nodes,positions);
}



bool TutteLayout::do_tutte(const graph& G, LayoutInterface& a,
              list<node>& fixed_nodes, list<DPoint>& fixed_positions)
// does the actual computation. fixed_nodes and fixed_positions
// contain the nodes with fixed positions.
{
  node v, w;
  edge e;

  // mark fixed nodes and set their positions in a
  node_array<bool> fixed(G,false);
  forall(v,fixed_nodes)
  {
    fixed[v] = true;
    DPoint p = fixed_positions.pop();   // slightly dirty...
    fixed_positions.append(p);          // ...
    a.set_position(v,p);
  }

  if(fixed_nodes.size() == G.number_of_nodes()) return true;
  // all nodes have fixed positions - nothing left to do

  // collect other nodes
  list<node> other_nodes;
  forall_nodes(v,G) if(!fixed[v]) other_nodes.append(v);

  node_array<int> ind(G);       // position of v in other_nodes and A

  int i = 0;
  forall(v,other_nodes) ind[v] = i++;

  int n = other_nodes.size();   // #other nodes
  vector coord(n);              // coordinates (first x then y)
  vector rhs(n);                // right hand side
  matrix A(n,n);                // equations

  // initialize non-zero entries in matrix A
  forall(v,other_nodes)
  {
    double one_over_d = 1.0/double(G.degree(v));
    forall_inout_edges(e,v)
    {
      // get second node of e
      w = (v == source(e)) ? target(e) : source(e);
      if(!fixed[w]) A(ind[v],ind[w]) = one_over_d;
    }
    A(ind[v],ind[v]) = -1;
  }

  if(!A.det()) return false;

  // compute right hand side for x coordinates
  forall(v,other_nodes)
  {
    rhs[ind[v]] = 0;
    double one_over_d = 1.0/double(G.degree(v));
    forall_inout_edges(e,v)
    {
      // get second node of e
      w = (v == source(e)) ? target(e) : source(e);
      if(fixed[w]) rhs[ind[v]] -= (one_over_d*a.get_x(w));
    }
  }

  // compute x coordinates
  coord = A.solve(rhs);
  forall(v,other_nodes) a.set_x(v,coord[ind[v]]);

  // compute right hand side for y coordinates
  forall(v,other_nodes)
  {
    rhs[ind[v]] = 0;
    double one_over_d = 1.0/double(G.degree(v));
    forall_inout_edges(e,v)
    {
      // get second node of e
      w = (v == source(e)) ? target(e) : source(e);
      if(fixed[w]) rhs[ind[v]] -= (one_over_d*a.get_y(w));
    }
  }

  // compute y coordinates
  coord = A.solve(rhs);
  forall(v,other_nodes) a.set_y(v,coord[ind[v]]);

  // translate coordinates, such that the center lies in
  // the center of the bounding box
  DPoint center = the_bbox.center();

  forall_nodes (v, G) {
	a.set_position(v, a.get_position(v)+center);
  }

  return true;
}
