/************************************************************************

         Filename   :   compute_wha.cc

         Version    :   09.1994

         Author     :   Sebastian Leipert

         Language   :   C++


 File containing all procedures for computing the [w,h,a]-numbering
 of all pertinent nodes in the pertinent subtree of a PQ-tree and for 
 deleting pertinent leaves to gain reducabillity of the PQ-tree. 

************************************************************************/ 



#include <AGD/impl/MaxPlanSub/proto.h>

     
extern  int      verticecount;
extern  int      edgecount;
extern  int      total_num_del_edges;
extern  int      node_count;
extern  pq_node  *root;

extern  int      _printDuringReduction;
extern  int      _printPQTreeIterative;
extern  int      _iterationNumber;
extern  char     _debugInformation[INFOSIZE];


/************************************************************************
                          compute1
************************************************************************/ 

void compute1(
              pq_node **pert_root,
              int i,
              int,
              stack<int> *pertinent_leaves,
              leaves *leave_ptr,
              adjlists *deleted_edges
             )


   // Computes the [w,h,a]-number of every pertinent node in the
   // pertinent subtree of the PQ-tree. If the minimum 
   // of the h- and a-number of the root of the pertinent
   // subtree is not 0, then the PQ-tree is not reducible.
   // According to the [w,h,a]-numbering, this procedure removes
   // pertinent leaves out of the PQ-tree to gain reducabillity.
   // Furthermore, the number of pertinent leaves in the frontier of
   // every pertinent node is computed. With that, it is possible to
   // detect the root of the pertinent subtree.

{
   pq_node    *node_ptr;
   int        helpnode;
   int        del_edges = 0;   // number of deleted pertinent leaves
   int        max_pert_leaf_count = pertinent_leaves->count();
                              // number of all pertinent leaves
                              
   queue<pq_node*> *Q = new queue<pq_node*>;
                              // Queue to store the nodes, of which the 
                              // [w,h,a]-number has to be computed next.
                              // A node only gets into Q, if for all of 
                              // its children the [w,h,a]-number has been
                              // computed.
                              
   stack<pq_node*> *compute_stack = new stack<pq_node*>;
                              // Stack containing all nodes where the
                              // [w,h,a]-number has been computed.
                              // Is needed to remove nodes out of tree.
                              
   pertinent_leaves->start_at_bottom();
   while (!pertinent_leaves->read_last())
                              // Place all pertinent leaves on to Q and
                              // compute_stack.
   {
      helpnode = pertinent_leaves->read_next();
      node_ptr = leave_ptr->get_ptr(helpnode,i);
      node_ptr->not_visited_child_count--;
      node_ptr->pert_leaf_count = 1;
      Q->enqueue(node_ptr);
      compute_stack->push(node_ptr);
   }// while
   
   while (!Q->queue_empty())
                               // Remove the first node out of Q and
                               // compute its [w,h,a]-number.
   {
      node_ptr = Q->dequeue();
      if (node_ptr->pert_leaf_count < max_pert_leaf_count)
                               // node_ptr is not the root.
      {
         node_ptr->parent->pert_leaf_count =
           node_ptr->parent->pert_leaf_count + node_ptr->pert_leaf_count;
         node_ptr->parent->not_visited_child_count--;
         if (!node_ptr->parent->not_visited_child_count)
                               // All children of node_ptr have their
                               // [w,h,a]-number, or have been placed on to
                               // Q. Place node_ptr on to Q and compute_stack.
         {
            Q->enqueue(node_ptr->parent);
            compute_stack->push(node_ptr->parent);
         }// if
      }// if

      // Compute the [w,h,a]-number of node_ptr.
      
      if (node_ptr->typ == LEAF)
                               // node_ptr is a leaf.
      {
         node_ptr->label = FULL;
         node_ptr->w_nr = 1;
         node_ptr->h_nr = 0;
         node_ptr->a_nr = 0;
         if (node_ptr->pert_leaf_count < max_pert_leaf_count)
            node_ptr->parent->full_children->push(node_ptr);        
      }// if

      else

                                // node_ptr is P- or Q-node.

      {
                                // Compute the w-number.
                                // The w-number is computed for P- and
                                // Q-nodes identically.      
         node_ptr->w_nr = sum_pert_child(node_ptr);
                                
         if (node_ptr->full_children->count() == node_ptr->child_count)

                                // node_ptr is a full node.
         
         {
            node_ptr->label = FULL;
            if (node_ptr->pert_leaf_count < max_pert_leaf_count)
               node_ptr->parent->full_children->push(node_ptr);
            node_ptr->h_nr = 0;
            node_ptr->a_nr = 0;
         }// if 

         else

                               // node_ptr is a partial node.

         {
            node_ptr->label = PARTIAL;
            if (node_ptr->pert_leaf_count < max_pert_leaf_count)
               node_ptr->parent->partial_children->push(node_ptr);

            if (node_ptr->typ == P_NODE)
                               // node_ptr is a partial P-node.
                               // Compute h- and a-number of node_ptr.
               h_a_num_p_node(node_ptr);

            else
                               // node_ptr is a partial Q-node.
                               // Compute h- and a-number of node_ptr.
               h_a_num_q_node(node_ptr);

         }// else
      }// else
   }// while
                               // Set the root of the pertinent subtree. 
   (*pert_root) = node_ptr; 
   if ((*pert_root)->h_nr < (*pert_root)->a_nr)
                               // Find the minimum of the h- and a-number
                               // of the root of the pertinent subtree.
      del_edges = (*pert_root)->h_nr;
   else
      del_edges = (*pert_root)->a_nr;
                              
   if (del_edges > 0)
                               // If the minimum is <> 0, then pertinent
                               // leaves have to be removed out of the tree. 
   {
      if ((*pert_root)->h_nr < (*pert_root)->a_nr)
         (*pert_root)->del_typ = H;
      else
         (*pert_root)->del_typ = A;
   }// if
   delete_leaves(*pert_root,compute_stack,i,leave_ptr,deleted_edges);
   
   delete Q;
   delete compute_stack;
}// compute1         
   


/************************************************************************
                          delete_leaves
************************************************************************/ 

void delete_leaves(
                   pq_node *,
                   stack<pq_node*> *compute_stack,
                   int i,
                   leaves *leave_ptr,
                   adjlists *deleted_edges
                  )

   // If the minimum number of the h- and a-numbers of the pertinent
   // root is not 0, some of the pertinent leaves have to be deleted.
   // This is done by delete_leaves using the [w,h,a]-numbering
   // and the compute_stack where every node is stored above its 
   // pertinent children.
{
   pq_node   *node_ptr;
   int       child_count;

   while (!compute_stack->stack_empty())
                               // Get the next node.
   {
      child_count = 0;
      node_ptr = compute_stack->pop();
      if (node_ptr->label == FULL && (node_ptr->del_typ == H
                                      || node_ptr->del_typ == A))
                               // A full node cannot be made type H or A,
                               // so it must be type B.
         node_ptr->del_typ = B;


      if (node_ptr->typ == LEAF)
                               // Check whether the leaf remains in the
                               // PQ-tree or not.
      {
         if (node_ptr->del_typ == W)
                               // If the leaf may not remain in the
                               // PQ-tree delete it. If every child
                               // of a node is going to be deleted
                               // the node will be deleted as well.
            kill_leaf(node_ptr,i,leave_ptr,deleted_edges);
      }// if

      
      else
                               // node_ptr is a Q- or a P-node.
      
      switch (node_ptr->del_typ)
      
      {
         case B:
                               // Nothing to do here.
                 break;

         case W:
                               // If a node is marked W, all pertinent
                               // children are marked W.
                 set_pert_child(node_ptr,PERTINENT,W);             
                 node_ptr->pert_child_count = 0;
                 break;
                 
         case H:
                              // Case whether node_ptr is a P-
                              // or a Q-node. 
                 if (node_ptr->typ == P_NODE)
                              // node_ptr is P-node marked H.
                              // Mark all partial children W.
                 {
                    set_pert_child(node_ptr,PARTIAL,W);
                              // Mark all full children B.
                    set_pert_child(node_ptr,FULL,B);
                              // Mark the h_child1 H, if it exists.
                    if (node_ptr->h_child1 != NULL)
                    {
                       node_ptr->h_child1->del_typ = H;
                       if (node_ptr->h_child1->h_nr <
                           node_ptr->h_child1->w_nr)
                          child_count = 1;
                    }// if
                    node_ptr->pert_child_count = node_ptr->pert_child_count
                         + child_count - node_ptr->partial_children->count();
                 }// if
                 
                 else
                              // node_ptr is a Q-node marked H.
                              // Mark all pertinent children W.
                 {
                    set_pert_child(node_ptr,PERTINENT,W);
                              // Mark all full children in the
                              // consecutive sequence of h_child1 B,
                              // except the one possible partial child
                              // which is marked H, if it exists.
                    node_ptr->pert_child_count=set_h_child(node_ptr->h_child1);
                 }// else
                 break;

                 
         case A:
                              // Case whether node_ptr is a P-
                              // or a Q-node.

            if (node_ptr->typ == P_NODE)
                              // node_ptr is a P-node marked A.
                 {
                    if (node_ptr->a_child != NULL)
                              // There exists an a_child.
                              // Mark this child A and all
                              // other children W.
                    {
                       set_pert_child(node_ptr,PERTINENT,W);
                       node_ptr->a_child->del_typ = A;
                       node_ptr->pert_child_count = 1;
                    }// if
                    
                    else
                              // There exists a consecutive sequence
                              // of possibly full children and possibly
                              // up to two partial children at the
                              // end of the sequence.
                              // Mark the full children B, if there
                              // exist any, and mark the partial children
                              // H, if there exist any. Notice that all
                              // full children are included into
                              // the consecutive sequence, since the
                              // father is a P-node.
                    {
                       set_pert_child(node_ptr,FULL,B);
                              // Mark all partial children W.
                       set_pert_child(node_ptr,PARTIAL,W);
                       if (node_ptr->h_child1 != NULL)
                              // If h_child1 exists, mark it H.
                       {
                          node_ptr->h_child1->del_typ = H;
                          if (node_ptr->h_child1->h_nr <
                                node_ptr->h_child1->w_nr)
                             child_count = 1;
                       }// if
                       if (node_ptr->h_child2 != NULL)
                              // If h_child2 exists, mark it H.
                       {
                          node_ptr->h_child2->del_typ = H;
                          if (node_ptr->h_child2->h_nr <
                                node_ptr->h_child2->w_nr)
                             child_count++;
                       }// if
                       node_ptr->pert_child_count=node_ptr->pert_child_count
                             - node_ptr->partial_children->count()
                             + child_count;
                    }// else
                 }// if
                 
            else
                              // node_ptr is a Q-node marked A.
                 
                 {
                    if (node_ptr->a_child != NULL)
                              // There exist a_child.
                              // Mark this child A and all other
                              // children W.
                    {
                       set_pert_child(node_ptr,PERTINENT,W);
                       node_ptr->a_child->del_typ = A;
                       node_ptr->pert_child_count = 1;
                    }// if
                    
                    else
                              // There exists a consecutive sequence
                              // of possibly full children and possibly
                              // up to two partial children at the
                              // end of the sequence.
                              // Mark the full children B, if there
                              // exist any, and mark the partial children
                              // H, if there exist any.
                              // First, mark all pertinent children W.
                    {
                       set_pert_child(node_ptr,PERTINENT,W);
                              // Find the consecutive sequence
                              // and mark the nodes as discussed.
                       node_ptr->pert_child_count =
                             set_a_children(node_ptr->h_child2);
                    }// else
                 }// else
                 break;
      }// switch

                              //  Cleanup the nodes
      node_ptr->pert_leaf_count = 0;
      node_ptr->full_children->empty_stack();
      node_ptr->partial_children->empty_stack();
      if (node_ptr->label != TO_BE_DELETED)
         node_ptr->label = EMPTY;
      node_ptr->h_child1 = NULL;
      node_ptr->h_child2 = NULL;
      node_ptr->a_child = NULL;
      node_ptr->w_nr = 0;
      node_ptr->h_nr = 0;
      node_ptr->a_nr = 0;
      node_ptr->del_typ = B;
   }// while
}// delete_leaves
                


/************************************************************************
                          kill_leaf
************************************************************************/ 

void kill_leaf(
               pq_node *node_ptr,
               int i,
               leaves *leave_ptr,
               adjlists *deleted_edges
              )

   // Removes a leaf from the PQ-Tree by updating the immediate
   // siblings and its parent and by actually deleting the node
   // representing the leaf.
   // If the remaining parent contains only one child, the
   // parent is removed as well, and the child is connected
   // to its grandparent, if there is one. By this way, all
   // nodes that are full and are marked W are actually removed
   // from the PQ-tree.
{
   pq_node  *sibling = NULL;
   int      node;
   int      neighbor;
   int      leafnr;

   neighbor = node_ptr->key;
   leafnr = node_ptr->leafnum;
   deleted_edges->insert_edge(neighbor,leafnr);
                            // Find an immediate sibling of node_ptr.
   if (node_ptr->parent->typ == P_NODE)
      sibling = node_ptr->sib_next;
   else
      if (!node_ptr->imm_siblings->stack_empty())
      {
         node_ptr->imm_siblings->start_at_bottom();
         sibling = node_ptr->imm_siblings->read_next();
      }// if

                            // Remove node_prt from siblings.
   if (node_ptr->parent->typ == P_NODE)
      remove_P_imm_sib(node_ptr);
   else
      remove_Q_imm_sib(node_ptr);
   node_ptr->parent->child_count--;

                            // Check if parent has only one son left
                            // (thats where we need the sibling, for
                            // then we have to connect it with
                            // its grandparent).
   check_if_one_son(node_ptr->parent,sibling);

                            // Find the pointer in the leave_ptr-list.
   node = node_ptr->key;
   leave_ptr->remove_ptr(node,i);
   
   destroy(node_ptr);       // Delete the node containing the leaf.

   total_num_del_edges++;
}// kill_leaf

   

/************************************************************************
                          set_h_child
************************************************************************/ 
   
int set_h_child(
                pq_node *h_child1
               )

   // Marks all full children of a Q-node in a consecutive sequence,
   // on the one side where the h_child1 points to, B.
   // If there is a partial child at the
   // end of the consecutive sequence, then it is marked H. The procedure
   // counts for the pertinent children of the Q-node at the same time.
{
   int        count = 0;     // counts the pertinent children
   int        full_label = FALSE; // true, as long as children with a
                                  // full label are found
   pq_node    *aktual_ptr = h_child1;
   pq_node    *new_sibling = NULL;
   pq_node    *old_sibling = NULL;
   
   if (h_child1 != NULL)
                             // There exists a consecutive sequence
                             // of pertinent children at one side of
                             // the Q-node.
      full_label = TRUE;
      
   while (full_label)
                             // The end of the consecutive sequence is 
                             // not found yet.
   {
      aktual_ptr->imm_siblings->start_at_bottom();
                             // Get next sibling.
      new_sibling = aktual_ptr->imm_siblings->read_next();
      if (new_sibling == old_sibling)
      {
         if (!aktual_ptr->imm_siblings->read_last())
            new_sibling = aktual_ptr->imm_siblings->read_next();
         else
                             // Second end of Q-node is reached.
            full_label = FALSE;
      }// if
                             // Check the label of the next sibling.
      if (aktual_ptr->label == FULL)
      {
         aktual_ptr->del_typ = B;
         count++;
      }// if
      else
                            // End of the consecutive sequence reached.
      {
         if (aktual_ptr->label == PARTIAL)
         {
             aktual_ptr->del_typ = H;
             if ((aktual_ptr->pert_leaf_count - aktual_ptr->h_nr) > 0)
                count++;
         }// if
         full_label = FALSE;
      }// else
      old_sibling = aktual_ptr;
      aktual_ptr = new_sibling;
   }// while
   return count;
}// set_h_child
   
   
   
/************************************************************************
                          set_a_children
************************************************************************/ 

int set_a_children(
                   pq_node *h_child2
                  )

   // Marks all full children of a Q-node in a consecutive sequence,
   // starting with h_child2, B. If there are partial children at the
   // ends of the consecutive sequence, then they are marked H.
   // The procedure counts for the pertinent children at the same time.
{
   int        pert_child = 0;
                             // counts the pertinent children
   int        reached_end = FALSE;
                             
                             // Observe that it is not clear, which
                             // direction we have to follow, to find
                             // the consecutive sequence starting at
                             // h_child2. So we have to trace both sides
                             // to find the maximal consecutice sequence.

   int        left_sum = 0;  // sum of pertinent leaves in the front
                             // on the left side of h_child2
   int        right_sum = 0; // sum of pertinent leaves in the front
                             // on the right side of h_child2
   pq_node    *aktual_ptr = h_child2;
   pq_node    *left_new_sib = NULL;
                             // pointer for tracing the left side
   pq_node    *left_old_sib = NULL;
                             // pointer keeping the last visited child
                             // on the left side in mind
   pq_node    *right_new_sib = NULL;
                             // pointer for tracing the right side
   pq_node    *right_old_sib = NULL;
                             // pointer for keeping the last visited
                             // child on the right side in mind
   

   h_child2->imm_siblings->start_at_bottom();
                             // Set del_typ of h_child2.
   if (h_child2->label == FULL)
      h_child2->del_typ = B;
   else if (h_child2->label == PARTIAL)
      h_child2->del_typ = H;
   else
      printf("set_a_children: 1. node of sequence is EMPTY\n");
   left_sum = right_sum = aktual_ptr->w_nr - aktual_ptr->h_nr;
   if (left_sum > 0)
      pert_child++;
      

   if (!h_child2->imm_siblings->read_last())
                             // Trace the left side of h_child2.
   {
      left_new_sib = h_child2->imm_siblings->read_next();
      left_old_sib = h_child2;
      aktual_ptr = left_new_sib;
      reached_end = FALSE;

      while(!reached_end)
      {
         if (aktual_ptr->label == FULL)
            left_sum = left_sum + aktual_ptr->w_nr - aktual_ptr->h_nr;
         else
                              // Reached the end of the left side
                              // of the consecutive sequence.
         {
            if (aktual_ptr->label == PARTIAL)
               left_sum = left_sum + aktual_ptr->w_nr - aktual_ptr->h_nr;
            reached_end = TRUE;
         }// else
         if (!reached_end)
                              // There are still siblings that must be
                              // visited.
         {
            aktual_ptr->imm_siblings->start_at_bottom();
            left_new_sib = aktual_ptr->imm_siblings->read_next();
            if (left_new_sib == left_old_sib)
            {
               if (!aktual_ptr->imm_siblings->read_last())
                  left_new_sib = aktual_ptr->imm_siblings->read_next();
               else
                              // The end of the Q-node is reached.
                  reached_end = TRUE;
            }// if
            left_old_sib = aktual_ptr;
            aktual_ptr = left_new_sib;
         }// if
      }// while
   }// if

   
   if (!h_child2->imm_siblings->read_last())
                              // Trace the right side of h_child2.
   {
      right_new_sib = h_child2->imm_siblings->read_next();
      right_old_sib = h_child2;
      aktual_ptr = right_new_sib;
      reached_end = FALSE;
      while(!reached_end)

      {
         if (aktual_ptr->label == FULL)
            right_sum = right_sum + aktual_ptr->w_nr - aktual_ptr->h_nr;
         else
                              // Reached the end of the right side of
                              // the consecutive sequence.
         {
            if (aktual_ptr->label == PARTIAL)
               right_sum = right_sum + aktual_ptr->w_nr - aktual_ptr->h_nr;
            reached_end = TRUE;
         }// else
         if (!reached_end)
                              // There are still siblings that must be
                              // visited. 
         {
            aktual_ptr->imm_siblings->start_at_bottom();
            right_new_sib = aktual_ptr->imm_siblings->read_next();
            if (right_new_sib == right_old_sib)
            {
               if (!aktual_ptr->imm_siblings->read_last())
                  right_new_sib = aktual_ptr->imm_siblings->read_next();
               else
                              // The end of the Q-node is reached.
                  reached_end = TRUE;
            }// if
            right_old_sib = aktual_ptr;
            aktual_ptr = right_new_sib;
         }// if
         
      }// while
   }// if

   
   h_child2->imm_siblings->start_at_bottom();
   if (left_sum >= right_sum)
                              // Left side has more leaves in the front
                              // of the cosecutice sequence. Mark all
                              // full children in the sequence B and the
                              // partial child H, if there exist one.
   {
      if (!h_child2->imm_siblings->read_last())
      {
         left_new_sib = h_child2->imm_siblings->read_next();
         left_old_sib = h_child2;
         aktual_ptr = left_new_sib;
         reached_end = FALSE;
         while (!reached_end)
         {
            if (aktual_ptr->label == FULL)
            {
               aktual_ptr->del_typ = B;
               pert_child++;
            }// if
            else
            {
               if (aktual_ptr->label == PARTIAL)
               {
                  aktual_ptr->del_typ = H;
                  if ((aktual_ptr->w_nr - aktual_ptr->h_nr) > 0)
                     pert_child++;
               }// if
               reached_end = TRUE;
            }// else
            if (!reached_end)
            {
               aktual_ptr->imm_siblings->start_at_bottom();
               left_new_sib = aktual_ptr->imm_siblings->read_next();
               if (left_new_sib == left_old_sib)
               {
                  if (!aktual_ptr->imm_siblings->read_last())
                     left_new_sib = aktual_ptr->imm_siblings->read_next();
                  else
                              // End of the Q-node is reached.
                     reached_end = TRUE;
               }// if
               left_old_sib = aktual_ptr;
               aktual_ptr = left_new_sib;
            }// if
         }// while
      }// if
   }// if

   
   else
                              // Right side has more leaves in the front
                              // of the consecutive sequence. Mark all
                              // full children in the sequence B and the
                              // partial child H, if there exists one.
   {
      if (!h_child2->imm_siblings->read_last())
         h_child2->imm_siblings->read_next();
      if(!h_child2->imm_siblings->read_last())
      {
         right_new_sib = h_child2->imm_siblings->read_next();
         right_old_sib = h_child2;
         aktual_ptr = right_new_sib;
         reached_end = FALSE;
         while(!reached_end)
         {
            if (aktual_ptr->label == FULL)
            {
               aktual_ptr->del_typ = B;
               pert_child++;
            }// if
            else
            {
               if (aktual_ptr->label == PARTIAL)
               {
                  aktual_ptr->del_typ = H;
                  if ((aktual_ptr->w_nr - aktual_ptr->h_nr) > 0)
                     pert_child++;
               }// if
               reached_end = TRUE;
            }// else
            if (!reached_end)
            {
               aktual_ptr->imm_siblings->start_at_bottom();
               right_new_sib = aktual_ptr->imm_siblings->read_next();
               if (right_new_sib == right_old_sib)
               {
                  if (!aktual_ptr->imm_siblings->read_last())
                     right_new_sib = aktual_ptr->imm_siblings->read_next();
                  else
                              // End of the Q-node is reached.
                     reached_end = TRUE;
               }// if
               right_old_sib = aktual_ptr;
               aktual_ptr = right_new_sib;
            }// if
         }// while
      }// if
   }// else

   return pert_child;   
}//set_a_child

   
 
/************************************************************************
                          set_pert_child
************************************************************************/ 
            
void set_pert_child(
                    pq_node *node_ptr,
                    int label,
                    int del_typ
                   )

   // Sets all full and/or partial children of node_ptr to
   // del_typ B, W, H or A.
{
   pq_node  *helpptr;
  
   if (label == PERTINENT)
                              // Set all full and partial
                              // children to del_typ.                        
   {
      node_ptr->partial_children->start_at_bottom();
      while (!node_ptr->partial_children->read_last())
      {
         helpptr = node_ptr->partial_children->read_next();
         helpptr->del_typ = del_typ;
      }// while
      node_ptr->full_children->start_at_bottom();
      while (!node_ptr->full_children->read_last())
      {
         helpptr = node_ptr->full_children->read_next();
         helpptr->del_typ = del_typ;
      }// while
   }// if
   
   else if (label == PARTIAL)
                              // Set all partial children to del_typ.
   {  node_ptr->partial_children->start_at_bottom();
      while (!node_ptr->partial_children->read_last())
      {
         helpptr = node_ptr->partial_children->read_next();
         helpptr->del_typ = del_typ;
      }// while
   }// else if
   
   else
                              // Set all full children to del_typ.
   {
      node_ptr->full_children->start_at_bottom();
      while (!node_ptr->full_children->read_last())
      {
         helpptr = node_ptr->full_children->read_next();
         helpptr->del_typ = del_typ;
      }// while
   }// else
}// set_pert_child
     
        

/************************************************************************
                          h_a_num_p_node
************************************************************************/ 

void h_a_num_p_node(
                    pq_node *node_ptr
                   )

   // Compute the h- and a-number of node_ptr, node_ptr is a P-node.
{
   int      sum_par_w = 0;       // = sum w_i, i partial child
   int      sum_max1 = 0;        // = max_1(w_i - h_i), i partial child
   int      sum_max2 = 0;        // = max_2(w_i - h_i), i partial child
   int      sum_help = 0;
   pq_node  *helpptr;
   pq_node  *h_child1 = NULL;
   pq_node  *h_child2 = NULL;
   pq_node  *a_child = NULL;
  
// Compute the h-number.

   node_ptr->partial_children->start_at_bottom();
   while (!node_ptr->partial_children->read_last())
                                // Compute of all partial children i
                                // of node_ptr the 1. and the 2. maximum
                                // of (w_i - h_i).
   {
      helpptr = node_ptr->partial_children->read_next();
      sum_par_w = sum_par_w + helpptr->w_nr;
      sum_help = helpptr->w_nr - helpptr->h_nr;
      if (sum_max1 <= sum_help)
      {
         sum_max2 = sum_max1;
         h_child2 = h_child1;
         sum_max1 = sum_help;
         h_child1 = helpptr;
      }// if
      else if (sum_max2 <= sum_help)
      {
         sum_max2 = sum_help;
         h_child2 = helpptr;
      }// else if 
       
   }// while
   node_ptr->h_child1 = h_child1;
   node_ptr->h_child2 = h_child2;
   node_ptr->h_nr = sum_par_w - sum_max1;
                                 // h-number was computed, h_child1 and
                                 // h_child2 were set.

// Compute the a-number.

   int  a2 = sum_par_w - sum_max1 - sum_max2;
   int  a1 = a1_b1_number(node_ptr,&a_child);
                                 // a1 and a2 have been computed.
                                
   if (a1 <= a2)
                                 // Find the minimum of a1 and a2 for the
                                 // a-number. Set a_child. 
   {
       node_ptr->a_nr = a1;
       node_ptr->a_child = a_child;
   }// if
   else
   {
      node_ptr->a_nr = a2;
      node_ptr->a_child = NULL;
   }// else
}// h_a_num_p_node
  
  

/************************************************************************
                          h_a_num_q_node
************************************************************************/ 

void h_a_num_q_node(
                    pq_node *node_ptr
                   )

   // Finds the h- and a-number of a partial Q-node.

{
   int       sum_all_w = 0;       // = sum w_i , i pertinent node
   int       sum_left = 0;        // = sum (w_i - h_i) at left end
   int       sum_right = 0;       // = sum (w_i - h_i) at right end
   int       full_label;
   pq_node   *a_child = NULL;
   pq_node   *left_child = NULL;
   pq_node   *right_child = NULL;
   pq_node   *old_sibling = NULL;
   pq_node   *new_sibling = NULL;
   
// Compute the h-number.

   sum_all_w = sum_pert_child(node_ptr);
   
   node_ptr->endmost_children->start_at_bottom();
                                 // Find the two endmost_children of the
                                 // Q-node.
                                 
   if (!node_ptr->endmost_children->read_last())
                                 // This is the left side of the Q-node.
      left_child = node_ptr->endmost_children->read_next();
      
   if (!node_ptr->endmost_children->read_last())
                                 // This is the right side of the Q-node.
      right_child = node_ptr->endmost_children->read_next();
      
   if (!node_ptr->endmost_children->read_last() || left_child == NULL
       || right_child == NULL)
      printf("h_a_num_q_node: Error in endmost_children, node %d",
              node_ptr->key);
                             
   full_label = TRUE;
   old_sibling = NULL;          
   new_sibling = NULL;           
                                 // Check left side of the Q-node
                                 // for the maximum consecutive sequence
                                 // of full nodes with only one partial
                                 // node at the end.
   while (full_label)            
   {
      if (left_child->label != FULL)
                                 // End of the consecutive sequence found.
         full_label = FALSE;
      if (left_child->label != EMPTY)
                                 // Get new sibling.
      {
         sum_left = sum_left + left_child->w_nr - left_child->h_nr;
         left_child->imm_siblings->start_at_bottom();
         new_sibling = left_child->imm_siblings->read_next();
         if (new_sibling == old_sibling)
         {
            if (!left_child->imm_siblings->read_last())
               new_sibling = left_child->imm_siblings->read_next();

            else
                                 // other end of Q-node reached
               full_label = FALSE;
         }// if
         old_sibling = left_child;
         left_child = new_sibling;
      }// if
   }// while
                            
   old_sibling = NULL;           
   new_sibling = NULL;           
   full_label = TRUE;            
                                 // Check the right side of the Q-node
                                 // for the maximum consecutive sequence
                                 // of full nodes with only one partial
                                 // node at the end.
   while (full_label)            
   {
      if (right_child->label != FULL)
                                 // End of the consecutive sequence found.
         full_label = FALSE;
      if (right_child->label != EMPTY)
                                 // Get next sibling.
      {
         sum_right = sum_right + right_child->w_nr - right_child->h_nr;
         right_child->imm_siblings->start_at_bottom();
         new_sibling = right_child->imm_siblings->read_next();
         if (new_sibling == old_sibling)
         {
            if (!right_child->imm_siblings->read_last())
               new_sibling = right_child->imm_siblings->read_next();
            else
               full_label = FALSE;
         }// if
         old_sibling = right_child;
         right_child = new_sibling;
      }// if
   }// while

                                 // Find the maximum of the
                                 // two consecutive sequences at the
                                 // both ends of the Q-node.
   node_ptr->endmost_children->start_at_bottom();
   if (!node_ptr->endmost_children->read_last())
      left_child = node_ptr->endmost_children->read_next();
   if (!node_ptr->endmost_children->read_last())
      right_child = node_ptr->endmost_children->read_next();
   if (sum_left == 0 && sum_right == 0)
   {
      node_ptr->h_nr = sum_all_w;
      node_ptr->h_child1 = NULL;
   }// if
   else if (sum_left < sum_right)
   {
      node_ptr->h_nr = sum_all_w - sum_right;
      node_ptr->h_child1 = right_child;
   }// else
   else
   {
      node_ptr->h_nr = sum_all_w - sum_left;
      node_ptr->h_child1= left_child;
   }// else

      
// Compute the a- number.

                                  // We now check for consecutive
                                  // sequences between all children
                                  // of the Q-node. Therefore the children
                                  // which form a consecutive sequence are
                                  // stored in a stack, which is emptied
                                  // as soon as the end of the sequence is
                                  // reached. When the stack is emptied, we
                                  // count for the pertinent leaves in the
                                  // front of the sequence, and update
                                  // the sequence with the maximum number
                                  // of pertinent leaves.
                                  // Observe that when the sequence ends
                                  // with a partial node, we have to push
                                  // this partial node back into the stack,
                                  // after it is emptied, since it may form
                                  // a consecutive sequence with its other
                                  // siblings.

   int             b1 = a1_b1_number(node_ptr,&a_child);
   int             b2 = 0;
   int             a_sum = 0;     // number of pertinent leaves of
                                  // the actual visited sequence 
   int             old_a_sum = 0; // number of pertinent leaves of the
                                  // actual maximum sequence
   int             reached_end;   // true, when reached the end of the Q-node
   //int             seq_found;     // true when a new consecutive sequence is
                                  // found, false as soon as an empty child
                                  // is found
   pq_node         *left_most = NULL;
   pq_node         *right_most = NULL;
                                  // pointers to keep the both ends of the
                                  // actual visited sequence in mind
   pq_node         *old_left_most = NULL;
   pq_node         *old_right_most = NULL;
                                  // pointers to keep the both ends of the
                                  // actual maximum sequence in mind
   pq_node         *aktual_ptr = NULL;
   pq_node         *helpptr;
   pq_node         *last_child = NULL;
                                  // pointer to the second endmost child

   stack<pq_node*> sequence;      // stack to store the actual visited
                                  // consecutive sequence
   
                             
   old_sibling = NULL;
   new_sibling = NULL;
   node_ptr->endmost_children->start_at_bottom();
   aktual_ptr = node_ptr->endmost_children->read_next();
   last_child = node_ptr->endmost_children->read_next();
                                  // Start at one side of the Q-node.
   reached_end = FALSE;
   while (!reached_end)
                                  // Check all children of the Q-node.          
   {
      if (sequence.stack_empty())
                                  // No actual consecutive sequence. 
      {
         if (aktual_ptr->label != EMPTY)
                                  // New consecutive sequence found.
            sequence.push(aktual_ptr);
      }// if
      else
                                  // Actual consecutive sequence.
      {
         if (aktual_ptr->label == FULL)
                                  // Full node of the actual sequence.
            sequence.push(aktual_ptr);
            
         else if (aktual_ptr->label == EMPTY)
                                  // End of the actual consecutive sequence
                                  // found. Pop children out of stack
                                  // and sum up the a-numbers of the sequence. 
         {
            a_sum = 0;
            right_most = old_sibling;
            while (!sequence.stack_empty())
            {
               helpptr = sequence.pop();
               a_sum = a_sum + helpptr->w_nr - helpptr->h_nr;
            }// while
            left_most = helpptr;
            if (old_a_sum < a_sum)
                                  // Update the maximum sequence.
            {
               old_a_sum = a_sum;
               old_left_most = left_most;
               old_right_most = right_most;
            }// if
         }// else if
         
         else
                                  // Actual child is partial.
                                  // End of the actual consecutive sequence
                                  // found. Pop children out of stack
                                  // and sum up the a-numbers of the sequence.
                                  // The last child was partial, so push it
                                  // back into the stack, since it may form
                                  // a sequence with some other children.
         {
            sequence.push(aktual_ptr);
                                  //  Push last child of the sequence
                                  //  into stack.
            a_sum = 0;
            right_most = aktual_ptr;
            while (!sequence.stack_empty())
            {
               helpptr = sequence.pop();
               a_sum = a_sum + helpptr->w_nr - helpptr->h_nr;
            }// while
            left_most = helpptr;
            if (old_a_sum < a_sum)
                                  // Update the maximum sequence.
            {
               old_a_sum = a_sum;
               old_left_most = left_most;
               old_right_most = right_most;
            }// if
            sequence.push(aktual_ptr);
                                  // Push partial child back into stack.
         }// else
      }// else


      if (aktual_ptr != last_child)
                                  // Get next sibling.
      {
         aktual_ptr->imm_siblings->start_at_bottom();
         new_sibling = aktual_ptr->imm_siblings->read_next();
         if (new_sibling == old_sibling)
            new_sibling = aktual_ptr->imm_siblings->read_next();
         old_sibling = aktual_ptr;
         aktual_ptr = new_sibling;
      }// if
      else
                                  // End of Q-node reached.
         reached_end = TRUE;
   }// while

                                  // If the last visited child of the Q-node
                                  // was part of a consecutive sequence,
                                  // then the stack, that stores this
                                  // sequence, was not emptied, and the last
                                  // child was not put into the stack.This is
                                  // done here.
   if (!sequence.stack_empty())
   {
      helpptr = sequence.pop();
      right_most = helpptr;
      a_sum = right_most->w_nr - right_most->h_nr;
      while (!sequence.stack_empty())
      {
         helpptr = sequence.pop();
         a_sum = a_sum + helpptr->w_nr - helpptr->h_nr;
      }// while
      left_most = helpptr;
      if (old_a_sum < a_sum)
                                 // Update maximum sequence.
      {
         old_a_sum = a_sum;
         old_left_most = left_most;
         old_right_most = right_most;
      }// if
   }// if

                                  // Get the minimum of the two possible
                                  // a-numbers b1 and b2 as the
                                  // a-number of the Q-node
   b2 = sum_all_w - old_a_sum;
   if (b2 < b1)
   {
      node_ptr->a_nr = b2;
      node_ptr->h_child2 = old_left_most;
      node_ptr->a_child = NULL;
   }// if
   else
   {
      node_ptr->a_nr = b1;
      node_ptr->h_child2 = NULL;
      node_ptr->a_child = a_child;
   }// else
}// h_a_num_q_node
                       
         
         
/************************************************************************
                          a1_b1_number
************************************************************************/ 
      
int a1_b1_number(
                 pq_node *node_ptr,
                 pq_node **a_child
                )

   // Computes the difference of the sum over all w-numbers of all
   // pertinent children and max(w_i - a_i), i being a pertinent child.
   // This difference is needed for computing the a-number of a partial
   // P-node and for computing the a-number of a partial Q-node. 

{
   int      sum_max_a = 0;     // = max(w_i - a_i), i pertinent child
   int      sum_all_w = 0;     // = sum w_i, i pertinent child

   int      sum_help = 0;
   pq_node  *helpptr = NULL;
  

   
   node_ptr->full_children->start_at_bottom();
   node_ptr->partial_children->start_at_bottom();
   while (!node_ptr->full_children->read_last())
                               // Compute sum_all_w and sum_max_a
                               // by checking all full children.
   {
      helpptr = node_ptr->full_children->read_next();
      sum_all_w = sum_all_w + helpptr->w_nr;
      sum_help = helpptr->w_nr - helpptr->a_nr;
      if (sum_max_a < sum_help)
      {
         sum_max_a = sum_help;
         (*a_child) = helpptr;
      }// if
   }// while
   
   while (!node_ptr->partial_children->read_last())
                               // Compute sum_all_w and sum_max_a
                               // by checking all partial children.
   {
      helpptr = node_ptr->partial_children->read_next();
      sum_all_w = sum_all_w + helpptr->w_nr;
      sum_help = helpptr->w_nr - helpptr->a_nr;
      if (sum_max_a < sum_help)
      {
         sum_max_a = sum_help;
         (*a_child) = helpptr;
      }// if
   }// while
   return (sum_all_w - sum_max_a);
}// a1_b1_number



/************************************************************************
                          sum_pert_child
************************************************************************/ 

int sum_pert_child(
                   pq_node *node_ptr
                  )

   // Compute the w-number of node_ptr by summing the w-numbers of
   // all pertinent children.
{
   pq_node   *helpptr;
   int       sum = 0;

   node_ptr->full_children->start_at_bottom();
   node_ptr->partial_children->start_at_bottom();
   
   while (!node_ptr->full_children->read_last())
                                  // Compute the sum of all w-numbers
                                  // of the full children.
   {
      helpptr = node_ptr->full_children->read_next();
      sum = sum + helpptr->w_nr;
   }// while
   
   while (!node_ptr->partial_children->read_last())
                                  // Add to sum the sum of all w-numbers
                                  // of the partial children.
   {
      helpptr = node_ptr->partial_children->read_next();
      sum = sum + helpptr->w_nr;
   }// while
   return sum;
}// sum_pert_child










