//-------------------------------------------------------------------
//
//                     random_tric_graph.cpp
//
//    author: Carsten Gutwenger
//    date:   february 1997
//-------------------------------------------------------------------


#include <AGD/generators.h>
#include <AGD/impl/misc.h>

#include <LEDA/edge_map.h>
#include <LEDA/planar_map.h>


// the class Embedding is used to create temporarily the random
// graph (represented as a planar map)

class Embedding : public planar_map {
public:
	Embedding(graph& G);

	// some simple access methods
	face F_out();
	face left_face (edge e);
	face right_face (edge e);
	int  outv (face f);
	int  oute (face f);
	void outv (face f, int x);
	void oute (face f, int x);
	void inc (face f, int dv, int de);
	int  size (face f);

	edge get_face_edge (face f, int pos);
	void assign_graph (graph &G);
	edge my_split_edge (edge e, edge &e1);
	edge my_new_edge (edge e1, edge e2);
	edge new_chain (edge e1, edge e2, edge &e_ret);
	
	face get_face (int k);

private:
   face outer_face;
   face_map<int> the_outv, the_oute;
};

inline face Embedding::F_out()
{
	return outer_face; 
}
   
inline face Embedding::left_face (edge e)
{ 
	return adj_face(reverse(e));
}

inline face Embedding::right_face (edge e)
{
	return adj_face(e);
}

inline int Embedding::outv (face f)
{
	return the_outv[f];
}

inline int Embedding::oute (face f)
{
	return the_oute[f];
}

inline void Embedding::outv (face f, int x)
{
	the_outv[f] = x;
}

inline void Embedding::oute (face f, int x)
{
	the_oute[f] = x;
}

inline void Embedding::inc (face f, int dv, int de)
{
	the_outv[f] += dv;
	the_oute[f] += de;
}

inline int Embedding::size(face f)
{
	return adj_edges(f).length();
}

// we start with graph G (a simple polygon with n nodes)
Embedding::Embedding (graph& G) : planar_map(G)
{
   const int n = number_of_nodes();

   the_outv.init (*this);
   the_oute.init (*this);

   edge e = first_edge();
   outer_face = left_face(e);
   face f = right_face(e);

   outv (f, n);
   oute (f, n);
   outv (left_face(e), 0);
   oute (left_face(e), 0);
}

// returns the k'th face
face Embedding::get_face (int k)
{
	int i = 1;
	face f;

	k = k % number_of_faces();
	for (f = first_face(); i <= k; i++, f = succ_face(f)) ;

	return f;
}

// creates our initial graph: a polygon with n nodes
// the created graph represents a planar map
static void ini_graph (graph& G, int n)
{
	node v1, v, w;
	int i;

	v1 = G.new_node();
	for (v = v1, i = 2; i <= n; v = w, i++) {
		w = G.new_node();
		G.new_edge(v, w);
		G.new_edge(w, v);
	}
	G.new_edge(v, v1);
	G.new_edge(v1, v);
}

// returns pos'th edge of face f
edge Embedding::get_face_edge (face f, int pos)
{
	int i;
	edge e;

	for (i = 1, e = first_face_edge(f); i <= pos; i++, e = succ_face_edge(e))
		;
	return e;
}

// assigns our temporary graph to the original graph G
void Embedding::assign_graph (graph &G)
{
	node_array<node> map(*this);
	edge_array<bool> visited(*this, false);
	node v;
	edge e;

	G.clear();
	forall_nodes (v, *this)
		map[v] = G.new_node();
   
	forall_edges (e, *this)
		if (visited[reversal(e)] == false) {
			visited[e] = true;
			G.new_edge (map[source(e)], map[target(e)]);
		}
}

// an extension to split_edge, which updates outv(f) and oute(f)
edge Embedding::my_split_edge (edge e, edge &e1)
{
	edge e2 = split_edge (e);
	e1 = pred_face_edge(e2);

	if (left_face(e2) == outer_face) {
		inc (right_face(e2), 1, 1);
	} else if (right_face(e) == outer_face) {
		inc (right_face(e), 1, 1);
	}

	return e2;
}

// an extension to new_edge, which updates outv(f) and oute(f)
edge Embedding::my_new_edge (edge e1, edge e2)
{
	edge e, ei;
	face f1, f, f_old = right_face(e1);

	e = new_edge (e1, e2);
	if (f_old == outer_face) {
		outer_face = right_face(e);
		f1 = right_face(e1);
		forall_face_edges (ei, f1)
			if (ei != reverse(e)) {
				inc (left_face(ei), 0, -1);
				if (ei != e1) {
					forall_adj_faces (f, source(ei))
						if (f != f1)
							inc (f, -1, 0);
				}
			}
		outv (f1, 2);
		oute (f1, 1);
	}

	return e;
}

// my_new_edge + my_split_edge
edge Embedding::new_chain (edge e1, edge e2, edge &e_ret)
{
	edge e, et;

	e = my_new_edge (e1, e2);
	e_ret = my_split_edge (e, et);
	return et;
}

bool sep_pair(edge e2, list_item it, list<edge> &deg2, Embedding &E)
{
	edge e;
	face f;

	for (e = e2; e != deg2[it]; e = E.succ_face_edge(e)) {
		f = E.left_face(e);
		if (E.outv(f) > E.oute(f)+1)
			return true;
	}
	return false;
}

void rand_seq (array<int> &R, int k, int m)
{
	array<bool> A (1, m);
	int i, j, l, r;
	const int d = 2*m/k;

	for (i = 1; i <= m; i++)
		A[i] = false;

	j = 1;
	for (i = 1; i <= k; i++) {
		r = 1+rand_int (0, d-1);
		for (l = 1; l <= r; l++)
			do
				j = (j < m) ? j+1 : 1;
			while (A [j] == true) ;
		A [j] = true;
    }
    j = 1;
    for (i = 1; i <= m; i++)
		if (A [i] == true) R [j++] = i;
}

void random_tric_graph (graph &G, int n, int m)
{
	edge e, e1, e2, e3;
	list_item it, it2;
	int n_s, k, i, j, l;
	face f;
   
	array<int> A;
	array<edge> B;
   
	graph Gi;
	ini_graph (Gi, 3);

	Embedding E(Gi);
   
	list<edge> deg2;

	if (m < 3*(n+1)/2) m = 3*(n+1)/2;
	m *= 2;
	k = (15*n-m)/12;
	n_s = k + rand_int (0, n-k-1);

	int max_deg_2 = 10, p_expand = 60;
	edge_map<list_item> link (E, nil);
   
	forall_face_edges (e, E.F_out())
		link[e] = deg2.append(e);

	while (E.number_of_nodes() + deg2.length()-2 < n_s) {
		if (deg2.length() <= 3 || (deg2.length() < max_deg_2 &&
			rand_int(0,99) < p_expand))
		{
			e = E.get_face_edge (E.F_out(), rand_int (0,E.size(E.F_out())-1));
		e2 = E.my_split_edge (e, e1);
		if (link[e] != nil) {
			deg2 [link[e]] = e1; link [e1] = link [e];
		}
		for (e = E.pred_face_edge(e2); E.outdeg(E.source(e)) != 2; e = E.pred_face_edge(e)) ;
			link [e2] = deg2.insert (e2, link[e], after);
		} else {
			it = deg2.get_item (rand_int(0,deg2.length()-1));
			e1 = deg2 [it];
			it2 = deg2.cyclic_succ (it);
			e2 = deg2 [it2];
			e = E.new_chain (e1, e2, e3);
			link [e3] = deg2.insert (e3, it, after);
			link [e1] = nil; deg2.del_item (it);
			link [e2] = nil; deg2.del_item (it2);
		}
	}
   
	// fge Rand hinzu
	i = l = 0; f = nil;
	for (it = deg2.last(); it != nil && i < 2; l++, it = deg2.pred (it))
		if (f != E.left_face(deg2 [it])) {
			i++; f = E.left_face(deg2 [it]);
		}
	l = Max (l, 3);

	while (!deg2.empty()) {
		k = 1 + rand_int (0,Min (deg2.length()-2, 4)-1);
		i = j = 0; f = nil;
		for (it = deg2.first(); it != nil && (i < k+2 || j < 2);
			i++, it = deg2.succ(it))
		{
			e2 = deg2[it];
			if (f != E.left_face(e2)) {
				j++; f = E.left_face(e2);
			}
		}
		while (it != nil && sep_pair(e2, it, deg2, E)) {
			i++; e2 = deg2[it];
			it = deg2.succ(it);
		}
		if (deg2.length()-i < l) {
			i = deg2.length(); e2 = deg2.tail();
		}
		k = i-2;
		e1 = deg2.pop();
		e2 = deg2.pop();
		E.new_chain (e1, e2, e);
		for (i = 2; i <= k; i++) {
			e1 = e; e2 = deg2.pop();
			E.new_chain (e1, e2, e);
		}
		e2 = deg2.pop();
		E.my_new_edge (e, e2);
	}
   
	while (E.number_of_nodes() < n) {
		f = E.get_face (rand_int(0,E.number_of_faces()-1));

		l = E.size (f);
		k = 3 + rand_int (0,(l-3)/2-1);
		A = array<int> (1,k);
		B = array<edge> (1,k);
		rand_seq (A, k, l);
  
		e = E.first_face_edge (f); i = 1;
		for (j = 1; j <= l && i <= k; j++, e = E.succ_face_edge(e))
			if (j == A[i]) B [i++] = e;
		E.new_chain (B[1], B[2], e);
		for (i = 3; i <= k; i++)
			e = E.my_new_edge (e, B[i]);
	}
   
	list<face> deg4;
	forall_faces (f, E)
		if (E.size(f) >= 4) deg4.append (f);
	while (E.number_of_edges() < m && !deg4.empty()) {
		it = deg4.get_item (rand_int (0,deg4.length()-1));
		f = deg4.del_item (it);
		l = rand_int (0,E.size(f)/2-1);
		e1 = E.first_face_edge (f);
		for (i = 1; i <= l; i++)
			e1 = E.succ_face_edge (e1);
		l = 2+rand_int(0,E.size(f)-3-1);
		for (e2 = e1, i = 1; i <= l; i++)
			e2 = E.succ_face_edge (e2);
		e = E.my_new_edge (e1, e2);
		if (E.size (E.left_face(e)) >= 4)
			deg4.append (E.left_face(e));
		if (E.size (E.right_face(e)) >= 4)
			deg4.append (E.right_face(e));
	}
   
	E.assign_graph (G);
}
