//-------------------------------------------------------------------
//
//                        ranking.cpp
//
//    author: Carsten Gutwenger
//            Michael Krger
//    date:   february 1998
//-------------------------------------------------------------------


#include <AGD/TopSortRanking.h>
#include <AGD/HierarchicalRanking.h>
#include <AGD/DfsRanking.h>
#include <LEDA/queue.h>


//-------------------------------------------------------------------
//                        RankAssignment
//-------------------------------------------------------------------

const int RankAssignment::in_graph = 1;
const int RankAssignment::out_ranking = 1;

bool RankAssignment::check (const graph& G, string& message)
{
	return check_graph (in_graph, G, message);
}


//-------------------------------------------------------------------
//                       HierarchicalRanking
//-------------------------------------------------------------------

HierarchicalRanking::HierarchicalRanking(bool sdl) : _sep_deg0(sdl)
{
	add_precondition (key::no_self_loops);
	add_precondition (key::directed);
}


AGDModule *HierarchicalRanking::clone() const
{
	return new HierarchicalRanking(_sep_deg0);
}


void HierarchicalRanking::call (const graph& G, node_array<int>& rank)
{
	edge_array<bool> reversed(G,false);
	list<edge> R;

	Is_Acyclic(G,R);

	while(!R.empty())
		reversed[R.pop()] = true;

	is_source.init(G,true);
	adjacent.init(G);

	node v;
	edge e;

	forall_nodes(v,G) {
		forall_out_edges(e,v)
			if (!reversed[e]) {
				adjacent[v].append(target(e));
				is_source[target(e)] = false;
			}
		forall_in_edges(e,v)
			if (reversed[e]) {
				adjacent[v].append(source(e));
				is_source[source(e)] = false;
			}
	}

	finished.init(G,false);
	ingoing.init(G,0);

	int min = 0, max = 0;
	MAX_N = G.number_of_nodes();

	forall_nodes(v,G)
		if (is_source[v]) {
			dfs(v);
			get_tmp_rank(v, rank);
			dfs_add(v, rank);

			if (rank[v] < min) min = rank[v];
		}

	forall_nodes(v,G) {
		if ((rank[v] -= min) > max) max = rank[v];
	}

	if (max > 0 && separate_deg0_layer()) {
		max++;
		forall_nodes(v,G)
			if (G.degree(v) == 0) rank[v] = max;
	}

	is_source.init();
	adjacent.init();
	finished.init();
	ingoing.init();
}

void HierarchicalRanking::dfs(node v)
{
	node w;

	ingoing[v]++;
	if (ingoing[v] == 1 && !finished[v])
		forall (w,adjacent[v])
			dfs(w);
}

void HierarchicalRanking::get_tmp_rank(node v, node_array<int>& rank)
{
	list<node> N;
	node u, w;

	offset = MAX_N;
	N.append(v);
	rank[v] = 0;

	while (!N.empty()) {
		w = N.pop();

		forall(u,adjacent[w]) {
			ingoing[u]--;
			if (ingoing[u] == 0) {
				if (finished[u])
					offset = Min(offset, rank[u] - (rank[w]+1));
				else {
					N.append(u);
					rank[u] = rank[w]+1;
				}
			}
		}
	}
	if (offset == MAX_N) offset = 0;
}

void HierarchicalRanking::dfs_add(node v, node_array<int>& rank)
{
	node w;

	if (!finished[v]) {
		finished[v] = true;
		rank[v] += offset;

		forall(w,adjacent[v])
			dfs_add(w,rank);
	}
}


//-------------------------------------------------------------------
//                        TopSortRanking
//-------------------------------------------------------------------

TopSortRanking::TopSortRanking()
{
	add_precondition (key::no_self_loops);
	add_precondition (key::directed);
}


AGDModule *TopSortRanking::clone() const
{
	return new TopSortRanking;
}


void TopSortRanking::call (const graph& G, node_array<int>& rank)
{
	edge_array<bool> reversed(G,false);
	list<edge> R;

	Is_Acyclic(G,R);

	while(!R.empty())
		reversed[R.pop()] = true;

	node v;
	forall_nodes(v,G) rank[v] = -1;

	forall_nodes(v,G)
		compute_depth(v,reversed,rank);
}


int TopSortRanking::compute_depth (node v,
	const edge_array<bool>& reversed,
	node_array<int>& rank)
{
	if (rank[v] >= 0) return rank[v];

	int r, max = -1;
	edge e;

	forall_in_edges(e,v)
		if(!reversed[e] && (r = compute_depth(source(e),reversed,rank)) > max)
			max = r;

	forall_out_edges(e,v)
		if(reversed[e] && (r = compute_depth(target(e),reversed,rank)) > max)
			max = r;

	return (rank[v] = max+1);
}


//-------------------------------------------------------------------
//                           DfsRanking
//-------------------------------------------------------------------

DfsRanking::DfsRanking()
{
	add_precondition (key::directed);
}


AGDModule *DfsRanking::clone() const
{
	return new DfsRanking;
}

void DfsRanking::call (const graph& G, node_array<int>& rank)
{
	node v;

	forall_nodes(v,G)
		rank[v] = -1;

	forall_nodes(v,G)
		dfs(v,0,rank);
}

void DfsRanking::dfs (node v, int r, node_array<int>& rank)
{
	if (rank[v] >= 0) return;

	rank[v] = r++;

	edge e;
	forall_inout_edges(e,v)
		dfs(opposite(v,e),r,rank);
}


//-------------------------------------------------------------------
//                           CORanking
//-------------------------------------------------------------------
/*
#include <AGD/CORanking.h>
#include <AGD/CanonicalOrder.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/PlanAug.h>
#include <AGD/GraphCopy.h>

CORanking::CORanking()
{
	add_precondition (key::directed);
}


AGDModule *CORanking::clone() const
{
	return new CORanking;
}

void CORanking::call (const graph& G, node_array<int>& rank)
{
	GraphCopy GC(G);
	PlanarSubgraph S;
	PlanAug A;

	S.call(GC);
	A.call(GC);
	Make_Bidirected(GC);
	PLANAR(GC,true);
	GC.compute_faces();

	CanonicalOrder C(GC,co_biconnected,GC.first_face());
	node_array<int> r(GC);
	int highest = 0;

	const int m = C.length();

	int i, k;
	node v;

	node_array<node> next(G);
	for (i = 1; i <= C.len(1); i++) {
		next[C(1,i)] = (i == C.len(1)) ? nil : C(1,i+1);
		r[C(1,i)] = 0;
	}

	for (k = 2; k <= m; k++) {
		node last = next[C.right(k)];
		int max = 0;
		for (v = C.left(k); v != last; v = next[v]) {
			if (r[v] > max) max = r[v];
		}
		max++;
		if (max > highest) highest = max;

		for (i = 1; i <= C.len(k); i++) {
			r[C(k,i)] = max;
			next[C(k,i)] = (i == C.len(k)) ? C.right(k) : C(k,i+1);
		}
		next[C.left(k)] = C(k,1);
	}

	forall_nodes(v,GC)
		rank[GC.original(v)] = highest-r[v];
}

*/