//-------------------------------------------------------------------
//
//                        SugiyamaLayout.cpp
//
//    author: Carsten Gutwenger
//            Michael Krger
//    date:   february 1998
//-------------------------------------------------------------------


#include <AGD/SugiyamaLayout.h>
#include <AGD/HierarchicalRanking.h>
#include <AGD/BarycenterHeuristic.h>
#include <AGD/FastHierarchyLayout.h>

#include <AGD/Hierarchy.h>
#include <AGD/GraphCopyLI.h>
#include <float.h>


#ifdef _MSC_VER
ostream& operator<< (ostream& os, const list<node>& L);
istream& operator>> (istream& is, list<node>& L);
int compare (const list<node>&, const list<node>&);
#endif

const int HierarchyLayoutModule::in_graph = 1;
const int HierarchyLayoutModule::in_hierarchy = 1;
const int HierarchyLayoutModule::out_layout = 1;


//-------------------------------------------------------------------
//                        SugiyamaLayout
//-------------------------------------------------------------------

SugiyamaLayout::SugiyamaLayout() : LayoutModule(true), _runs(3)
{
	// precondition
	add_precondition (key::simple);
	add_precondition (key::no_self_loops);
	add_precondition (key::no_reversals);

	// initialize used modules
	array<KeySet> PRE_1(1,1), PRE_2(1,2), POST_1(0,1);

	PRE_1 [1] << key::simple << key::no_self_loops <<
		key::no_reversals << key::directed;
	ranking.init (PRE_1,POST_1,HierarchicalRanking());

	PRE_2 [1] = PRE_1[1]; 
	cross_min.init (PRE_2,POST_1,BarycenterHeuristic());

	comp_coord.init (PRE_2,POST_1,FastHierarchyLayout());
}


AGDModule *SugiyamaLayout::clone() const
{
	return new SugiyamaLayout;
}


void SugiyamaLayout::_call (GraphCopy& GC, LayoutInterface& A)
{
	// we work always on a directed graph
	if (GC.is_undirected()) GC.make_directed();

	// compute the ranking
	node_array<int> rank(GC.original());
	ranking.get().call(GC.original(),rank);

	// build proper hierarchy and minimize crossings
	Hierarchy H(GC,rank);
	minimize_crossings(GC,H);

	// assign final coordinates
	comp_coord.get().call(GC,H,A);
}

void SugiyamaLayout::ext_call (const graph& G,
	node_array<int>& rank, LayoutInterface& A)
{
	GraphCopy GC(G);
	GraphCopyLI A_GC(GC,A);

	if (GC.is_undirected()) GC.make_directed();

	// build proper hierarchy and minimize crossings
	Hierarchy H(GC,rank);
	minimize_crossings(GC,H);

	// assign final coordinates
	comp_coord.get().call(GC,H,A_GC);
	A_GC.update();
}

void SugiyamaLayout::reduce_crossings (GraphCopy& GC, Hierarchy& H)
{
	minimize_crossings(GC,H);
}


int SugiyamaLayout::traverse_top_down(const graph& G,
	Hierarchy& H, TwoLayerCrossMin& minimizer)
{
	int i;

	H.direction(Hierarchy::downward);
	for (i = 1; i < H.number_of_layers(); i++) {
		H.set_changeable_layer(i);
		minimizer.call(G,H[i]);
	}
	return H.calc_crossings();
}

int SugiyamaLayout::traverse_bottom_up(const graph& G,
	Hierarchy& H, TwoLayerCrossMin& minimizer)
{
	int i;

	H.direction(Hierarchy::upward);
	for (i = H.number_of_layers()-2; i >= 0; i--) {
		H.set_changeable_layer(i);
		minimizer.call(G,H[i]);
	}
	return H.calc_crossings();
}

void SugiyamaLayout::minimize_crossings(const graph& G, Hierarchy& H)
{
	best_nr_crossings = G.number_of_edges()*G.number_of_edges();

	if (G.number_of_edges() == 0) return;

	TwoLayerCrossMin &minimizer = cross_min.get();
	node_map<int> best_pos;

	minimizer.init(G,H);

	for (int i = 1; ;i++ ) {
		bool has_changed;
		int new_nr_crossings, old_nr_crossings;

		old_nr_crossings = H.calc_crossings();
		if (old_nr_crossings < best_nr_crossings) {
			H.store_pos (best_pos);
			best_nr_crossings = old_nr_crossings;
		}

		//minimizer.init(G,H);

		do {
			has_changed = false;

			new_nr_crossings = traverse_top_down(G,H,minimizer);

			if (new_nr_crossings < old_nr_crossings) {
				if (new_nr_crossings < best_nr_crossings) {
					H.store_pos(best_pos);
					best_nr_crossings = new_nr_crossings;
				}
				old_nr_crossings = new_nr_crossings;
				has_changed = true;
			}

			new_nr_crossings = traverse_bottom_up(G,H,minimizer);

			if (new_nr_crossings < old_nr_crossings) {
				if (new_nr_crossings < best_nr_crossings) {
					H.store_pos(best_pos);
					best_nr_crossings = new_nr_crossings;
				}
				old_nr_crossings = new_nr_crossings;
				has_changed = true;
			}

		} while (has_changed);

		if (best_nr_crossings == 0 || i >= runs())
			break;

		H.shuffle();
	}

	H.restore_pos(best_pos);
	minimizer.cleanup();
	cout << best_nr_crossings << " crossings\n";
	cout.flush();
}


//-------------------------------------------------------------------
//                            Layer
//-------------------------------------------------------------------

Layer::Layer (Hierarchy& H, int i, const list<node>& nodes)
{
	_hierarchy = &H;
	_index = i;
	_nodes = array<node>(nodes.length());

	node v;
	int j = 0;
	forall(v,nodes) {
		_nodes[j] = v;
		H._pos[v] = j++;
	}
}

const Layer& Layer::neighbored_layer() const
{
	return (_hierarchy->direction() == Hierarchy::downward) ?
		*_hierarchy->_layer[_index-1] :
		*_hierarchy->_layer[_index+1];
}


const list<node>& Layer::adj_nodes (node v) const
{
	Hierarchy::Direction d = _hierarchy->direction();
	int ch = _hierarchy->_changeable;

	return ((d == Hierarchy::upward && _index == ch) ||
		(d == Hierarchy::downward && _index != ch)) ?
		_hierarchy->_lower_adj_nodes[v] :
		_hierarchy->_upper_adj_nodes[v];
}

const list<edge>& Layer::adj_edges (node v) const
{
	Hierarchy::Direction d = _hierarchy->direction();
	int ch = _hierarchy->_changeable;

	return ((d == Hierarchy::upward && _index == ch) ||
		(d == Hierarchy::upward && _index != ch)) ?
		_hierarchy->_lower_adj_edges[v] :
		_hierarchy->_upper_adj_edges[v];
}

void Layer::swap (int i, int j)
{
	node v = _nodes[i];
	_hierarchy->_pos[_nodes[i] = _nodes[j]] = i;
	_hierarchy->_pos[_nodes[j] = v] = j;  
}

void Layer::quicksort (const node_array<double>& weight)
{
	if (length() == 0) return;

	_qs(0,length()-1,weight);

	int i;
	for (i = 0; i < length(); i++) {
		_hierarchy->_pos[_nodes[i]] = i;
	}
}

int Layer::cmp_qs (double w_v, int p_v, double w_w, int p_w)
{
	if ( w_w - w_v >= DBL_EPSILON ) return -1;
	if ( w_v - w_w >= DBL_EPSILON ) return 1;
	return p_v - p_w;
}
		 
void Layer::_qs (int l, int r, const node_array<double>& weight)
{
	const node_map<int>& P = _hierarchy->_pos;
	int    i = l, j = r, m = rs(l,r), xp = P[_nodes[m]];
	double x = weight[_nodes[m]];

	do {
		while (cmp_qs(weight[_nodes[i]],P[_nodes[i]],x,xp) < 0) i++;
		while (cmp_qs(weight[_nodes[j]],P[_nodes[j]],x,xp) > 0) j--;
		if (i <= j) {
			leda_swap(_nodes[i++],_nodes[j--]);
		}
	} while (i <= j);

	if (l < j) _qs(l,j,weight);
	if (i < r) _qs(i,r,weight);
}

void Layer::set_pos (const node_map<int>& position)
{
	int i;
	node v;
	array<node> N(length());

	for (i = 0; i < length(); i++) {
		v = _nodes[i];
		N[_hierarchy->_pos[v] = position[v]] = v;
	}

	_nodes = N;
}

void Layer::shuffle()
{
	_nodes.permute();

	for (int i = 0; i < length(); i++)
		_hierarchy->_pos[_nodes[i]] = i;
}

ostream& operator<< (ostream& os, const Layer& L)
{
	int i, l = L.length();

	os << "[ ";
	for (i = 0; i < l; i++) {
		os << index(L[i]);
		if (i < l-1) os << ",";
	}
	os << " ]";

	return os;
}

istream& operator>> (istream& is, Layer& L)
{
	return is;
}


//-------------------------------------------------------------------
//                            Hierarchy
//-------------------------------------------------------------------

Hierarchy::Hierarchy (GraphCopy& GC, const node_array<int>& R)
{
	gc_p = &GC;
	_direction = downward;

	_rank.init(GC);
	_pos.init(GC);

	node v, w;
	edge e, e_orig;;

	forall_nodes(v,GC)
		_rank[v] = R[GC.original(v)];

	int i, rv, rw;
	list<edge> edges = GC.all_edges();

	forall(e,edges) {
		e_orig = GC.original(e);
		v = source(e); w = target(e);
		rv = _rank[v]; rw =  _rank[w];

		if (rv > rw) {
			GC.rev_edge(e); leda_swap(rv,rw);
		}

		if (rv == rw) {
			error_handler(1,"Hierarchy::Hierarchy: rv == rw");
			//_rank[GC.split_edge(e,e1,e)] = rv+1;
			//_chain[e_orig].append(e1);
			//GC.rev_edge(e);

		} else {
			for (rv++; rv < rw; rv++) {
				_rank[source(e = GC.split(e))] = rv;
			}
		}
	}

	int max_rank = 0;
	forall_nodes(v,GC)
		if (_rank[v] > max_rank) max_rank = _rank[v];

	array<list<node> > L(0,max_rank);
	forall_nodes(v,GC)
		L[_rank[v]].append(v);

	_layer = array<Layer *>(0,max_rank);
	for (i = 0; i <= max_rank; i++)
		_layer[i] = new Layer(*this,i,L[i]);

	_lower_adj_edges.init(GC);
	_lower_adj_nodes.init(GC);
	_upper_adj_edges.init(GC);
	_upper_adj_nodes.init(GC);

	forall_nodes(v,GC) {
		forall_out_edges(e,v) {
			_lower_adj_edges[v].append(e);
			_lower_adj_nodes[v].append(target(e));
		}
		forall_in_edges(e,v) {
			_upper_adj_edges[v].append(e);
			_upper_adj_nodes[v].append(source(e));
		}
	}

	// initialize node arrays used in crossing calculation only once
	high_occ.init(GC);
	sorted_adj_nodes.init(GC);
}

Hierarchy::~Hierarchy()
{
	int i;

	for (i = 0; i < number_of_layers(); i++)
		delete _layer[i];
}

void Hierarchy::store_pos (node_map<int>& old_pos)
{
	old_pos = _pos;
}

void Hierarchy::restore_pos (const node_map<int>& new_pos)
{
	node v;

	_pos = new_pos;

	forall_nodes(v, *gc_p)
		_layer[_rank[v]]->_nodes[_pos[v]] = v;
}

// caculate crossings between layer i and i+1
int Hierarchy::calc_crossings (int i)
{
	if (length(i) <= 1 || length(i+1) <= 1)
		return 0;

	const graph& G = *gc_p;
	int j, k;
	int deactivated, nodecross,
		allcross, crossing_number = 0;
	node uk, ul;
	list<node> upper_lower_list[2];
	list_item li_uj;
	int index[2] = { 0, 0 };

	for (j = i; j <= i+1; j++) {
		for (k = 0; k < length(j); k++) {
			uk = (*_layer[j])[k];
			high_occ[uk] = nil;
			sorted_adj_nodes[uk].clear();
		}
	}

	for (j = i; j <= i+1; j++) {
		for (k = 0; k < length(j); k++) {
			uk = (*_layer[j])[k];
			const list<node>& N = (j == i) ? _lower_adj_nodes[uk] : _upper_adj_nodes[uk];
			forall (ul, N)
				sorted_adj_nodes[ul].append(uk);
		}
	}
  
	j = 0;
	do {
		deactivated = nodecross = allcross = 0;
		ul = (*_layer[i+j])[index[j]];

		if (high_occ[ul] != nil) {
			forall_items (li_uj, upper_lower_list[j]) {
				if (upper_lower_list[j].inf(li_uj) == ul) {
					deactivated++;
					allcross += nodecross;
					upper_lower_list[j].erase(li_uj); // deletion of current element allowed since LEDA 3.5.1!
				} else
					nodecross++;
				if (high_occ[ul] == li_uj)
					break; // loop until high_occ[ ul ] (incl.)
			}
			crossing_number +=
				deactivated * upper_lower_list[(j == 0) ? 1 : 0].size() + allcross;
		}
    
		forall (uk, sorted_adj_nodes[ul])
			if (pos(uk) >  pos(ul) ||
				pos(uk) == pos(ul) && j == 0)
				high_occ[uk] = upper_lower_list[(j == 0) ? 1 : 0].append(uk);

		index[j]++;
		if (index[1-j] < length(i+1-j))
			j = 1-j;
	} while (index[j] < length(i+j));

	return crossing_number;
}

int Hierarchy::calc_crossings()
{
	int i, sum = 0;

	for (i = 0; i < number_of_layers()-1; i++)
		sum += calc_crossings(i);

	return sum;
}

void Hierarchy::shuffle()
{
	for (int i = 0; i < number_of_layers(); i++)
		_layer[i]->shuffle();
}
