//-------------------------------------------------------------------
//
//                        MaximumPlanarSubgraph.c
//
//
// author: Sergey Fialko
// date:   march 1998
//-------------------------------------------------------------------

#include <AGD/MaximumPlanarSubgraph.h>
#include <AGD/PlanarSubgraph.h>
#include <AGD/impl/planarisator.h>
#include <LEDA/node_array.h>
#include <LEDA/node_map2.h>
#include <LEDA/ugraph.h>


//*******************************************************************
class kuratowski_edge {
 private:
    node source_;
    node target_;
    list<edge> edges_;
    list<edge> min_cut_;

 public:
    kuratowski_edge (node s, node t, edge e);
    kuratowski_edge (node, node, list<kuratowski_edge*>&);
    inline list<edge> edges () { return edges_; }
    inline edge first_edge () { return edges_.head (); }
    inline int width () { return min_cut_.size (); }
    inline int number_of_edges () { return edges_.size (); }
    list<edge> min_cut ()  { return min_cut_; }
    void add_parallel (kuratowski_edge*);
};
//******************************************************************
kuratowski_edge::kuratowski_edge (node s, node t, edge e) : 
		 source_ (s), target_ (t) {
  edges_.push (e);
  min_cut_.push (e);
}
//******************************************************************
kuratowski_edge::kuratowski_edge (node s, node t, list<kuratowski_edge*>& chain) :
			source_ (s), target_ (t)  {
  kuratowski_edge *ke;
  min_cut_ = chain.head()->min_cut ();
  forall (ke, chain)  {
	if (ke->width () < width ())
		min_cut_ = ke->min_cut ();
	list<edge> new_edges = ke->edges ();
	edges_.conc (new_edges);
  }
}
//******************************************************************
void kuratowski_edge::add_parallel (kuratowski_edge *ke)  {
  list<edge> new_cut = ke->min_cut ();
  min_cut_.conc (new_cut);
  new_cut = ke->edges ();
  edges_.conc (new_cut);
}
//*****************************************************************
struct e_info {
  node target;
  edge original;
  e_info (node n, edge e) : target (n), original (e) {}
};
//******************************************************************
void permute_graph (GRAPH<node, edge>& g) {
  int i = 0;
  
  for ( i = 0; i < g.number_of_nodes ()/2; i++) {
    node n = g.choose_node ();

    list<e_info*> in, out;

    edge e;

    forall_out_edges (e, n)
      out.push (new e_info (target (e), g.inf (e)));

    forall_in_edges (e, n) 
      in.push (new e_info (source (e), g.inf (e)));

    node save_info = g.inf (n);
    g.del_node (n);

    in.permute ();
    out.permute ();

    n = g.new_node (save_info);

    e_info *ei;

    forall (ei, in) {
      g.new_edge (ei->target, n, ei->original);
      delete ei;
    }

    forall (ei, out) {
      g.new_edge (n, ei->target, ei->original);
      delete ei;
    }
  }
}
//********************************************************************
//********************************************************************
MaximumPlanarSubgraph::MaximumPlanarSubgraph() : SubgraphModule(),
nHeuristicCalls_ (1)
{
	add_post_rule (key::planar);
	add_post_rule (key::simple, 1, key::simple);
	add_post_rule (key::no_self_loops, 1, key::no_self_loops);
	add_post_rule (key::no_reversals, 1, key::no_reversals);
}
//********************************************************************
void MaximumPlanarSubgraph::_call (const graph& G, list<edge>& UnplanarEdges)
{
  bctree BC (&G, true, false);
  block B;

  forall (B, BC.all_blocks ()) {
    list<node> nodes;
    node n;

// Try to sort out little blocks without processing
    forall (n, BC.all_nodes (B))
	if (degree (n) > 2) //n can be one of Kuratowski base nodes (n is not a chain node)
		nodes.push (n);

    if (nodes.size () + BC.degree (B) < 5 )
	continue;// B is little planar block

    nodes = BC.all_nodes (B);
    
    bc_node cp;
    forall (cp, BC.adj_nodes (B))
	nodes.push (first_node (cp));

// Make copy from B, test whether copy is planar, if not 
// simplify and planarize it
    UGRAPH<node, kuratowski_edge*> copy;

//************************************************************************
// Kopiere block b in copy.
//************************************************************************
    edge e, ed;
    node_array<node> o2copy (G, nil);

// first, copy all nodes from B to graph copy and save relations to original nodes
// in o2copy
    forall (n, nodes)
	o2copy[n] = copy.new_node (n);

//save connections between nodes of copy in this node_matrix
    node_matrix<kuratowski_edge*> k_edges (copy, nil);

// Copy edges which belong to B in original graph to copy graph (avoid possibly
// parallel edges and selfloops). Assign to each copied edge a new kuratowski_edge.
    forall (n, nodes)
	forall_out_edges (e, n) {
		node t = opposite (n, e);
		if ( nodes.search (t) && !k_edges (o2copy[n], o2copy[t]) &&
		    !k_edges (o2copy[t], o2copy[n]) && n != t)  {
			ed = copy.new_edge (o2copy [n], o2copy[t]);
			copy[ed] = new kuratowski_edge (n, t, e);
			k_edges (source (ed), target (ed)) = copy.inf (ed);
			k_edges (target (ed), source (ed)) = copy.inf (ed);
		}
	}
    if (PLANAR (copy))
        continue;

    int heurSol = 0;
    list<edge> sol;

    if ( nHeuristicCalls_ ) {
         PlanarSubgraph PS;
	 heurSol = MAXINT;
	 list<edge> tmp_sol;
	 GRAPH<node, edge> CopyOfcopy;
	 CopyGraph (CopyOfcopy, copy);
	 edge e;

         for ( int i = 1; i <= nHeuristicCalls_; i++) {
	    if (heurSol == 1 ) 
	       break;

	    PS.call (CopyOfcopy, tmp_sol);
	    if ( tmp_sol.size () < heurSol ) {
	        heurSol = tmp_sol.size ();
		sol.clear ();
		while (!tmp_sol.empty ())
		   sol.push (copy.inf ((CopyOfcopy.inf (tmp_sol.pop ())))->first_edge ());
	    }
	    float uT = used_time (); //debug
	    permute_graph (CopyOfcopy);
	    uT = used_time (uT);
	    tmp_sol.clear ();
//	    cout << i << "  " << uT << "\n";
	 }
    }

    node v;
   
       forall_nodes (v, copy) { //Simplify copy
	if ( degree (v) != 2 )
		continue;

	n = v;
	while (degree (n) == 2)  { // n is a chain node

		list<kuratowski_edge*> chain;
		edge e = copy.first_adj_edge (n);
		node t = opposite (n, e);
		chain.push (copy.inf (e));
		copy.del_edge (e);
// Short cut in the direction of first adj edge		
		while ( degree (t) == 1) {
		        e = copy.first_adj_edge (t);
			t =  opposite (t, e);
			chain.push (copy.inf (e));
			copy.del_edge (e);
		}
		node s = t; //Save first end node of the chain
		t = n;
//Short cut in direction of second adj edge
		while ( degree (t) == 1) {
		        e = copy.first_adj_edge (t);
			t =  opposite (t, e);
			chain.push (copy.inf (e));
			copy.del_edge (e);
		}

		kuratowski_edge *kur_e,*ke = new kuratowski_edge (s, t, chain);
		forall (kur_e, chain)
			delete kur_e;

		if ( k_edges (s, t) ) {//A (kuratowski_)edge exists already between s and t.
			k_edges (s, t)->add_parallel (ke);
			delete ke;
// It's possible, that after removing edges in the found chain the end nodes (s and t) are
// chain nodes now and have to be short cut once again.
			n = s;
			if (degree (s) == 2)  n = s;
			if (degree (t) == 2)  n = t;
		}
		else  {
			copy.new_edge (s, t, ke);
			k_edges (s, t) = ke;
			k_edges (t, s) = ke;
		}
	}
    }

    forall_nodes (n, copy) //Remove short cut nodes (they have degree 0).
	if (degree (n) == 0)
		copy.del_node (n);

      copy.make_directed ();

      PLANARISATOR opt (copy, heurSol);
      
      edge_array<int> weights (copy);
      forall_edges (e, copy)
	weights[e] = copy.inf(e)->width ();

      opt.set_weights (weights);
      opt.optimize ();

      list<node> solution;
      if ( opt.found ()) {
         solution = opt.solution ();
      
// Solution is a list of pairs of nodes. Source node is before target.
        while (!solution.empty ()) {
           node s = solution.pop (), t = solution.pop ();
           if ( !k_edges (s, t)) 
	      cerr << "k_edges - Error -- " << k_edges (s, t) << endl;
	   kuratowski_edge* ke = k_edges (s, t);
	   list<edge> cut = ke->min_cut ();
//	cout << ke->width () << " ";
	   while ( !cut.empty ())
		UnplanarEdges.push (cut.pop ());

        }
      }
      else
         UnplanarEdges = sol;//Heuristics solution
//      cout << "\n";
      forall_edges (e, copy) 
	delete (copy.inf (e));

 } //forall (B, BC.all_blocks ()) 
}
//***************************************************************************
AGDModule *MaximumPlanarSubgraph::clone() const 
{
  return new MaximumPlanarSubgraph;
}
